// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

namespace Twig
{
    //
    // Bowlby-esque Attachment behavioral system
    //
    partial class Mammal
    {
        /// <summary>
        /// Creature's estimate of the availability of caregiver when no caregiver in view.
        /// This is large for adults, small for children.
        /// </summary>
        public float generalCaretakerAvailability = 0.95f;

        /// <summary>
        /// Creature's primary caregiver.
        /// </summary>
        public Mammal primaryCareGiver;

        /// <summary>
        /// True if object is a caregiver of this creature.
        /// </summary>
        public virtual bool IsCareGiver(TwigObject o) {
            return o!=null && (o==primaryCareGiver || (o is Doll && IsHolding(o)));
        }

        AttachmentBehavior attachmentBehavior;
        class AttachmentBehavior : Behavior<ApproachParameters>
        {
            public override string Name
            {
                get
                {
                    return "Attach";
                }
            }
            Mammal me;
            public AttachmentBehavior(Mammal me)
            {
                this.me = me;
            }

            public float activationThreshold = 0.4f;
            public float deactivationThreshold = 0.1f;
            public bool activated = false;

            protected override void Update()
            {
                // Update Activation
                if (activated)
                {
                    Activation = 0.5f * (1f + me.Anxiety);
                    if (me.hugging && me.Anxiety < deactivationThreshold)
                        activated = false;
                }
                else
                {
                    Activation = 0;
                    if (me.primaryCareGiver!=null && me.Anxiety > activationThreshold)
                        activated = true;
                }

                // Update MotorOutput
                if (me.IsCareGiver(me.Attention.Focus))
                    MotorOutput.Approach(me.Attention.Focus, me.armLength/2f);

                if (MotorOutput.target != null)
                {
                    MotorOutput.reach = true;
                    me.WaitingFor("Attach: approach caregiver");
                    if (me.IsHolding(MotorOutput.target) || me.FacingToward(MotorOutput.target))
                        MotorOutput.hug = true;
                    else
                        me.WaitingFor("Attach: face caregiver");
                }
                else
                {
                    MotorOutput.reach = MotorOutput.hug = false;
                }
                //me.Chart("reach", MotorOutput.reach ? 1 : -1, 3f);
            }
        }

        /// <summary>
        /// Distance from caregiver at which child begins to become anxious.
        /// </summary>
        public float explorationRange = 5f;

        /// <summary>
        /// True if looking at caregiver.
        /// </summary>
        public bool SeeCareGiver
        {
            get
            {
                return IsCareGiver(gazeTarget);
            }
        }

        /// <summary>
        /// True of being looked at by caregiver
        /// </summary>
        public bool CareGiverWatching
        {
            get
            {
                return SeeCareGiver && TwigObject.LookingAt(gazeTarget, this);
            }
        }

        /// <summary>
        /// True if touching caregiver
        /// </summary>
        public bool TouchingCareGiver
        {
            get
            {
                return IsCareGiver(upperSpine.contactObject) || IsCareGiver(upperSpine.contactObject)
                        || IsCareGiver(head.contactObject) || IsCareGiver(arms.left.lower.contactObject) || IsCareGiver(arms.right.lower.contactObject)
                        || (hugging && (IsCareGiver(arms.left.GraspedObject) || IsCareGiver(arms.right.GraspedObject)));
            }
        }

        /// <summary>
        /// Fraction of explorationDistance child is from nearest caregiver.
        /// </summary>
        public float NormalizedCaregiverDistance
        {
            get
            {
                float distance = (primaryCareGiver==null)?9999f:(Position - primaryCareGiver.Position).Length();
                if (hugging && (IsCareGiver(arms.left.GraspedObject) || IsCareGiver(arms.right.GraspedObject)))
                    distance = Math.Min(distance, explorationRange / 5f);
                return distance / explorationRange;
            }
        }

        /// <summary>
        /// Overall level of security/anxiety produced by proximity/distance of caregiver
        /// </summary>
        public float AttachmentSecurity
        {
            get
            {
                float seeWeight = attachmentBehavior.activated ? 0.25f : 1f;
                return generalCaretakerAvailability+0.5f*Convert.ToSingle(TouchingCareGiver)
                    + (seeWeight*(Convert.ToInt32(SeeCareGiver) + Convert.ToInt32(CareGiverWatching))
                    + 5f*Math.Max(0f, 1f - NormalizedCaregiverDistance))*0.25f;
            }
        }
    }
}
