// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

#define ballSpin

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

#pragma warning disable 1591

namespace Twig
{
    /// <summary>
    /// Ball object played with by children.
    /// </summary>
    public class Ball : TwigObject
    {
        protected float radius = 0.25f;
        Node node;
#if ballSpin
        Quaternion orientation = Quaternion.Identity;
#endif
        SimpleMesh sphere;

        // For calling from the script
        public Ball(TwigGame g, string name, Vector3 position, Vector3 color)
            : this(g, name, position, new Color((byte)color.X, (byte)color.Y, (byte)color.Z)) { }

        public Ball(TwigGame g, string name, Vector3 position, Color c)
            : base(g, name, true)
        {
            node = new Node(this, name+" center");
            node.Position = position + new Vector3(0f, radius, 0f);
            node.Velocity = new Vector3(0, 0, 0);
            nodes = new Node[] { node };
            sphere = SimpleMesh.Sphere(g.Device, radius, 2, 0);
            Color = c;

#if DEBUG
            node.PositionUpdated += delegate(Vector3 newp)
            {
                Debug.Assert(newp.Y > 0, "Ball moved below ground");
            };
#endif

            collisionElements = new CollisionElement[] { new BallCollisionElement(this) };
        }

#if MDX
        public override Material DefaultMaterial
        {
            get
            {
                Material m = new Material();
                m.DiffuseColor = new ColorValue(255, 255, 0);
                m.AmbientColor = m.DiffuseColor;
                m.Specular = Color.White;
                m.SpecularSharpness = 0.1f;
                return m;
            }
        }
#endif

        public override Vector3 Position
        {
            get { 
                Vector3 p = node.Position;
                p.Y = 0;
                return p;
            }
        }

        public override float Mass
        {
            get { return 0.01f; }
        }

        public override Vector3 CenterOfMass
        {
            get { return Position; }
        }

        public override Vector3 LookPoint
        {
            get { return node.Position; }
        }
        public override void Constrain()
        {
            Vector3 p = node.Position;
            if (p.Y < radius+node.GroundLevel)
            {
                p.Y = radius+node.GroundLevel;
                node.Position = p;
            }

#if OldCollision
            foreach (TwigObject o in TwigObject.AllTwigObjects)
                o.TestCollisionWithBall(this);
#endif
        }

        public override void  Update(GameTime gameTime)
        {
            // Translate the node
            node.Update();
#if ballSpin
            // Apply spin to the orientation
            Vector3 v = node.Velocity;
            v.Y = 0;
            if (v.LengthSquared() > 0.01f)
            {
                Vector3 axis = Vector3.Cross(Vector3.Up, v);
                if (axis.LengthSquared() > 0)
                {
                    axis.Normalize();
                    orientation = Quaternion.Concatenate(orientation, Quaternion.CreateFromAxisAngle(axis, Physics.deltaT * v.Length() / radius));
                    orientation.Normalize();
                }
            }
#endif
            base.Update(gameTime);
        }

#if OldCollision
        public override bool TestCollisionWithRagdoll(Ragdoll h)
        {
            bool collision = false;
            foreach (Link l in h.Links)
            {
                collision |= TestCollisionWithLink(l);
            }
            return collision;
        }

        public bool TestCollisionWithLink(Link l)
        {
            Vector3 center = node.Position;
            Vector3 l1 = l.node1.Position, l2 = l.node2.Position;
            float realLinkLength = (l1 - l2).LengthSquared();
            Debug.Assert(realLinkLength != 0, "Link length is zero");
            float cpaParameter = Vector3.Dot(center-l1, l2 - l1) / realLinkLength;

            if (cpaParameter < 0 || cpaParameter > 1)
                return false;

            Vector3 cpa = l1 + cpaParameter * (l2 - l1);
            float minDistance = radius + l.radius;
            float minDistanceSq = minDistance * minDistance;
            Vector3 offset = center - cpa;
            if (offset.LengthSquared() < minDistanceSq) // || (cpa - l1).LengthSquared() < minDistanceSq
                //|| (cpa - l2).LengthSquared() < minDistanceSq)
            {
                // Collision
                float dist = offset.Length();
                offset *= (dist - minDistance) / Math.Max(dist, 0.01f);
                // Treat ball as massless so that it gets all the correction
                node.Position -= offset;
                // Correct velocity
                MathUtil.SafeNormalize(ref offset);
                node.Velocity -= 2 * Vector3.Dot(node.Velocity, offset) * offset;
                node.Velocity += 5 * (l.node1.Velocity + l.node2.Velocity); 

                return true;
            }
            return false;
        }
#endif

        public override void Draw(GameTime gameTime)
        {
            ConfigureMaterial();
#if ballSpin
            BasicEffect.World = Matrix.Multiply(Matrix.CreateFromQuaternion(orientation), Matrix.CreateTranslation(node.Position));
#else
            BasicEffect.World = Matrix.CreateTranslation(node.Position);
#endif
            BasicEffect.CommitChanges();
            BasicEffect.Begin();
            foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                sphere.Draw();
                pass.End();
            }
            BasicEffect.End();
        }

#if MDX
        static Mesh sphere;
        public static void PrepareDevice(Device d)
        {
            sphere = Mesh.Sphere(d, radius, 10, 10);
        }
#endif

        public override float CollisionRadius
        {
            get { return radius; }
        }

        /// <summary>
        /// Collision proxy for Balls
        /// </summary>
        [System.Diagnostics.DebuggerDisplay("{ball.name}")]
        class BallCollisionElement : CollisionElement
        {
            Ball ball;
            /// <summary>
            /// Make a new collision element representing a Ball TwigObject
            /// </summary>
            /// <param name="b"></param>
            public BallCollisionElement(Ball b) : base(CollisionElement.TypeCode.Sphere, b)
            {
                ball = b;
            }

            protected override GeometryPrimitives.Sphere SphereGeometry
            {
                get
                {
                    return new GeometryPrimitives.Sphere(ball.node.Position, ball.radius);
                }
            }

            protected override void HandleCollision(CollisionElement collider, Vector3 contactPoint, Vector3 contactNormal, float penetrationDepth)
            {
                ball.node.Position -= contactNormal * penetrationDepth;
                ball.node.Velocity += collider.Velocity;
                //Debug.Assert(ball.node.Velocity.LengthSquared() < 500, "Ball velocity is abnormally large after collision");
            }

            public override Vector3 Velocity
            {
                get { return ball.Velocity; }
            }

            public override float Mass
            {
                get { return 0.1f; }
            }

            public override float InvMass
            {
                get { return 10f; }
            }

            public override float PainMultiplier
            {
                get { return 1f; }
            }
        }
    }
}
