// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Twig
{
    /// <summary>
    /// This is an encapsulation of the old GRL behavior and behavior-max abstractions
    /// as abstract generic classes.
    ///
    /// Activation is the equivalent of activation-level and MotorOutput is motor-vector
    /// </summary>
    /// <typeparam name="T"></typeparam>
    [System.Diagnostics.DebuggerDisplay("Activation={Activation}")]
    public abstract class Behavior<T> where T : new()
    {
        /// <summary>
        /// Does nothing.
        /// </summary>
        public Behavior() { }

        /// <summary>
        /// Informative name.  Used only for debugging.
        /// </summary>
        public virtual string Name
        {
            get
            {
                return GetType().Name;
            }
        }

        /// <summary>
        /// Internal state variable for activation.
        /// </summary>
        protected float mActivation;
        /// <summary>
        /// Internal state variable for motor output.
        /// </summary>
        protected T mOutput = new T();
        /// <summary>
        /// Activation level of the behavior.
        /// 0 means completely inactive, 1 means completely active.
        /// NB: Reading this can cause behavior to update itself.
        /// </summary>
        public float Activation
        {
            get
            {
                MaybeUpdate();
                return mActivation;
            }
            set
            {
                mActivation = value;
            }
        }

        /// <summary>
        /// Motor output of behavior.  This is the control information to be used
        /// if the behavior is selected for activation.
        /// NB: Reading this can cause behavior to update itself.
        /// </summary>

        public T MotorOutput
        {
            get
            {
                MaybeUpdate();
                return mOutput;
            }
            set
            {
                mOutput = value;
            }
        }

        /// <summary>
        /// Recomputes Activation and/or MotorOutput of behavior.
        /// </summary>
        protected abstract void Update();

        int lastUpdateFrame;
        void MaybeUpdate()
        {
            if (lastUpdateFrame!=Physics.tickCount)
            {
                lastUpdateFrame = Physics.tickCount;
                Update();
            }
        }
    }

#if DelegateBehaviors
    public class DelegateBehavior<T> : Behavior<T> where T : new()
    {
        public delegate void Updater(ref float activation, ref T motorOutput);
        public DelegateBehavior(string name, Updater updater)
        {
            this.name = name;
            this.updater = updater;
        }

        string name;
        Updater updater;

        protected override void Update()
        {
            updater(ref mActivation, ref mOutput);
        }
    }
#endif

    /// <summary>
    /// Implements Winner-Take-All behavior selection between a
    /// set of component behaviors.  Activation level and motor
    /// output are the activation and output of whichever child
    /// behavior is most active.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    [System.Diagnostics.DebuggerDisplay("Selected={maxBehavior}")]
    public class MaxBehavior<T> : Behavior<T> where T : new()
    {
        /// <summary>
        /// Child behaviors.
        /// </summary>
        public readonly List<Behavior<T>> behaviors = new List<Behavior<T>>();
        /// <summary>
        /// Identify of the child behavior with maximum activation.
        /// This is made a member primarilty for debugging purposes.
        /// </summary>
        public Behavior<T> maxBehavior;
        double switchTime;
        double switchTimeout;

        /// <summary>
        /// Create a winner-take-all selector with the specified child behaviors.
        /// </summary>
        public MaxBehavior(float switchTime, Behavior<T>[] behaviors)
        {
            this.switchTime = switchTime;
            this.behaviors.AddRange(behaviors);
        }

        /// <summary>
        /// Add a new behavior to the list of behaviors being checked.
        /// </summary>
        /// <param name="behavior"></param>
        public void Add(Behavior<T> behavior)
        {
            this.behaviors.Add(behavior);
        }

        /// <summary>
        /// Recomputes MotorOutput and Activation
        /// </summary>
        protected override void Update()
        {
            if (Physics.time >= switchTimeout)
            {
                Behavior<T> newMaxBehavior = behaviors[0];
                for (int i = 1; i < behaviors.Count; i++)
                {
                    Behavior<T> b = behaviors[i];
                    if (b.Activation > newMaxBehavior.Activation)
                        newMaxBehavior = b;
                }
                if (newMaxBehavior != maxBehavior)
                {
                    // Switch
                    maxBehavior = newMaxBehavior;
                    switchTimeout = Physics.time + switchTime;
                }
            }
            Activation = maxBehavior.Activation;
            MotorOutput = maxBehavior.MotorOutput;
        }
    }

    /// <summary>
    /// Simple timer object.  Activate field remains true for specified amount of time.
    /// </summary>
    public class Fuse
    {
        double expiration;
        /// <summary>
        /// Creates a simple timer.
        /// </summary>
        /// <param name="time">Period of time for Active to remain true.</param>
        public void Fire(float time) {
            expiration = Physics.time+time;
        }

        /// <summary>
        /// True if fuse hasn't yet run out.
        /// </summary>
        public bool Active
        {
            get
            {
                return Physics.time <= expiration;
            }
        }
    }
}
