// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A simple box
    /// </summary>
    public class BoxModel : TwigObjectWithModel
    {
        #region Basic instance variables
        Link[] links;
        float collisionRadius;
        /// <summary>
        /// Mass of the object
        /// </summary>
        protected float mass;
        /// <summary>
        /// Dimensions of object as computed from bounding box
        /// </summary>
        protected float width, height, depth;

        /// <summary>
        /// Offset of what it thinks of as the center of the object
        /// </summary>
        protected Vector3 centerOffset = Vector3.Zero;
        #endregion

        #region Constructors
        /// <summary>
        /// Makes a new box with the specified model.
        /// </summary>
        public BoxModel(TwigGame g, string name, string modelName, Vector3 center, bool centerOnBottom)
            : this(g, name, g.Content.Load<Model>(modelName), center, 0, 0, 0, true, centerOnBottom)
        { }

        /// <summary>
        /// Makes a new box with the specified model.
        /// </summary>
        public BoxModel(TwigGame g, string name, Model m, Vector3 center, bool centerOnBottom)
            : this(g, name, m, center, 0, 0, 0, true, centerOnBottom)
        { }

        /// <summary>
        /// Makes a new box with the specified model, if height is non-zero, it overrides the height of the bounding-box.
        /// </summary>
        public BoxModel(TwigGame g, string name, Model m, float height, Vector3 center, bool centerOnBottom)
            : this(g, name, m, center, 0, height, 0, true, centerOnBottom)
        { }

        /// <summary>
        /// Makes a new box
        /// </summary>
        public BoxModel(TwigGame g, string name, Model m, Vector3 center, float width, float height, float depth, bool centerOnBottom)
            : this(g, name, m, center, width, height, depth, false, centerOnBottom)
        { }

        BoxModel(TwigGame g, string name, Model m, Vector3 center, float width, float height, float depth, bool compute, bool centerOnBottom)
            : base(g, name, m, true)
        {
            float realHeight = height;
            if (compute)
            {
                GeometryPrimitives.AABB bb = m.BoundingBox();
                width = bb.Width;
                realHeight = bb.Height;
                if (height==0)
                    height = bb.Height;
                depth = bb.Depth;
                centerOffset = bb.Center;
            }
            collisionRadius = Math.Max(width, Math.Max(height, depth));

            Vector3 X = Vector3.UnitX*(width/2);
            Vector3 Y = Vector3.UnitY*(height/2);
            Vector3 Z = Vector3.UnitZ*(depth/2);
            if (centerOnBottom)
            {
                Vector3 offset = Vector3.UnitY * (realHeight / 2);
                center += offset;
                centerOffset -= offset;
            }

            mass = width * height * depth * Physics.WaterDensity;
            //mass = Single.PositiveInfinity;

            this.width = width;
            this.height = height;
            this.depth = depth;

            nodes = new Node[8] {
                new Node(this, name+"---", center-X-Y-Z),  // 0
                new Node(this, name+"--+", center-X-Y+Z),  // 1
                new Node(this, name+"-+-", center-X+Y-Z),  // 2
                new Node(this, name+"-++", center-X+Y+Z),  // 3
                new Node(this, name+"+--", center+X-Y-Z),  // 4
                new Node(this, name+"+-+", center+X-Y+Z),  // 5
                new Node(this, name+"++-", center+X+Y-Z),  // 6
                new Node(this, name+"+++", center+X+Y+Z)   // 7
            };
            links = new Link[16] {
                // X axis
                new Link("name", this, nodes[0], nodes[4], width, 0),
                new Link("name", this, nodes[1], nodes[5], width, 0),
                new Link("name", this, nodes[2], nodes[6], width, 0),
                new Link("name", this, nodes[3], nodes[7], width, 0),
                // Y axis
                new Link("name", this, nodes[0], nodes[2], height, 0),
                new Link("name", this, nodes[1], nodes[3], height, 0),
                new Link("name", this, nodes[4], nodes[6], height, 0),
                new Link("name", this, nodes[5], nodes[7], height, 0),
                // Z axis
                new Link("name", this, nodes[0], nodes[1], depth, 0),
                new Link("name", this, nodes[2], nodes[3], depth, 0),
                new Link("name", this, nodes[4], nodes[5], depth, 0),
                new Link("name", this, nodes[6], nodes[7], depth, 0),
                // Crosspieces
                new Link("name", this, nodes[0], nodes[7], (float)Math.Sqrt(width*width+height*height+depth*depth), 0),
                new Link("name", this, nodes[1], nodes[6], (float)Math.Sqrt(width*width+height*height+depth*depth), 0),
                new Link("name", this, nodes[2], nodes[5], (float)Math.Sqrt(width*width+height*height+depth*depth), 0),
                new Link("name", this, nodes[3], nodes[4], (float)Math.Sqrt(width*width+height*height+depth*depth), 0)
            };
            foreach (Node n in nodes)
                n.Mass = mass / 8;

            collisionElements = new CollisionElement[] { new BoxCollisionElement(this) };

            //targetCenter = Position;
            //targetNormal = FacingDirection;
            Atlas["top"] = new TopChart(this);
        }
        #endregion

        #region Geometric and physical properties
        /// <summary>
        /// Change of basis matrix from model coordinates to global coordinates
        /// </summary>
        public override Matrix WorldMatrix
        {
            get {
                return Matrix.CreateTranslation(-centerOffset) * MathUtil.ChangeOfBasis(LocalX, LocalY, LocalZ, Vector3.Zero) * Matrix.CreateTranslation(Position);
            }
        }
        
        /// <summary>
        /// Global coordinates of local X axis
        /// </summary>
        protected Vector3 LocalX
        {
            get
            {
                Vector3 v = nodes[4].Position - nodes[0].Position;
                v.Normalize();
                return v;
            }
        }

        /// <summary>
        /// Global coordinates of local Y axis
        /// </summary>
        protected virtual Vector3 LocalY
        {
            get
            {
                Vector3 v = nodes[2].Position - nodes[0].Position;
                v.Normalize();
                return v;
            }
        }

        /// <summary>
        /// Global coordinates of local Z axis
        /// </summary>
        protected Vector3 LocalZ
        {
            get
            {
                Vector3 v = nodes[1].Position - nodes[0].Position;
                v.Normalize();
                return v;
            }
        }


        /// <summary>
        /// Returns the position of the center of the box.
        /// </summary>
        public override Vector3 Position
        {
            get { return 0.5f*(nodes[0].Position+nodes[7].Position); }
        }

        /// <summary>
        /// Returns the position of the center of the bottom of the box.
        /// </summary>
        public Vector3 BottomPosition
        {
            get { return 0.5f * (nodes[0].Position + nodes[5].Position); }
        }

        /// <summary>
        /// Returns the position of the center of the top of the box.
        /// </summary>
        public Vector3 TopPosition
        {
            get { return 0.5f * (nodes[2].Position + nodes[7].Position); }
        }


        /// <summary>
        /// Center of mass of the box (same as its geometric center)
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get { return Position; }
        }

        /// <summary>
        /// Point to stare at on the box, in this case, its center
        /// </summary>
        public override Vector3 LookPoint
        {
            get { return Position; }
        }

        /// <summary>
        /// The Box's mass
        /// </summary>
        public override float Mass
        {
            get { return mass; }
        }

        /// <summary>
        /// radius of a bounding circle around the box's projection to the ground plane.
        /// </summary>
        public override float CollisionRadius
        {
            get { return collisionRadius; }
        }
        #endregion

        #region Charts
        class TopChart : PlanarChart
        {
            BoxModel me;
            public TopChart(BoxModel me)
            {
                this.me = me;
            }

            public override Matrix GlobalTransform
            {
                get { return Matrix.CreateTranslation(new Vector3(0, 0.5f * me.height, 0)) * me.WorldMatrix; }
            }
            public override Vector3 Normal(Vector3 local)
            {
                return me.LocalY;
            }
        }
        #endregion

        #region Update and control
        /// <summary>
        /// Update node positions.
        /// </summary>
        public override void Update(GameTime gameTime)
        {
            //nodes[0].Trace = TraceMode.Position;
            Control();
            foreach (Node n in nodes)
                n.Update();
        }

        /// <summary>
        /// Constrain all links
        /// </summary>
        public override void Constrain()
        {
            if (IsGrasped && orientationFace == null)
            {
                Mammal g = Grasper as Mammal;
                if (g != null)
                {
                    // Kluge: constrain all nodes to be "past" the legs along the transverse axis of the body.
                    if (g.arms.left.IsGripping(this))
                    {
                        float projection = Math.Min(Vector3.Dot(g.legs.left.root.Position, g.bodyRight), Vector3.Dot(g.legs.left.end.Position, g.bodyRight))
                                             - g.legs.left.upper.radius;
                        foreach (Node n in nodes)
                            n.ConstrainLE(g.bodyRight, projection);
                    }
                    else
                    {
                        float projection = Math.Max(Vector3.Dot(g.legs.right.root.Position, g.bodyRight), Vector3.Dot(g.legs.right.end.Position, g.bodyRight))
                                                + g.legs.right.upper.radius;
                        foreach (Node n in nodes)
                            n.ConstrainGE(g.bodyRight, projection);
                    }
                }
            }
            foreach (Link l in links)
                l.Constrain();
        }

        /// <summary>
        /// The limb currently grasping this object
        /// </summary>
        public override Limb GraspingLimb
        {
            set
            {
                orientationFace = null;
                base.GraspingLimb = value;
            }
        }

        /// <summary>
        /// Unit normal of the top face of the paper.
        /// </summary>
        Vector3 FacingDirection
        {
            get
            {
                Vector3 normal = Vector3.Cross(LocalZ, LocalX);
                normal.Normalize();
                return normal;
            }
        }

#if OLDPOSITIONING
        float centerGain = 1f;
        float orientationGain = 1f;
#endif

        Chart positionTargetChart;
        Vector3 positionTarget;
        Chart orientationFace;
        Node normalTarget;
        Vector3 normalInChart;

        /// <summary>
        /// Tell BoxModel to move its center to the specified position within the specified coordinate frame
        /// and to orient its specified face toward the specified target.
        /// </summary>
        public virtual void Place(Chart positionChart, Vector3 position, string faceName, Node normalTarget)
        {
            positionTargetChart = positionChart;
            if (positionChart == null)
            {
                orientationFace = null;
            }
            else
            {
                positionTarget = position;
                orientationFace = Atlas[faceName];
                this.normalTarget = normalTarget;
            }
        }

        /// <summary>
        /// Tell BoxModel to move its center to the specified position within the specified coordinate frame
        /// and to orient its specified face in the specified direction relative to the chart.
        /// </summary>
        public virtual void Place(Chart positionChart, Vector3 position, string faceName, Vector3 normalInChart)
        {
            this.normalInChart = normalInChart;
            Place(positionChart, position, faceName, null);
        }

        /// <summary>
        /// Adjust position and orientation of paper
        /// </summary>
        public virtual void Control()
        {
            if (!IsGrasped)
                orientationFace = null;

            if (orientationFace == null)
            {
                foreach (Node n in nodes)
                    n.Acceleration = Physics.gVector;
            }
            else
            {
                System.Diagnostics.Debug.Assert(orientationFace != null, "orientationFace null");

#if OLDPOSITIONING
            Vector3 center = Position;
            Vector3 facing = FacingDirection;
            Vector3 targetNormal = normalTarget.Position - Position;
            MathUtil.SafeNormalize(ref targetNormal);

            Vector3 orientationError = targetNormal - facing;
            Vector3 targetPosition = positionTargetChart.ToGlobal(positionTarget);
            Vector3 centerError = targetPosition - center;
            Vector3 centerForce = centerError * centerGain;

            // Figure out which way to point the top of the paper
            Mammal grasper = Grasper as Mammal;
            Vector3 grasperUp = (grasper != null) ? grasper.bodyUp : Vector3.Up;
            Vector3 up = Vector3.Cross(targetNormal, Vector3.Cross(grasperUp, targetNormal));
            MathUtil.SafeNormalize(ref up);
            float upError = Vector3.Dot(up, LocalZ);   // These should be perpendicular, if the roll is correct

            for (int i = 0; i < 8; i++)
            {
                Node n = nodes[i];
                Vector3 offset = n.Position - center;
                n.Position = targetPosition + offset;
                offset.Normalize();
                Vector3 facingTorque = orientationGain * Vector3.Dot(offset, orientationError) * facing;
                Vector3 rollTorque = 3*orientationGain * upError * Vector3.Cross(offset, Vector3.Cross(offset, targetNormal));
                n.Velocity = facingTorque + rollTorque;
            }
#else
                // Figure out the target center
                Vector3 targetPosition = positionTargetChart.ToGlobal(positionTarget);
                // Target Y axis
                Vector3 targetNormal = (normalTarget==null)?
                    positionTargetChart.ToGlobalDirection(normalInChart) :
                    normalTarget.Position - Position;
                MathUtil.SafeNormalize(ref targetNormal);
                // Target X axis
                Mammal grasper = Grasper as Mammal;
                Vector3 grasperUp = (grasper == null) ? Vector3.Up : grasper.bodyUp;
                Vector3 up = Vector3.Cross(targetNormal, Vector3.Cross(grasperUp, targetNormal));
                MathUtil.SafeNormalize(ref up);

                Vector3 targetZ = Vector3.Cross(up, targetNormal);

                // Now move all the nodes.  For this version, we just teleport them to the right places
                for (int i = 0; i < 8; i++)
                {
                    nodes[i].Position = NodePosition(i, targetPosition, up, targetNormal, targetZ);
                    nodes[i].Velocity = Vector3.Zero;
                }
#endif
                // Not necessary for a plain box, but needed for things like signs before we snap the hand to the grasp point.
                Constrain();

                if (IsGrasped)
                    GraspingLimb.SnapToGraspedNode();
            }
        }

        Vector3 NodePosition(int nodeNumber, Vector3 center, Vector3 localX, Vector3 localY, Vector3 localZ)
        {
            return center
                + (((nodeNumber & 4) == 0) ? -0.5f : 0.5f) * width * localX
                + (((nodeNumber & 2) == 0) ? -0.5f : 0.5f) * height * localY
                + (((nodeNumber & 1) == 0) ? -0.5f : 0.5f) * depth * localZ;
        }
        #endregion

        #region Drawing
        /// <summary>
        /// Controls display of corner markers
        /// </summary>
        [DebugSwitch("Display corners of BoxModels")]
        public static bool displayCorners = false;

        /// <summary>
        /// Draws the box.
        /// </summary>
        public override void Draw(GameTime gameTime)
        {
            if (displayCorners)
            {
                foreach (Node n in nodes)
                    TwigGame.DrawDebugMarker(n.Position, Color.Yellow);
            }
            //TwigGame.DrawDebugMarker(BottomPosition, Color.Yellow);
            base.Draw(gameTime);
        }
        #endregion
        
        #region Collision detection
        class BoxCollisionElement : CollisionElement
        {
            BoxModel me;
            public BoxCollisionElement(BoxModel me) : base(TypeCode.OBB, me)
            {
                this.me = me;
            }

            public override float Mass
            {
                get { return me.mass; }
            }

            public override float InvMass
            {
                get { return 1/me.mass; }
            }

            protected override GeometryPrimitives.OBB OBBGeometry
            {
                get
                {
                    return new GeometryPrimitives.OBB(me.Position,
                        me.LocalX, me.LocalY, me.LocalZ,
                        me.width * 0.5f, me.height * 0.5f, me.depth * 0.5f);
                }
            }

            public override Vector3 Velocity
            {
                get { return me.Velocity; }
            }

            public override float PainMultiplier
            {
                get { return 1; }
            }

            protected override void HandleCollision(CollisionElement collider, Vector3 contactPoint, Vector3 contactNormal, float penetrationDepth)
            {
                penetrationDepth = Math.Min(me.width, Math.Min(me.height, Math.Min(me.depth, penetrationDepth)));
                Vector3 offset = contactNormal * penetrationDepth * InvMass / (InvMass + collider.InvMass) ;
                foreach (Node n in me.nodes)
                    n.Position += offset;

#if KnockDown
                Person victim = collider.TwigObject as Person;
                if (victim != null)
                {
                    victim.standUp = false;
                    victim.sitUp = false;
                }
#endif
            }
        }
        #endregion
    }
}
