// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Twig
{
    /// <summary>
    /// Represents a local coordinate system (potentially one of many)
    /// on the surface of a TwigObject that implements IManifold.
    /// </summary>
    public abstract class Chart
    {
        /// <summary>
        /// Translates from local coordinates to world coordinates.
        /// </summary>
        public abstract Vector3 ToGlobal(Vector3 local);

        /// <summary>
        /// Convert a direction vector from local to global coordinates
        /// </summary>
        public Vector3 ToGlobalDirection(Vector3 local)
        {
            // Ugly, but it works.
            return ToGlobal(local) - ToGlobal(Vector3.Zero);
        }
        /// <summary>
        /// Translates from world coordinates to local.
        /// </summary>
        public abstract Vector3 ToLocal(Vector3 global);

        /// <summary>
        /// Surface normal at specified point, if defined.
        /// </summary>
        public abstract Vector3 Normal(Vector3 local);
    }

    /// <summary>
    /// A chart for a planar surface.
    /// </summary>
    public abstract class PlanarChart : Chart
    {
        /// <summary>
        /// Transformation matrix from local to global coordinates.
        /// </summary>
        public abstract Matrix GlobalTransform { get; }

        /// <summary>
        /// Transformation matrix from global to local coordinates.
        /// </summary>
        public virtual Matrix LocalTransform
        {
            get { return Matrix.Invert(GlobalTransform); }
        }

        /// <summary>
        /// Translates from local coordinates to world coordinates.
        /// </summary>
        public override Vector3 ToGlobal(Vector3 local)
        {
            return Vector3.Transform(local, GlobalTransform);
        }

        /// <summary>
        /// Translates from world coordinates to local.
        /// </summary>
        public override Vector3 ToLocal(Vector3 global)
        {
            return Vector3.Transform(global, LocalTransform);
        }
    }
}
