// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Twig
{
    /// <summary>
    /// Implements graphing of character state variables as a HUD
    /// </summary>
    public class ChartRecorder : HUD
    {
        const int chartLength = 780;
        const int chartSections = 2;
        const int chartChannels = 5;
        const int totalCharts = chartChannels * chartSections;
        const int chartZero = 65;
        const int chartMiddle = 130;
        string[] chartLabels = new string[totalCharts];
        float[] chartScales = new float[totalCharts];

        VertexPositionColor[][] charts = new VertexPositionColor[totalCharts][];
        VertexPositionColor[][] upperShadow = new VertexPositionColor[totalCharts][];
        VertexPositionColor[][] lowerShadow = new VertexPositionColor[totalCharts][];

        Color[] channelColors = new Color[chartChannels];
        int annotationChart = -1;
        int annotationTimepoint;

        /// <summary>
        /// Creates a chart recorder HUD
        /// </summary>
        /// <param name="g"></param>
        public ChartRecorder(TwigGame g) : base(g)
        {
            Init(g);
        }

        /// <summary>
        /// Creates a chart recorder HUD
        /// </summary>
        public ChartRecorder(TwigGame g, Keys hotKey, bool display)
            : base(g, hotKey, "Toggle display of chart recorder", display)
        {
            Init(g);
        }

        void Init(TwigGame g)
        {
            TwigGame = g;
            g.ChartRecorder = this;
            channelColors[0] = Color.Red;
            channelColors[1] = Color.Green;
            channelColors[2] = Color.Blue;
            channelColors[3] = Color.Purple;
            channelColors[4] = Color.Yellow;
            for (int channel = 0; channel < totalCharts; channel++)
            {
                VertexPositionColor[] chart = new VertexPositionColor[chartLength];
                charts[channel] = chart;
                VertexPositionColor[] upper = new VertexPositionColor[chartLength];
                upperShadow[channel] = upper;
                VertexPositionColor[] lower = new VertexPositionColor[chartLength];
                lowerShadow[channel] = lower;

                int zero = chartZero + chartMiddle * (channel % chartChannels);
                for (int t = 0; t < chartLength; t++)
                {
                    chart[t] = new VertexPositionColor(new Vector3(t, zero, 0), channelColors[channel % chartChannels]);
                    upper[t] = new VertexPositionColor(new Vector3(t, zero - 1, 0), Color.Black);
                    lower[t] = new VertexPositionColor(new Vector3(t, zero + 1, 0), Color.Black);
                }
            }
        }

        /// <summary>
        /// Updates chart recorder channel
        /// </summary>
        public void Chart(int section, string label, float v, float maxvalue)
        {
            int channel = Array.IndexOf<string>(chartLabels, label, section * chartChannels, chartChannels);
            if (channel < 0)
            {
                channel = Array.IndexOf<string>(chartLabels, null, section * chartChannels, chartChannels);
                chartLabels[channel] = label;
                chartScales[channel] = maxvalue;
            }
            System.Diagnostics.Debug.Assert(channel >= 0);

            float y = ChartToYCoordinate(section, v, maxvalue);
            int x = Physics.tickCount % chartLength;
            charts[channel][x].Position.Y = y;
            upperShadow[channel][x].Position.Y = y - 1;
            lowerShadow[channel][x].Position.Y = y + 1;
        }

        float ChartToYCoordinate(int section, float value, float maxvalue)
        {
            return section * chartMiddle + chartZero - value * chartZero / maxvalue;
        }

        float YCoordinateToChart(float chartY, int chart)
        {
            return (chartMiddle * (chart / chartChannels) + chartZero - chartY) * chartScales[chart] / chartZero;
        }

        /// <summary>
        /// Draws the chart recorder content in the HUD
        /// </summary>
        public override void Draw(GameTime gameTime)
        {
            HUDEffect.Begin();
            foreach (EffectPass pass in HUDEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                HUDSprites.Begin();
                CheckChartAnnotation();
                HUDEffect.DiffuseColor = Color.Black.ToVector3();
                HUDEffect.CommitChanges();
                for (int chart = 0; chart < totalCharts; chart++)
                {
                    Device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineStrip, upperShadow[chart], 0, chartLength - 1);
                    Device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineStrip, lowerShadow[chart], 0, chartLength - 1);
                }

                for (int chart = 0; chart < totalCharts; chart++)
                {
                    HUDEffect.DiffuseColor = channelColors[chart % chartChannels].ToVector3();
                    HUDEffect.CommitChanges();
                    Device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineStrip, charts[chart], 0, chartLength - 1);
                }

                // Draw time mark
                DrawLine(Physics.tickCount % chartLength, 0, Physics.tickCount % chartLength, 300, Color.White);
                // Draw X axes
                DrawLine(0, chartZero, chartLength, chartZero, Color.White);
                DrawLine(0, chartMiddle + chartZero, chartLength, chartMiddle + chartZero, Color.White);

                if (annotationChart >= 0)
                {
                    float chartY = charts[annotationChart][annotationTimepoint].Position.Y;
                    float chartValue = YCoordinateToChart(chartY, annotationChart);
                    HUDSprites.DrawString(HUDFont,
                        String.Format("{0} = {1}", chartLabels[annotationChart], chartValue),
                        new Vector2(annotationTimepoint, (int)chartY - 15), channelColors[annotationChart % chartChannels]);
                }
                HUDSprites.End();
                pass.End();
            }
            HUDEffect.End();
        }

        void CheckChartAnnotation()
        {
#if Windows
            MouseState mouse = Mouse.GetState();
            annotationChart = -1;
            // Yes, I know I shouldn't have to check if mouse.X > 0, but is isn't always.
            if (mouse.X > 0 && mouse.X < chartLength)
            {
                float matchError = 3;
                annotationTimepoint = mouse.X;
                for (int chart = 0; chart < totalCharts; chart++)
                {
                    if (chartLabels[chart] == null)
                        continue;

                    float error = Math.Abs(charts[chart][mouse.X].Position.Y - mouse.Y);
                    if (error < matchError)
                    {
                        annotationChart = chart;
                        matchError = error;
                    }
                }
            }
#endif
        }

    }
}
