// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace Twig
{
    /// <summary>
    /// Contains collision information about a TwigObject or a portion of a TwigObject
    /// </summary>
    public abstract class CollisionElement
    {
        /// <summary>
        /// Create a collision element for specified twig object and link (if any).
        /// </summary>
        protected CollisionElement(TypeCode type, TwigObject tObject, Link l)
        {
            typeCode = type;
            twigObject = tObject;
            link = l;
        }

        /// <summary>
        /// Create a collision element for specified twig object, but not for a link.
        /// </summary>
        protected CollisionElement(TypeCode type, TwigObject tObject) : this(type, tObject, null) { }

        /// <summary>
        /// The TwigObject for which this is a CollisionElement
        /// </summary>
        private TwigObject twigObject;

        /// <summary>
        /// TwigObject with which this collision element is associated.
        /// </summary>
        public TwigObject TwigObject
        {
            get { return twigObject; }
        }

        TwigObject connectedObject;

        /// <summary>
        /// Suppress collisions with specified object.  Intended for suppressing
        /// collisions with grasped objects.   Only one object can be suppressed at a time.
        /// Call with null, to reenable collision.
        /// </summary>
        /// <param name="o"></param>
        public void SuppressCollisionWith(TwigObject o) {
            connectedObject = o;
        }

        /// <summary>
        /// Link with which this collision element is associated, if any.
        /// </summary>
        private Link link;

        /// <summary>
        /// Link with which this collision element is associated, if any.
        /// </summary>
        public Link Link
        {
            get { return link; }
        }

        /// <summary>
        /// Indicates the shape of the bounding volume of a collision element
        /// </summary>
        protected enum TypeCode :byte {
            /// <summary>
            /// Bounding volume is a sphere
            /// </summary>
            Sphere,

            /// <summary>
            /// Bounding volume is a capsule
            /// </summary>
            Capsule,

            /// <summary>
            /// Bounding volume is an oriented bounding box
            /// </summary>
            OBB,
            /// <summary>
            /// Maximum code value
            /// </summary>
            MaxCode // THIS MUST COME AT THE END!
        };

        const byte maxType = (byte)TypeCode.MaxCode;

        /// <summary>
        /// Indicates the shape of the bounding volume of this collision element
        /// </summary>
        protected TypeCode typeCode;

        /// <summary>
        /// Returns geometry information if this collision element's bounding volume is a sphere
        /// </summary>
        protected virtual GeometryPrimitives.Sphere SphereGeometry
        {
            get
            {
                throw new MissingMethodException("Called SphereGeometry method on class " + this.GetType().Name + " which isn't a sphere");
            }
        }

        /// <summary>
        /// Returns geometry information if this collision element's bounding volume is a capsule
        /// </summary>
        protected virtual GeometryPrimitives.Capsule CapsuleGeometry
        {
            get
            {
                throw new MissingMethodException("Called CapsuleGeometry method on class " + this.GetType().Name + " which isn't a capsule");
            }
        }

        /// <summary>
        /// Returns geometry information if this collision element's bounding volume is an oriented bounding box
        /// </summary>
        protected virtual GeometryPrimitives.OBB OBBGeometry
        {
            get
            {
                throw new MissingMethodException("Called OBBGeometry method on class " + this.GetType().Name + " which isn't an oriented bounding box");
            }
        }

        /// <summary>
        /// The velocity of the center of mass of the collision element.
        /// </summary>
        public abstract Vector3 Velocity { get; }

        /// <summary>
        /// Mass of the collision element
        /// </summary>
        public abstract float Mass { get; }

        /// <summary>
        /// 1/mass of the collision element
        /// </summary>
        public abstract float InvMass { get; }

        /// <summary>
        /// How painful it is to get hit by this collision element
        /// </summary>
        public abstract float PainMultiplier { get; }

        /// <summary>
        /// Called when this collision element has collided with something.
        /// </summary>
        protected abstract void HandleCollision(CollisionElement collider, Vector3 contactPoint, Vector3 contactNormal, float penetrationDepth);

        /// <summary>
        /// Test collision between two CollisionElements and call their handlers if the collision occurred.
        /// </summary>
        public static void CheckAndHandleCollision(CollisionElement e1, CollisionElement e2)
        {
            if (e1.connectedObject == e2.TwigObject || e2.connectedObject == e1.TwigObject)
                return;

            TypeCode t1 = e1.typeCode;
            TypeCode t2 = e2.typeCode;

            // Canonicalize to have e1's type code less than e2's
            if (t2<t1) {
                TypeCode temp = t1;
                t1 = t2;
                t2 = temp;
                CollisionElement te = e1;
                e1 = e2;
                e2 = te;
            }

            Vector3 e1Contact;
            Vector3 e2Contact;
            Vector3 contactNormal = Vector3.Zero;
            float penetrationDepth = 0;

            // Dispatch on type codes
            switch ((byte)t1 * maxType + (byte)t2)
            {
                // Capsule/Capsule collision
                case (byte)TypeCode.Capsule*maxType+(byte)TypeCode.Capsule:
                    if (GeometryPrimitives.TestCollision(e1.CapsuleGeometry, e2.CapsuleGeometry,
                                                                out e1Contact, out e2Contact, out contactNormal, out penetrationDepth))
                    {
                        if (contactNormal == Vector3.Zero)
                        {
                            contactNormal = e2.Velocity - e1.Velocity;
                            MathUtil.SafeNormalize(ref contactNormal);
                        }
                        e1.HandleCollision(e2, e1Contact, contactNormal, penetrationDepth);
                        e2.HandleCollision(e1, e2Contact, -contactNormal, penetrationDepth);
                    }
                    break;

                // Sphere/Sphere collision
                case (byte)TypeCode.Sphere * maxType + (byte)TypeCode.Sphere:
                    if (GeometryPrimitives.TestCollision(e1.SphereGeometry, e2.SphereGeometry,
                                                                out e1Contact, out e2Contact, out contactNormal, out penetrationDepth))
                    {
                        if (contactNormal == Vector3.Zero)
                        {
                            contactNormal = e2.Velocity - e1.Velocity;
                            MathUtil.SafeNormalize(ref contactNormal);
                        }
                        e1.HandleCollision(e2, e1Contact, contactNormal, penetrationDepth);
                        e2.HandleCollision(e1, e2Contact, -contactNormal, penetrationDepth);
                    }
                    break;

                // Sphere/Capsule collision
                case (byte)TypeCode.Sphere * maxType + (byte)TypeCode.Capsule:
                    if (GeometryPrimitives.TestCollision(e1.SphereGeometry, e2.CapsuleGeometry,
                                                                out e1Contact, out e2Contact, out contactNormal, out penetrationDepth))
                    {
                        if (contactNormal == Vector3.Zero)
                        {
                            contactNormal = e2.Velocity - e1.Velocity;
                            MathUtil.SafeNormalize(ref contactNormal);
                        }
                        e1.HandleCollision(e2, e1Contact, contactNormal, penetrationDepth);
                        e2.HandleCollision(e1, e2Contact, -contactNormal, penetrationDepth);
                    }
                    break;

                case (byte)TypeCode.Sphere*maxType + (byte)TypeCode.OBB:
                    GeometryPrimitives.Sphere s = e1.SphereGeometry;
                    if (e2.OBBGeometry.TestPenetrationSphere(s, ref contactNormal, ref penetrationDepth))
                    {
                        e2.HandleCollision(e1, s.c + (-contactNormal * (s.r-penetrationDepth)), contactNormal, penetrationDepth);
                        e1.HandleCollision(e2, s.c+(-contactNormal*s.r), -contactNormal, penetrationDepth);
                    }
                    break;

                case (byte)TypeCode.Capsule * maxType + (byte)TypeCode.OBB:
                    GeometryPrimitives.Capsule cap = e1.CapsuleGeometry;
                    GeometryPrimitives.Sphere s1 = new GeometryPrimitives.Sphere(cap.a, cap.r);
                    GeometryPrimitives.Sphere s2 = new GeometryPrimitives.Sphere(cap.b, cap.r);
                    if (e2.OBBGeometry.TestPenetrationSphere(s1, ref contactNormal, ref penetrationDepth))
                    {
                        e2.HandleCollision(e1, s1.c + (-contactNormal * (s1.r - penetrationDepth)), contactNormal, penetrationDepth);
                        e1.HandleCollision(e2, s1.c + (-contactNormal * s1.r), -contactNormal, penetrationDepth);
                    }
                    else if (e2.OBBGeometry.TestPenetrationSphere(s2, ref contactNormal, ref penetrationDepth))
                    {
                        e2.HandleCollision(e1, s2.c + (-contactNormal * (s2.r - penetrationDepth)), contactNormal, penetrationDepth);
                        e1.HandleCollision(e2, s2.c + (-contactNormal * s2.r), -contactNormal, penetrationDepth);
                    }
                    break;

                case (byte)TypeCode.OBB * maxType + (byte)TypeCode.OBB:
                    break;

                default:
                    System.Diagnostics.Debug.Fail("Unsupported collision type: " + t1.ToString() + " with " + t2.ToString());
                    break;
            }
        }

        /// <summary>
        /// Handle any collisions between collision elements of o1 and o2.
        /// </summary>
        public static void CheckAndHandleCollisions(TwigObject o1, TwigObject o2)
        {
            Vector3 offset = o1.Position - o2.Position;
            float minimumCollisionDistance = o1.CollisionRadius+o2.CollisionRadius;
            offset.Y = 0;
            if (offset.LengthSquared() > minimumCollisionDistance * minimumCollisionDistance)
                return;

            UpdateSupport(o1, o2);
            UpdateSupport(o2, o1);

            foreach (CollisionElement e1 in o1.CollisionElements)
                foreach (CollisionElement e2 in o2.CollisionElements)
                    CheckAndHandleCollision(e1, e2);
        }

        /// <summary>
        /// Check nodes of obj for support by supportCandidate
        /// </summary>
        static void UpdateSupport(TwigObject supportCandidate, TwigObject obj)
        {
            if (supportCandidate.IsASupportingSurface)
                foreach (Node n in obj.Nodes)
                    if (supportCandidate.IsSupporting(n.Position))
                    {
                        //if (supportCandidate is SimpleTable && obj is Mammal)
                        //    System.Diagnostics.Debugger.Break();
                        n.SupportingSurface = supportCandidate;
                    }
        }
    }
}
