// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using Microsoft.Xna.Framework;

namespace Twig
{
    /// <summary>
    /// Provides functionality similar to System.Diagnostics.Debug, but specialized to the Twig environment
    /// </summary>
    public class Debug
    {
        /// <summary>
        /// Game being debugged.  This game will get paused, e.g. when an assert fails.
        /// </summary>
        public static TwigGame debugGame;

        /// <summary>
        /// Pauses game and loops recorded node motions if condition is false.
        /// </summary>
        [Conditional("DEBUG"), System.Diagnostics.DebuggerHidden()]
        public static void Assert(bool condition, string message)
        {
            if (!condition && !IsIgnoredAssertion(message))
            {
#if Windows
                System.Windows.Forms.DialogResult d 
                    = System.Windows.Forms.MessageBox.Show(message+"\nEnter debugger (yes), pause(no), or ignore for 15 seconds (cancel)?",
                    "Assertion failed", System.Windows.Forms.MessageBoxButtons.YesNoCancel);
                if (d == System.Windows.Forms.DialogResult.Yes)
                    System.Diagnostics.Debugger.Break();
                if (d != System.Windows.Forms.DialogResult.Cancel)
                {
#if NodeLogging
                    Node.PlaybackLog = true;
#endif
                    debugGame.Pause = true;
                }
                else
                {
                    IgnoreAssertion(message);
                }
#endif
            }
        }

        /// <summary>
        /// Table of times when ignored assertions expire
        /// </summary>
        static Dictionary<string, int> ignoredAssertionTable = new Dictionary<string, int>();
        static void IgnoreAssertion(string message)
        {
            ignoredAssertionTable[message] = Physics.tickCount + 15 * 60;
        }

        static bool IsIgnoredAssertion(string message)
        {
            int expirationTime;
            return ignoredAssertionTable.TryGetValue(message, out expirationTime) && Physics.tickCount < expirationTime;
        }

        /// <summary>
        /// Asserts that v's components must all be finite proper numbers.
        /// </summary>
        [Conditional("DEBUG"), System.Diagnostics.DebuggerHidden()]
        public static void AssertProper(Vector3 v, string message)
        {
            Assert(MathUtil.IsProperVector(v), message);
        }

        /// <summary>
        /// Asserts that v is a finite, proper number.
        /// </summary>
        [Conditional("DEBUG"), System.Diagnostics.DebuggerHidden()]
        public static void AssertProper(float v, string message)
        {
            Assert(MathUtil.IsProperNumber(v), message);
        }

        /// <summary>
        /// Asserts that v is a finite, proper number.
        /// </summary>
        [Conditional("DEBUG"), System.Diagnostics.DebuggerHidden()]
        public static void AssertProper(double v, string message)
        {
            Assert(MathUtil.IsProperNumber(v), message);
        }

        /// <summary>
        /// Asserts that v is a unit vector.
        /// </summary>
        [Conditional("DEBUG"), System.Diagnostics.DebuggerHidden()]
        public static void AssertUnitVector(Vector3 v, string message)
        {
            Assert(Math.Abs(v.LengthSquared() - 1) < 0.1f, message);
        }

        /// <summary>
        /// Drop into the debugger if condition is true.
        /// </summary>
        [Conditional("DEBUG"), System.Diagnostics.DebuggerHidden()]
        public static void BreakWhen(bool condition)
        {
            if (condition)
                System.Diagnostics.Debugger.Break();
        }
    }
}
