﻿// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Slot = Twig.Mammal.AttentionBuffer.Slot;
using AttentionBuffer = Twig.Mammal.AttentionBuffer;

namespace Twig
{
    /// <summary>
    /// A (unary) function over TwigObjects.  Its value is only
    /// computed for objects in the attention buffer.
    /// </summary>
    [System.Diagnostics.DebuggerDisplay("{name}")]
    public class Function<T>
    {
        string name;

        /// <summary>
        /// Name of function for debugging purposes
        /// </summary>
        public string Name
        {
            get { return name; }
        }


        T[] values = new T[AttentionBuffer.AttentionCapacity];

        /// <summary>
        /// A procedure for updating the function by directly modifying the values array.
        /// </summary>
        public delegate void UpdateProcedure(T[] valuesArray);

        /// <summary>
        /// A procedure that can update a function value given a slot number.
        /// </summary>
        public delegate T SlotUpdateProcedure(Slot s);
        /// <summary>
        /// A procedure that can update a function value given a TwigObject.
        /// </summary>
        public delegate T ObjectUpdateProcedure(TwigObject o);

        /// <summary>
        /// Character to which this function belongs
        /// </summary>
        public Mammal owner;

        /// <summary>
        /// Character to which this belongs.
        /// </summary>
        public Mammal Owner
        {
            get { return owner; }
        }

        /// <summary>
        /// The actual procedure to update this particular function.
        /// </summary>
        protected UpdateProcedure updateProcedure;

        /// <summary>
        /// Makes a function with the specified update procedure
        /// </summary>
        public Function(string name, Mammal owner, UpdateProcedure u)
        {
            this.name = name;
            this.updateProcedure = u;
            this.owner = owner;
            owner.RegisterUpdate(new Mammal.InferenceThunk(Update));
        }

        void Update()
        {
            updateProcedure(values);
        }

        /// <summary>
        /// Makes a function with the specified procedure for updating values by slot
        /// </summary>
        public Function(string name, Mammal owner, SlotUpdateProcedure slotUpdater)
            : this(name, owner, delegate(T[] values)
            {
                for (int i = 0; i < values.Length; i++)
                    values[i] = slotUpdater(new Slot((byte)i));
            })
        { }

        /// <summary>
        /// Makes a function that updates itself given the objects in the specified character's attention buffer,
        /// using the specified procedure.
        /// </summary>
        public Function(string name, Mammal m, ObjectUpdateProcedure o)
            : this(name, m, delegate(Slot s)
            {
                return o(m.Attention.SlotObject(s));
            })
        { }


        /// <summary>
        /// Returns the values of the function on all objects as an array indexed by attention slot number
        /// </summary>
        public T[] ValueArray
        {
            get { return values; }
            set { values = value; }
        }

        /// <summary>
        /// Returns the value of the function for the particular object.
        /// </summary>
        public T this[Slot s]
        {
            get
            {
                return values[s.Number];
            }
        }
    }
}
