// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

#pragma warning disable 1591

namespace Twig
{
    //
    // Gaze controller
    // Selects object for fixation
    // Head steering is in HeadController().
    //
    partial class Mammal
    {
        /// <summary>
        /// Object being fixated by creature.
        /// </summary>
        public TwigObject gazeTarget;
        /// <summary>
        /// Direction of the object being fixated by the creature.
        /// </summary>
        public Vector3 gazeDirection;
        double gazeShiftTimeout;
        /// <summary>
        /// Minimum number of seconds character will gaze at an object before switching to a new object.
        /// </summary>
        public float minimumGazeTime = 0.5f;

        private bool gazeLock = false;

        /// <summary>
        /// Prevents gaze shifts when true.
        /// </summary>
        public bool GazeLock
        {
            get { return gazeLock; }
            set { gazeLock = value; }
        }

        /// <summary>
        /// Used by the LookAt command to force the gaze to a specific object.
        /// </summary>
        TwigObject lookAtCommandTarget;

        public override bool LookingAt(TwigObject o)
        {
            return gazeTarget == o;
        }

        /// <summary>
        /// Amount we *don't* want to look at an object (e.g. because we're trying
        /// to face in the opposite direction.
        /// </summary>
        float GazePenalty(TwigObject o)
        {
            int slot = Attention.FastObjectSlotNumber(o);
            if (slot < 0)
                return 0;
            else
                return facingAwayFrom.Goal.ValueArray[slot];
        }

        /// <summary>
        /// Determine the importance of looking at this object
        /// </summary>
        /// <param name="objectInfo">object</param>
        /// <returns>Relative priority</returns>
        float GazePriority(AttentionBuffer.AttentionInfo objectInfo)
        {
            ObjectAppraisal app = Appraisal(objectInfo.obj);
            float monitoringPriority = app.monitoringPriority;
            float vigilanceCoefficient = (app.valence > 0) ? 0.01f : (-app.valence*Anxiety);
            if (objectInfo.obj == approach.target)
                // Don't let the target get trumped in vigilance by scary things
                vigilanceCoefficient = Math.Max(vigilanceCoefficient, Anxiety* 1.5f);
            return app.Salience + 0.1f * (monitoringPriority + vigilanceCoefficient) * (float)objectInfo.timeSinceLastLook
                + ((objectInfo.obj==Attention.Focus)?0.2f:0);
        }

        /// <summary>
        /// Shift gaze to new object, if necessary.
        /// </summary>
        protected void UpdateGaze()
        {
            if (lookAtCommandTarget != null &&
                locomotionController.Activation > 0 && locomotionController.maxBehavior == approach
                && locomotionController.MotorOutput != Vector3.Zero)
                // The character is trying to walk to something; cancel the LookAt command.
                gazeTarget = lookAtCommandTarget = null;

            if (lookAtCommandTarget != null)
                gazeTarget = lookAtCommandTarget;
            else
            {
                //Chart("GazeLock", GazeLock ? 1 : -1, 1.75f);
                // Don't let the gaze shift more frequently than every gazeShiftTimeout seconds
                if (Physics.time >= gazeShiftTimeout && !GazeLock)
                {
                    //
                    // Find the highest priority gaze target
                    //
                    TwigObject winner = null;
                    float winPriority = -1f;

                    // Preload the winner with any signaled goal
                    AttentionBuffer.Slot s;
                    lookingAt.Goal.FindWinner(out winner, out winPriority, out s);
                    if (winPriority == 0)
                        winner = null;

                    foreach (AttentionBuffer.AttentionInfo obj in Attention.AttendedObjectsInfo)
                    {
                        float priority = GazePriority(obj) - GazePenalty(obj.obj);
                        //Printf(string.Format("{0} {1} {2}", obj.obj.Name, priority, obj.timeSinceLastLook));
                        if (priority > winPriority && obj.obj != sittingSurface &&
                            (!IsHolding(obj.obj) || obj.obj.IsBeingManipulated || obj.obj == Attention.Focus))
                        {
                            winPriority = priority;
                            winner = obj.obj;
                        }
                    }

                    //
                    // Shift gaze to winner, if necessary
                    //
                    if (winner != gazeTarget && winPriority > 0)
                    {
                        gazeTarget = winner;
                        gazeShiftTimeout = Physics.time + minimumGazeTime;
                    }
                }
            }
            if (gazeTarget == null)
                gazeDirection = pelvisForward;
            else
            {
                gazeDirection = gazeTarget.Position - spineTop.Position;
                //if (gazeTarget is Marker)
                //    gazeDirection.Y = 0;
                MathUtil.SafeNormalize(ref gazeDirection);
                if (gazeDirection == Vector3.Zero)
                    gazeDirection = bodyForward;
            }
        }
    }
}
