﻿// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using SlotSet = Twig.Mammal.AttentionBuffer.SlotSet;
using Slot = Twig.Mammal.AttentionBuffer.Slot;

namespace Twig
{
    /// <summary>
    /// Represents for which objects in STM some predicate is a goal.
    /// A strength of 0 means it's not a goal.
    /// </summary>
    [System.Diagnostics.DebuggerDisplay("{DebugString}")]
    public class GoalAssertion : Function<float>
    {
        float[] newValues = new float[Mammal.AttentionBuffer.AttentionCapacity];

        Predicate[] preconditions;

        Predicate goal;

        /// <summary>
        /// Sets the preconditions for whatever behavior(s) achieves this goal.
        /// </summary>
        public void SetPreconditions(Predicate[] preconditions)
        {
            this.preconditions = preconditions;
        }

        /// <summary>
        /// Creates a new GoalAssertion for the specified character
        /// </summary>
        public GoalAssertion(string name, Predicate p, Mammal m)
            : base(name, m, (Function<float>.UpdateProcedure)null)
        {
            goal = p;
            // We have to fill the update procedure in here because the this argument isn't in
            // in the call to the base constructor (it compiles but fails the bytecode verifier
            // in .NET 3.5).
            updateProcedure = delegate(float[] valuesArray)
            {
                if (preconditions == null)
                {
                    Array.Copy(this.newValues, valuesArray, this.newValues.Length);
                    Array.Clear(this.newValues, 0, this.newValues.Length);
                }
                else
                    PropagatePreconditions(valuesArray);
                UpdateWaiting();
            };
        }

        /// <summary>
        /// Handles update of goal function when predicate has preconditions
        /// </summary>
        /// <param name="valuesArray"></param>
        void PropagatePreconditions(float[] valuesArray)
        {
            byte satisfiedSoFar = Owner.boundp.Extension.Mask;

            // First, spread activation to the preconditions
            foreach (Predicate p in preconditions)
            {
                byte preconditionExtension = p.Extension.Mask;
                p.Goal.Assert(newValues, (byte)(satisfiedSoFar & ~preconditionExtension));
                satisfiedSoFar &= preconditionExtension;
            }

            // Then spread activation to ourselves iff all preconditions are satisfied
            for (int i = 0; i < valuesArray.Length; i++)
                valuesArray[i] = ((satisfiedSoFar & (1 << i)) > 0) ? newValues[i] : 0;
            Array.Clear(this.newValues, 0, this.newValues.Length);
        }

        /// <summary>
        /// Asserts that the predicate is a goal for the specified SlotSet with specified strength.
        /// </summary>
        public void Assert(SlotSet s, float goalStrength)
        {
            for (byte i = 0; i < newValues.Length; i++)
                if (s[new Slot(i)])
                    newValues[i] += goalStrength;
        }

        /// <summary>
        /// Asserts the goal for the extension of the specified predicate with specified strength.
        /// </summary>
        public void Assert(Predicate p, float goalStrength)
        {
            Assert(p.Extension, goalStrength);
        }

        /// <summary>
        /// Assert the goal for the specified object with the specified strength.
        /// </summary>
        public void Assert(TwigObject o, float goalStrength)
        {
            newValues[Owner.Attention.ObjectSlotNumber(o)] += goalStrength;
        }

        /// <summary>
        /// Adds goal strengths to this goal, but only for those slots specified in mask.
        /// </summary>
        public void Assert(float[] goalStrengths, byte mask)
        {
            for (int i = 0; i < newValues.Length; i++)
                if ((mask & (1 << i))>0)
                    newValues[i] += goalStrengths[i];
        }

        /// <summary>
        /// Return the object for which this goal is most strongly asserted, along with its strength.
        /// </summary>
        public void FindWinner(out TwigObject target, out float strength, out Slot s)
        {
            float[] va = ValueArray;
            byte winner = 0;
            strength = va[winner];

            for (byte i = 1; i < va.Length; i++)
                if (va[i] > strength)
                {
                    winner = i;
                    strength = va[i];
                }

            s = new Slot(winner);
            target = Owner.Attention.SlotObject(s);
        }
        /// <summary>
        /// Renders the extension of the predicate as a string
        /// </summary>
        public string ExtensionAsString
        {
            get
            {
                StringBuilder b = new StringBuilder();
                bool gotone = false;
                for (int i = 0; i < Mammal.AttentionBuffer.AttentionCapacity; i++)
                {
                    Slot s = new Slot((byte)i);
                    if (ValueArray[i]>0)
                    {
                        TwigObject o = owner.Attention.SlotObject(s);
                        if (o != null)
                        {
                            if (gotone)
                                b.Append(", ");
                            else
                                b.Append('{');
                            b.Append(o.Name);
                            gotone = true;
                        }
                    }
                }
                if (gotone)
                {
                    b.Append('}');
                    return b.ToString();
                }
                else
                    return "{}";
            }
        }

        /// <summary>
        /// Shows value as a string.
        /// </summary>

        public string DebugString
        {
            get
            {
                return Name + "=" + ExtensionAsString;
            }
        }

        byte Mask
        {
            get
            {
                byte m = 0;
                float[] v = ValueArray;
                for (byte i = 0; i < v.Length; i++)
                    if (v[i] > 0)
                        m |= (byte)(1 << i);
                return m;
            }
        }

        void UpdateWaiting()
        {
            byte waitingFor = (byte)(Mask & ~goal.Extension.Mask);
            bool firstOne=true;
            if (waitingFor != 0)
            {
                StringBuilder b = new StringBuilder();
                b.Append(goal.Name);
                b.Append("(");
                for (byte i = 0; i < Twig.Mammal.AttentionBuffer.AttentionCapacity; i++)
                {
                    if ((waitingFor & (1 << i))!=0)
                    {
                        if (!firstOne)
                            b.Append(", ");
                        b.Append(owner.Attention.SlotObject(new Slot(i)).Name);
                        firstOne = false;
                    }
                }
                b.Append(")");
                owner.WaitingFor(b.ToString());
            }
        }
    }
}
