// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using System.Diagnostics;

#pragma warning disable 1591

namespace Twig
{
    /// <summary>
    /// A limb consisting of three notes (root, joint, and end) joined by two links (upper and lower).
    /// </summary>
    [Serializable]
    [DebuggerDisplay("{name}")]
    public class Limb : Constraint
    {
        /// <summary>
        /// Bone endpoint.
        /// </summary>
        public Node root, joint, end;
        /// <summary>
        /// Bone making up Limb.
        /// </summary>
        public Link upper, lower;
        /// <summary>
        /// Link from end to gripped object, if any.
        /// </summary>
        public Link grip;
        /// <summary>
        /// Object gripped by Limb, if any.
        /// </summary>
        TwigObject grippedObject;
        /// <summary>
        /// Position endpoint is moving to, if ballistic motion is in progress.
        /// </summary>
        Vector3 target;
        /// <summary>
        /// Time remaining for ongoing ballistic motion, if any.
        /// </summary>
        float ballisticTimeRemaining = 0f;
        bool inBallisticMotion = false;
        ControlMode finalMode;
        public TwigObject TwigObject;
        /// <summary>
        /// Length of limb when fully extended and not overstretched.
        /// </summary>
        public readonly float length;
        public readonly string name;
        Mammal me;
        /// <summary>
        /// Used for computing positions of normalized poses.
        /// -1 means this is the left side limb, +1 means the right-side
        /// </summary>
        float leftRight;

        LimbPose currentPose = null;
        LimbAnimation currentAnimation = null;

        /// <summary>
        /// Animation currently being played by limb
        /// </summary>
        public LimbAnimation CurrentAnimation
        {
            get { return currentAnimation; }
        }
        double currentAnimationStartTime;

        /// <summary>
        /// Pose to hold limb in
        /// </summary>
        public LimbPose CurrentPose
        {
            get { return currentPose; }
        }

        /// <summary>
        /// Speed with which to move into the current pose.
        /// </summary>
        float poseGain;
        
        public Limb(Mammal me, float leftRight, string name, TwigObject o, float radius, float upperLength, float lowerLength)
        {
            this.me = me;
            this.leftRight = leftRight;
            this.name = name;
            TwigObject = o;
            root = new Node(me, name+" root");
            joint = new Node(me, name+" joint");
            end = new Node(me, name+" endpoint");
            upper = new Link(name+" upper", o, root, joint, upperLength, radius);
            lower = new Link(name+" lower", o, joint, end, lowerLength, radius*0.85f);
            upper.painSensitivity = lower.painSensitivity = 0.4f;
            length = upper.Length + lower.Length;
            joint.BodyReference = root;
            end.BodyReference = root;
        }

        /// <summary>
        /// Transforms normalized node pose into this limb's local coordinate system.
        /// </summary>
        Vector3 PosePosition(NodePose p)
        {
            Vector3 X = p.X*leftRight*(p.xNormalization==LimbNormalization.Body?me.shoulderWidth:length)*me.bodyRight;
            Vector3 Y = p.Y * (p.yNormalization == LimbNormalization.Body ? me.spineLength : length) * me.bodyUp;
            Vector3 Z = p.Z * (p.zNormalization == LimbNormalization.Body ? me.shoulderWidth : length) * me.bodyForward;
            return root.Position + X + Y + Z;
        }

        /// <summary>
        /// Test whether this limb is gripping this object.
        /// </summary>
        /// <param name="o">Object to test for gripping</param>
        /// <returns>True if object is being gripped.</returns>
        public bool IsGripping(TwigObject o)
        {
            return Gripping && grippedObject == o;
        }

        /// <summary>
        /// True if limb is gripping something.
        /// </summary>
        public bool Gripping {
            get
            {
                return grip != null;
            }
        }

        /// <summary>
        /// Node being gripped, if any.
        /// </summary>
        public Node GripTarget
        {
            get
            {
                if (grip == null)
                    return null;
                else
                    return grip.node2;
            }
        }

        /// <summary>
        /// TwigObject being gripped, if any
        /// </summary>
        public TwigObject GraspedObject
        {
            get
            {
                return grippedObject;
            }
        }

        /// <summary>
        /// Attaches end node of limb to object using a zero-length Link.
        /// </summary>
        /// <param name="n">Node within object to attach to</param>
        /// <param name="obj">Object to grasp</param>
        public void Grab(Node n, TwigObject obj)
        {
            lower.CollisionElement.SuppressCollisionWith(obj);
            grip = new WeightedLink(name+" grip", TwigObject, end, n, 0, 0f);
            grippedObject = obj;
            if (obj!=null)
                obj.GraspingLimb = this;
        }

        /// <summary>
        /// Force end of limb to position of the node it's grasping.
        /// </summary>
        public void SnapToGraspedNode()
        {
            for (int i = 0; i < 2; i++)
            {
                end.Position = grip.node2.Position;
                TwigObject.Constrain();
                GraspedObject.Constrain();
            }
            root.Velocity = Vector3.Zero;
            joint.Velocity = Vector3.Zero;
            end.Velocity = Vector3.Zero;
        }

        /// <summary>
        /// Releases grasp.
        /// </summary>
        public void Ungrab()
        {
            lower.CollisionElement.SuppressCollisionWith(null);
            grip = null;
            if (grippedObject!=null)
                grippedObject.GraspingLimb = null;
            grippedObject = null;
        }

        public override void Constrain()
        {
            lower.Constrain();
            upper.Constrain();
            if (Gripping)
            {
                if (grippedObject == null)
                    grip.Constrain();
                else
                {
                    for (int i = 0; i < 3; i++)
                    {
                        grip.Constrain();
                        grippedObject.Constrain();
                    }
                }
            }
        }

        public void Draw()
        {
            upper.Draw();
            lower.Draw();
        }

        public void Update()
        {
            root.Update();
            joint.Update();
            if (InBallisticMotion)
            {
                float motionTime = Physics.deltaT;
                if (motionTime > ballisticTimeRemaining)
                {
                    motionTime = ballisticTimeRemaining;
                    inBallisticMotion = false;
                    end.ControlMode = finalMode;
                }
                float fractionOfDistance = motionTime / ballisticTimeRemaining;
                Vector3 remainingDistance = target - end.Position;
                Vector3 motion = fractionOfDistance * remainingDistance;
                end.Position += motion;
                end.Acceleration = new Vector3(0, 0, 0);
                ballisticTimeRemaining -= motionTime;
            } else if (currentAnimation != null) {
                if (currentAnimation.end != null)
                    MoveNodeToPose(end, currentAnimation.end.Position((float)(Physics.time - currentAnimationStartTime)));
                if (currentAnimation.joint != null)
                    MoveNodeToPose(joint, currentAnimation.joint.Position((float)(Physics.time - currentAnimationStartTime)));
                if (Physics.time > currentAnimationStartTime + currentAnimation.Length)
                    currentAnimation = null;
            } else if (currentPose != null)
            {
                if (currentPose.end!=null)
                    MoveNodeToPose(end, (NodePose)currentPose.end);
                if (currentPose.joint != null)
                    MoveNodeToPose(joint, (NodePose)currentPose.joint);
                else
                {
                    joint.Acceleration = Vector3.Zero;
                    joint.Velocity = root.Velocity;
                }
            }
            end.Update();
            upper.Update();
            lower.Update();
        }

        void MoveNodeToPose(Node n, NodePose p)
        {
            Vector3 current = n.Position;
            Vector3 goal = PosePosition(p);
            Vector3 error = goal - current;
            Vector3 velocity = n.Velocity - root.Velocity;

            //me.TwigGame.DrawDebugMarker(goal, Microsoft.Xna.Framework.Graphics.Color.Yellow);

            n.Acceleration = error * poseGain * length - velocity * 0.1f * poseGain * length;
            //n.Trace = TraceMode.Acceleration | TraceMode.Position | TraceMode.Velocity | TraceMode.GroundVelocity;
        }

        /// <summary>
        /// True if endpoint is engaged in a ballistic motion.
        /// </summary>
        public bool InBallisticMotion
        {
            get
            {
                return inBallisticMotion;
            }
        }

        /// <summary>
        /// Aborts and pending ballistic motion of this limb.
        /// </summary>
        public void CancelBallisticMotion() {
            ballisticTimeRemaining = 0f;
        }

        /// <summary>
        /// Initiates ballistic motion of limb endpoint.
        /// </summary>
        /// <param name="target">Location to move to</param>
        /// <param name="time">Duration of motion</param>
        /// <param name="finalMode">Control mode for endpoint upon termination</param>
        public void Goto(Vector3 target, float time, ControlMode finalMode)
        {
            end.ControlMode = ControlMode.Free;
            this.target = target;
            ballisticTimeRemaining = time;
            inBallisticMotion = true;
            currentPose = null;
            this.finalMode = finalMode;
        }

        /// <summary>
        /// Initiates ballistic motion of limb endpoint.
        /// </summary>
        /// <param name="target">Location to move to</param>
        /// <param name="time">Duration of motion</param>
        /// <param name="finalMode">Control mode for endpoint upon termination</param>
        public void Goto(NodePose target, float time, ControlMode finalMode)
        {
            Goto(PosePosition(target), time, finalMode);
        }

        /// <summary>
        /// Initiates ballistic motion of limb into the specified pose.
        /// </summary>
        /// <param name="target">Location to move to</param>
        /// <param name="time">Duration of motion</param>
        /// <param name="finalMode">Control mode for endpoint upon termination</param>
        public void Goto(LimbPose target, float time, ControlMode finalMode)
        {
            Goto((NodePose)target.end, time, finalMode);
        }

        /// <summary>
        /// Place limb in specified pose
        /// </summary>
        public void Pose(LimbPose pose, float gain)
        {
            CancelBallisticMotion();
            if (pose == null)
                Limp();
            currentPose = pose;
            poseGain = gain;
        }

        /// <summary>
        /// Sets the limb to play this animation.
        /// </summary>
        public void Play(LimbAnimation animation)
        {
            currentAnimation = animation;
            currentAnimationStartTime = Physics.time;
            this.poseGain = currentAnimation.PoseGain;
        }

        /// <summary>
        /// Tell the limb to go limp
        /// </summary>
        public void Limp()
        {
            root.ControlMode = joint.ControlMode = end.ControlMode = ControlMode.Free;
            CancelBallisticMotion();
        }

        /// <summary>
        /// Fraction of the limb's overall length that it's extended.
        /// 0: fully folded, although it would require self-collisions)
        /// 1: full extension
        /// >1: ripped apart
        /// </summary>
        public float Extension
        {
            get
            {
                Vector3 offset = root.Position - end.Position;
                return offset.Length() / (upper.Length + lower.Length);
            }
        }

        /// <summary>
        /// Enumerates links comprising limb.
        /// </summary>
        public IEnumerable<Link> Links
        {
            get
            {
                yield return upper;
                yield return lower;
            }
        }
    }
}