// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using System.Diagnostics;

#pragma warning disable 1591

namespace Twig
{
    /// <summary>
    /// A pair of limbs separated by a shoulder/pelvis link.
    /// </summary>
    [Serializable]
    [DebuggerDisplay("{name}")]
    public class LimbPair : Constraint
    {
        /// <summary>
        /// Left/right limb.
        /// </summary>
        public Limb left, right;
        /// <summary>
        /// Link joining and separating the two limbs
        /// </summary>
        public LinkWithMidpoint linkage;
        /// <summary>
        /// Distance between limbs
        /// </summary>
        public readonly float spacing;
        public readonly string name;
        
        public LimbPair(Mammal me, string name, TwigObject o, Node spine, float spacing, float radius, float upper, float lower)
        {
            this.name = name;
            left = new Limb(me, -1, name+"/left", o, radius, upper, lower);
            right = new Limb(me, 1, name+"/right", o, radius, upper, lower);
            linkage = new LinkWithMidpoint(name+" linkage", o, left.root, spine, right.root, spacing, radius);
            this.spacing = spacing;
        }

        public override void Constrain()
        {
            left.Constrain();
            right.Constrain();
            linkage.Constrain();
        }

        /// <summary>
        /// Draw both limbs and their linkage.
        /// </summary>
        public void Draw()
        {
            left.Draw();
            right.Draw();
            linkage.Draw();
        }

        /// <summary>
        /// Updates both arms.
        /// </summary>
        public void Update()
        {
            left.Update();
            right.Update();
            linkage.Update();
        }

        /// <summary>
        /// Enumerates all Links in the LimbPair
        /// </summary>
        public IEnumerable<Link> Links
        {
            get
            {
                foreach (Link l in left.Links)
                    yield return l;
                foreach (Link l in right.Links)
                    yield return l;
                yield return linkage;
            }
        }
    }
}
