// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace Twig
{
    /// <summary>
    /// Specifies how a NodePose is normalized from body to body
    /// </summary>
    public enum LimbNormalization {
        /// <summary>
        /// Coordinate is normalized to the length of the limb.
        /// </summary>
        Limb,
        /// <summary>
        /// Coordinate is normalized to the height/width/depth of the body.
        /// </summary>
        Body
    };

    /// <summary>
    /// Specifies position of a node of a limb in normalized coordinates
    /// </summary>
    public struct NodePose
    {
        /// <summary>
        /// Create a NodePose with uniform normalization along all axes
        /// </summary>
        public NodePose(LimbNormalization normalization, float x, float y, float z)
        {
            xNormalization = yNormalization = zNormalization = normalization;
            X = x;
            Y = y;
            Z = z;
        }

        /// <summary>
        /// Create a NodePose with the specified coordinates and normalizations
        /// </summary>
        public NodePose(LimbNormalization xnormalization, float x, LimbNormalization ynormalization, float y, LimbNormalization znormalization, float z)
        {
            xNormalization = xnormalization;
            X = x;
            yNormalization = ynormalization;
            Y = y;
            zNormalization = znormalization;
            Z = z;
        }

        /// <summary>
        /// Position along the left/right axis of the body
        /// </summary>
        public float X;
        /// <summary>
        /// Normalization of the X coordinate (limb/body).
        /// Positive values always mean outward from the body, negative values mean in toward the torso.
        /// </summary>
        public LimbNormalization xNormalization;
        /// <summary>
        /// Position along the principal axis of the torso
        /// </summary>
        public float Y;
        /// <summary>
        /// Normalization of the Y coordinate (limb/body)
        /// </summary>
        public LimbNormalization yNormalization;
        /// <summary>
        /// Position along the dorsal/ventral axis of the body
        /// </summary>
        public float Z;
        /// <summary>
        /// Normalization of the Z coordinate (limb/body)
        /// </summary>
        public LimbNormalization zNormalization;
    }

    /// <summary>
    /// Specifies a pose of a limb.
    /// Can specify positions for the end and/or joint nodes
    /// </summary>
    public class LimbPose
    {
        /// <summary>
        /// A limb pose with the joint (elbow/knee) position left free
        /// </summary>
        public LimbPose(NodePose end)
        {
            this.end = end;
            this.joint = null;
        }

        /// <summary>
        /// A limb pose with the joint (elbow/knee) position specified
        /// </summary>
        public LimbPose(NodePose joint, NodePose end)
        {
            this.end = end;
            this.joint = joint;
        }

        /// <summary>
        /// A limb pose with the endpoint in the specified normalized position
        /// </summary>
        public LimbPose(LimbNormalization normalization, float x, float y, float z) : this(new NodePose(normalization, x, y, z)) { }

        /// <summary>
        /// Pose for the limb endpoint
        /// </summary>
        public Nullable<NodePose> end;
        /// <summary>
        /// Pose for the limb joint (null is joint is free)
        /// </summary>
        public Nullable<NodePose> joint;
    }

    /// <summary>
    /// Represents a single keyframe in the animation of a node
    /// </summary>
    public struct NodeKey
    {
        /// <summary>
        /// Creates a keyframe of a node animation.
        /// </summary>
        public NodeKey(float time, float x, float y, float z)
        {
            this.time = time;
            this.X = x;
            this.Y = y;
            this.Z = z;
        }

        private float time;

        /// <summary>
        /// Time of this keyframe relative to the start of the animation.
        /// </summary>
        public float Time
        {
            get { return time; }
            set { time = value; }
        }
        /// <summary>
        /// Normalized X coordinate of this keyframe.
        /// </summary>
        [ContentSerializerIgnore]
        //[ContentSerializer(FlattenContent = true)]
        public float X;
        /// <summary>
        /// Normalized Y coordinate of this keyframe.
        /// </summary>
        [ContentSerializerIgnore]
        //[ContentSerializer(FlattenContent = true)]
        public float Y;
        /// <summary>
        /// Normalized Z coordinate of this keyframe.
        /// </summary>
        [ContentSerializerIgnore]
        //[ContentSerializer(FlattenContent = true)]
        public float Z;

        /// <summary>
        /// Sets normalized position of key frame
        /// </summary>
        //[ContentSerializerIgnore]
        public Vector3 Position
        {
            get
            {
                return new Vector3(X, Y, Z);
            }
            set
            {
                X = value.X;
                Y = value.Y;
                Z = value.Z;
            }
        }
    }

    /// <summary>
    /// Represents a fixed animation for node
    /// </summary>
    public class NodeAnimation
    {
        /// <summary>
        /// Used only for content pipeline
        /// </summary>
        private NodeAnimation()
        {
            X = new Curve();
            Y = new Curve();
            Z = new Curve();
            xNormalization = yNormalization = zNormalization = LimbNormalization.Limb;
        }

        /// <summary>
        /// Returns the position of the NodePose of the node at the specified time.
        /// </summary>
        public NodePose Position(float time)
        {
            return new NodePose(xNormalization, X.Evaluate(time), yNormalization, Y.Evaluate(time), zNormalization, Z.Evaluate(time));
        }

        /// <summary>
        /// Create a NodeMotion with uniform normalization along all axes
        /// </summary>
        public NodeAnimation(LimbNormalization normalization, params NodeKey[] keyFrames)
        {
            xNormalization = yNormalization = zNormalization = normalization;
            X = new Curve();
            Y = new Curve();
            Z = new Curve();
            Keys = keyFrames;
        }

        /// <summary>
        /// Create a NodeMotion with the specified coordinates and normalizations
        /// </summary>
        public NodeAnimation(LimbNormalization xnormalization, LimbNormalization ynormalization, LimbNormalization znormalization,
            params NodeKey[] keyFrames)
        {
            xNormalization = xnormalization;
            yNormalization = ynormalization;
            zNormalization = znormalization;
            X = new Curve();
            Y = new Curve();
            Z = new Curve();
            Keys = keyFrames;
        }

        /// <summary>
        /// Position along the left/right axis of the body
        /// </summary>
        [ContentSerializerIgnore]
        public Curve X;
        /// <summary>
        /// Normalization of the X coordinate (limb/body).
        /// Positive values always mean outward from the body, negative values mean in toward the torso.
        /// </summary>
        [ContentSerializer(Optional=true)]
        public LimbNormalization xNormalization;
        /// <summary>
        /// Position along the principal axis of the torso
        /// </summary>
        [ContentSerializerIgnore]
        public Curve Y;
        /// <summary>
        /// Normalization of the Y coordinate (limb/body)
        /// </summary>
        [ContentSerializer(Optional = true)]
        public LimbNormalization yNormalization;
        /// <summary>
        /// Position along the dorsal/ventral axis of the body
        /// </summary>
        [ContentSerializerIgnore]
        public Curve Z;
        /// <summary>
        /// Normalization of the Z coordinate (limb/body)
        /// </summary>
        [ContentSerializer(Optional = true)]
        public LimbNormalization zNormalization;

        /// <summary>
        /// Uniformly sets normalization along all axes.
        /// </summary>
        [ContentSerializer(Optional = true)]
        public LimbNormalization Normalization
        {
            get
            {
                throw new NotImplementedException("Normalization property is only supported for XML deserialization");
            }
            set
            {
                xNormalization = yNormalization = zNormalization = value;
            }
        }

        float length;

        /// <summary>
        /// Length of this animation in seconds.
        /// </summary>
        public float Length
        {
            get
            {
                return length;
            }
        }

        /// <summary>
        /// Keys the key frames of the animation
        /// </summary>
        [ContentSerializer(CollectionItemName="Key")]
        public NodeKey[] Keys
        {
            get
            {
                throw new NotImplementedException("Getting on Keys is not supported");
            }
            set
            {
                float maxTime = 0;
                foreach (NodeKey k in value)
                {
                    X.Keys.Add(new CurveKey(k.Time, k.X));
                    Y.Keys.Add(new CurveKey(k.Time, k.Y));
                    Z.Keys.Add(new CurveKey(k.Time, k.Z));
                    if (k.Time > maxTime)
                        maxTime = k.Time;
                }
                X.ComputeTangents(CurveTangent.Smooth);
                Y.ComputeTangents(CurveTangent.Smooth);
                Z.ComputeTangents(CurveTangent.Smooth);
                length = maxTime;
            }
        }

        /// <summary>
        /// Deserializer for content pipeline
        /// </summary>
        public class Reader : ContentTypeReader<NodeAnimation>
        {
            /// <summary>
            /// Read it from the file
            /// </summary>
            protected override NodeAnimation Read(
                    ContentReader input,
                    NodeAnimation existingInstance)
            {
                if (existingInstance == null)
                    existingInstance = new NodeAnimation();
                existingInstance.xNormalization = (LimbNormalization)input.ReadByte();
                existingInstance.yNormalization = (LimbNormalization)input.ReadByte();
                existingInstance.zNormalization = (LimbNormalization)input.ReadByte();
                existingInstance.length = input.ReadSingle();
                existingInstance.X = input.ReadObject<Curve>();
                existingInstance.Y = input.ReadObject<Curve>();
                existingInstance.Z = input.ReadObject<Curve>();
                return existingInstance;
            }
        }
    }

    /// <summary>
    /// A whole-limb keyframe animation
    /// </summary>
    public class LimbAnimation
    {
        /// <summary>
        /// Used only by content pipeline
        /// </summary>
        private LimbAnimation()
        {
            joint = null;
            loop = false;
        }

        /// <summary>
        /// Creates a new limb animation
        /// </summary>
        public LimbAnimation(NodeAnimation end, NodeAnimation joint, bool loop)
        {
            this.end = end;
            this.joint = joint;
            this.loop = loop;
        }
        /// <summary>
        /// Path for the limb endpoint
        /// </summary>
        public NodeAnimation end;

        /// <summary>
        /// Path for the limb joint (elbow or knee)
        /// </summary>
        [ContentSerializer(Optional=true)]
        public NodeAnimation joint;

        /// <summary>
        /// Total number of seconds in the animation
        /// </summary>
        public float Length
        {
            get
            {
                return end.Length;
            }
        }

        /// <summary>
        /// Whether the animation is intended to be looped.
        /// </summary>
        [ContentSerializer(Optional=true)]
        public bool loop;

        float poseGain = 1000f;

        /// <summary>
        /// Tracking gain for following the animation path.
        /// </summary>
        [ContentSerializer(Optional = true)]
        public float PoseGain
        {
            get { return poseGain; }
            set { poseGain = value; }
        }


        /// <summary>
        /// Deserializer for content pipeline
        /// </summary>
        public class Reader : ContentTypeReader<LimbAnimation>
        {
            /// <summary>
            /// Read it from the file
            /// </summary>
            protected override LimbAnimation Read(
                    ContentReader input,
                    LimbAnimation existingInstance)
            {
                if (existingInstance == null)
                    existingInstance = new LimbAnimation();

                existingInstance.end = input.ReadObject<NodeAnimation>();
                existingInstance.joint = input.ReadObject<NodeAnimation>();
                existingInstance.loop = input.ReadBoolean();
                existingInstance.PoseGain = input.ReadSingle();
                return existingInstance;
            }
        }
    }
}
