// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

#pragma warning disable 1591

namespace Twig
{
    /// <summary>
    /// A link (bone) connecting two Nodes at a fixed distance.
    /// </summary>
    [Serializable]
    [DebuggerDisplay("{name}")]
    public class Link : Constraint
    {
        public Link(string name, TwigObject o, Node n1, Node n2, float length, float radius)
        {
            this.name = name;
            this.TwigObject = o;
            node1 = n1;
            node2 = n2;
            this.radius = radius;
            this.length = length;
            this.LengthSquared = length * length;
            //this.cylinder = RenderWindow.Cylinder(radius, radius, length, 10, 1);
            if (radius!=0)
                this.cylinder = CylinderMesh.Get(o.Device, radius, length, 20);
            collisionElement = new LinkCollisionElement(this);
        }
        string name;

#if DEBUG
        [DebugSwitch("Break when pain is registered")]
        public static bool breakOnPain = false;
#endif

        LinkCollisionElement collisionElement;

        /// <summary>
        /// Collision Element corresponding to this Link
        /// </summary>
        public CollisionElement CollisionElement
        {
            get { return collisionElement; }
        }

        /// <summary>
        /// Ideal distance between nodes.
        /// </summary>
        protected float length, LengthSquared;
        /// <summary>
        /// Radius of bone comprising link.
        /// </summary>
        public float radius;

        /// <summary>
        /// Ideal distance between nodes.
        /// </summary>
        public float Length
        {
            get
            {
                return length;
            }
            set
            {
                length = value;
                LengthSquared = value * value;
            }
        }

        /// <summary>
        /// Returns a unit vector pointing along the axis of the link (from node1 to node2)
        /// CAN BE ZERO (if link is zero-length)
        /// </summary>
        public Vector3 Orientation
        {
            get
            {
                Vector3 o = node1.Position - node2.Position;
                MathUtil.SafeNormalize(ref o);
                return o;
            }
        }

        /// <summary>
        /// Nodes connected by link.
        /// </summary>
        public Node node1, node2;
        /// <summary>
        /// The object to which this link belongs.
        /// </summary>
        public TwigObject TwigObject;
        const float parallelThreshold = 0.0001f;
        [NonSerialized]
        CylinderMesh cylinder;

        /// <summary>
        /// Amount of pain registered on this limb
        /// </summary>
        public float pain;
        /// <summary>
        /// Amount of pain inflicted by this limb.
        /// This is a cheat to keep me from having to make the physics
        /// calculations more accurate.  This gets increased when
        /// punching or grappling.
        /// </summary>
        public float painMultiplier = 1f;

        /// <summary>
        /// Threshold below which pain doesn't register.
        /// </summary>
        public static float painThreshold = 100000f;
        /// <summary>
        /// Multiplier on pain sensing.
        /// </summary>
        public float painSensitivity = 1f;

        /// <summary>
        /// Link being collided with.
        /// </summary>
        public Link contactLink;
        /// <summary>
        /// Object being collided with.
        /// </summary>
        public TwigObject contactObject;
        Link newContactLink;
        TwigObject newConactObject;

        /// <summary>
        /// Mass of the link/bone.
        /// </summary>
        public float Mass
        {
            get
            {
                return node1.Mass + node2.Mass;
            }
        }

        public float InvMass
        {
            get
            {
                return 1 / Mass;
            }
        }

        /// <summary>
        /// Momentum of the link/bone.
        /// </summary>
        public Vector3 Momentum
        {
            get
            {
                return node1.Momentum + node2.Momentum;
            }
        }

        public Vector3 Velocity
        {
            get
            {
                return (node1.Velocity + node2.Velocity)/2f;
            }
        }

        /// <summary>
        /// Position of midpoint in world coordinates
        /// </summary>
        public Vector3 MidPoint
        {
            get
            {
                return (node1.Position + node2.Position) * 0.5f;
            }
        }

        /// <summary>
        /// KE of link/bone
        /// </summary>
        public float KineticEnergy
        {
            get
            {
                return node1.KineticEnergy + node2.KineticEnergy;
            }
        }

        /// <summary>
        /// Update contact and pain information.
        /// </summary>
        public void Update()
        {
            contactLink = newContactLink;
            contactObject = newConactObject;
            newContactLink = null;
            newConactObject = null;
            pain *= (1 - Physics.deltaT);
        }

        public override void Constrain()
        {
            Vector3 delta = node1.Position - node2.Position;
            delta *= LengthSquared / (delta.LengthSquared() + LengthSquared) - 0.5f;
            node1.Position += delta;
            node2.Position -= delta;
        }

        /// <summary>
        /// Signal the fact that the specified link of the specified TwigObject has contacted this object.
        /// </summary>
        public void DeclareContact(TwigObject o, Link l)
        {
            newConactObject = o;
            newContactLink = l;
        }

        /// <summary>
        /// Renders objects in wireframe mode.
        /// </summary>
        public static bool renderLines = false;
        static VertexPositionColor[] lineBuffer = new VertexPositionColor[2];
#if MDX
        static CustomVertex.PositionColored[] lineBuffer = new CustomVertex.PositionColored[2];

        public static void DrawLine(Device d, Vector3 from, Vector3 to, ColorValue color)
        {
            d.Transform.World = Matrix.Identity;
            d.VertexFormat = CustomVertex.PositionColored.Format;
            lineBuffer[0].Position = from;
            lineBuffer[0].Color = color.ToArgb();
            lineBuffer[1].Position = to;
            lineBuffer[1].Color = color.ToArgb();
            d.DrawUserPrimitives(PrimitiveType.LineList, 1, lineBuffer);
        }
#endif
        /// <summary>
        /// Renders the Link.
        /// </summary>
        public void Draw()
        {
            if (renderLines)
            {
                lineBuffer[0].Position = node1.PositionForDisplay;
                lineBuffer[0].Color = Color.Red;
                lineBuffer[1].Position = node2.PositionForDisplay;
                lineBuffer[1].Color = Color.Red;
                this.TwigObject.Device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, lineBuffer, 0, 1);
            }
            else
                DrawCylinder();
        }

        //
        // Render link as a cylinder (or line if renderLines is true)
        //
        void DrawCylinder()
        {
            //
            // Determine the matrix needed to transform a cylinder centered at the origin
            // and aligned with the Y axis into the pose required for this link.
            //

            // Offset between the nodes
            Vector3 offset = node2.PositionForDisplay - node1.PositionForDisplay;

            // Find orthonormal vectors perpendicular to offset (b1, b2)
            Vector3 b1 = Vector3.Cross(offset, Vector3.UnitY);

            if (b1.LengthSquared() < 0.001)
                b1 = Vector3.Cross(offset, Vector3.UnitX);
            b1.Normalize();
            Vector3 b2 = Vector3.Cross(b1, offset);
            b2.Normalize();

            Matrix transform = Matrix.CreateTranslation(node1.PositionForDisplay + offset * 0.5f);
            MathUtil.SafeNormalize(ref offset);
            transform.M11 = b1.X;
            transform.M12 = b1.Y;
            transform.M13 = b1.Z;
            transform.M31 = b2.X;
            transform.M32 = b2.Y;
            transform.M33 = b2.Z;
            transform.M21 = offset.X;
            transform.M22 = offset.Y;
            transform.M23 = offset.Z;
            TwigObject.BasicEffect.World = transform;
            TwigObject.BasicEffect.CommitChanges();

            cylinder.Draw();
        }

#if MDX
        //static Mesh cylinder;
        public static void PrepareDevice(Device d)
        {
            //cylinder = Mesh.Cylinder(d, 0.1f, 0.1f, 1.0f, 10, 1);
            //cylinder = Mesh.Sphere(d, 0.5f, 5, 5);
        }
#endif

#if OldCollision
        /// <summary>
        /// Test for collision between two links.
        /// If contact occurs, applies collision corrections to them and computes pain.
        /// </summary>
        /// <returns>True if the links are in contact.</returns>
        public static bool TestCollision(Link L1, Link L2)
        {
            // We can do a fast and dirty test for distance using a cross product
            // dot product and subtraction.  We might be better off doing that
            // even though we couldn't reuse any of that work ...


            // Line intersection code adapted from http://softsurfer.com/Archive/algorithm_0106/algorithm_0106.htm
            // Copyright notice from softsurfer:
            // Copyright 2001, softSurfer (www.softsurfer.com)
            // This code may be freely used and modified for any purpose
            // providing that this copyright notice is included with it.
            // SoftSurfer makes no warranty for this code, and cannot be held
            // liable for any real or imagined damage resulting from its use.
            // Users of this code must verify correctness for their application.

            Vector3 u = L1.node2.Position - L1.node1.Position;
            Vector3 v = L2.node2.Position - L2.node1.Position;
            Vector3 w = L1.node1.Position - L2.node1.Position;
            float a = Vector3.Dot(u, u);        // always >= 0
            float b = Vector3.Dot(u, v);
            float c = Vector3.Dot(v, v);        // always >= 0
            float d = Vector3.Dot(u, w);
            float e = Vector3.Dot(v, w);
            float D = a * c - b * b;       // always >= 0
            // sc is the fractional distance along L1 of the CPA
            // tc is the fractional distance along L2 of the CPA
            // In non-parallel case, the links intersect if 0<= sc, tc <= 1
            // and distance is less than the sum of the radii of the links
            float sc, tc;
            const float smallNum = 0.00000001f; 

            // compute the line parameters of the two closest points
            if (D < smallNum)
            {         // the lines are almost parallel
                sc = 0.0f;
                tc = (b > c ? d / b : e / c);   // use the largest denominator
            }
            else
            {
                sc = (b * e - c * d) / D;
                tc = (a * e - b * d) / D;
            }

            if (sc >= 0f && sc <= 1f && tc >= 0f && tc <= 1f)
            {
                // get the difference of the two closest points
                Vector3 dP = w + (sc * u) - (tc * v);  // = L1(sc) - L2(tc)
                float minDist = L1.radius+L2.radius;
                // BUG: If dP=0, then the spines of the cylinders are exactly coincident
                // and so we can't figure out what direction to move them in.
                // So we'll punt and move them randomly.
                while (dP == Vector3.Zero)
                    dP = MathUtil.NoiseVector(0.1f);
                if (dP.LengthSquared() < (minDist * minDist))
                {
                    // Intersection
                    float distance = dP.Length();
                    float dError = minDist-distance;
                    Vector3 correction = dP * (0.5f * dError / (dP.Length()*(L1.node1.InvMass+L2.node1.InvMass)));
                    // We really ought to weight the corrections based on sc and tc
                    // but this will let me at least debug the intersection code.
                    L1.ApplyCollisionCorrection(correction);
                    L2.ApplyCollisionCorrection(-correction);
                    if (L1.TwigObject != L2.TwigObject)
                    {
                        correction.Normalize();
                        float impactVelocity = Math.Abs(Vector3.Dot(L1.Velocity - L2.Velocity, correction));
                        L1.ApplyPain(impactVelocity * L2.Mass / L1.Mass, L2);
                        L2.ApplyPain(impactVelocity * L1.Mass / L2.Mass, L1);
                    }
                    return true;
                }
            }
            return false;
        }

        // Move by link by correction, weighted by inverse weight
        // increment pain if appropriate
        void ApplyCollisionCorrection(Vector3 correction)
        {
            Vector3 weighted = node1.InvMass * correction;
            node1.Position += weighted;
            node2.Position += weighted;
            float weight = weighted.LengthSquared();
        }

        void ApplyPain(float momentum, Link collider)
        {
            if (!Physics.IsSettled)
                return;
            float deltaPain = painSensitivity * collider.Mass * momentum / Mass * collider.painMultiplier;
            Debug.AssertProper(deltaPain, "Pain level due to collision is an invalid number.");
            if (deltaPain > painThreshold)
                pain += deltaPain;
        }

        /// <summary>
        /// Checks for collision between this link and the specified link of the specified TwigObject.
        /// If collision is detected, the link positions are corrected to depenetrate them, contact is
        /// declared, and pain is computed.
        /// </summary>
        /// <param name="l"></param>
        /// <param name="o"></param>
        /// <returns></returns>
        public bool CheckCollision(Link l, TwigObject o)
        {
            bool contact = TestCollision(this, l);
            if (contact)
                DeclareContact(o, l);
            return contact;
        }
#else
        public static void TestCollision(Link l1, Link l2)
        {
            CollisionElement.CheckAndHandleCollision(l1.CollisionElement, l2.CollisionElement);
        }

        void ApplyPain(float KE, float painMultiplier)
        {
            if (!Physics.IsSettled)
                return;
            float deltaPain = painSensitivity * KE * painMultiplier;
            Debug.AssertProper(deltaPain, "Pain level due to collision is an invalid number.");
            if (deltaPain > painThreshold)
            {
#if DEBUG
                if (breakOnPain)
                    System.Diagnostics.Debugger.Break();
#endif
                pain += deltaPain;
            }
        }
#endif

        [DebuggerDisplay("{link.name}")]
        class LinkCollisionElement : CollisionElement
        {
            public LinkCollisionElement(Link l)
                : base(CollisionElement.TypeCode.Capsule, l.TwigObject, l)
            { }

            protected override GeometryPrimitives.Capsule CapsuleGeometry
            {
                get
                {
                    Vector3 axis = Link.node2.Position - Link.node1.Position;
                    MathUtil.SafeNormalize(ref axis);
                    Vector3 offset = axis * Link.radius;
                    return new GeometryPrimitives.Capsule(Link.node1.Position + offset, Link.node2.Position - offset, Link.radius);
                }
            }

            protected override void HandleCollision(CollisionElement collider, Vector3 contactPoint, Vector3 contactNormal, float penetrationDepth)
            {
                bool selfCollision = false;

                if (collider.TwigObject != null && collider.TwigObject != TwigObject)
                {
                    if (collider.Link != null)
                        Link.contactLink = collider.Link;
                    Link.contactObject = collider.TwigObject;
                }
                else
                    selfCollision = true;
                float correction = (InvMass/(InvMass+collider.InvMass))*penetrationDepth;
                Link.node1.Position -= correction * contactNormal;
                Link.node1.Position -= correction * contactNormal;
                float normalSpeed = Vector3.Dot((Velocity - collider.Velocity), contactNormal);
                if (!selfCollision)
                    Link.ApplyPain(0.5f * (collider.Mass / Mass) * normalSpeed * normalSpeed, collider.PainMultiplier);
            }

            public override Vector3 Velocity
            {
                get { return Link.Velocity; }
            }

            public override float Mass
            {
                get { return Link.Mass; }
            }

            public override float InvMass
            {
                get { return Link.InvMass; }
            }

            public override float PainMultiplier
            {
                get { return Link.painMultiplier; }
            }
        }
    }

    /// <summary>
    /// Link that honors the relative masses of the two nodes when resolving distance constraint.
    /// </summary>
    [DebuggerDisplay("{name}")]
    public class WeightedLink : Link
    {
        public WeightedLink(string name, TwigObject o, Node n1, Node n2, float length, float radius)
            : base(name, o, n1, n2, length, radius) { }

        public override void Constrain()
        {
            float invSum = (node1.InvMass + node2.InvMass);
            Vector3 delta = node1.Position - node2.Position;
            delta *= (LengthSquared / (delta.LengthSquared() + LengthSquared) - 0.5f)/invSum;
            node1.Position += node1.InvMass*delta;
            node2.Position -= node2.InvMass*delta;
        }
    }

    /// <summary>
    /// Imposes the constraint that two nodes must be at least a certain distance apart.
    /// </summary>
    [DebuggerDisplay("{name}")]
    public class DistanceLowerBound : Link
    {
        public DistanceLowerBound(string name, TwigObject o, Node n1, Node n2, float length, float radius)
            : base(name, o, n1, n2, length, radius)
        { }

        public override void Constrain()
        {
            Vector3 delta = node1.Position - node2.Position;
            float ls = delta.LengthSquared();
            if (ls < LengthSquared)
            {
                delta *= LengthSquared / (ls + LengthSquared) - 0.5f;
                node1.Position += delta;
                node2.Position -= delta;
            }
        }
    }

    /// <summary>
    /// Imposes the constraint that two nodes must be at least a certain distance apart.
    /// </summary>
    [DebuggerDisplay("{name}")]
    public class DistanceUpperBound : Link
    {
        public DistanceUpperBound(string name, TwigObject o, Node n1, Node n2, float length, float radius)
            : base(name, o, n1, n2, length, radius)
        { }

        public override void Constrain()
        {
            Vector3 delta = node1.Position - node2.Position;
            float ls = delta.LengthSquared();
            if (ls > LengthSquared)
            {
                delta *= LengthSquared / (ls + LengthSquared) - 0.5f;
                node1.Position += delta;
                node2.Position -= delta;
            }
        }
    }


    /// <summary>
    /// Link a link, but maintains a midpoint Node between the two endpoints.
    /// Midpoint is held at the midpoint of the two Nodes plus offset.
    /// </summary>
    [Serializable]
    [DebuggerDisplay("{name}")]
    public class LinkWithMidpoint : Link
    {
        /// <summary>
        /// Node to constrain between the two endpoints.
        /// </summary>
        public Node midpoint;
        /// <summary>
        /// Displacement from the midpoint of the endpoints at which to hold the midpoint node.
        /// </summary>
        public Vector3 offset;
        public LinkWithMidpoint(string name, TwigObject o, Node n1, Node midpoint, Node n2, float length, float radius)
            : base(name, o, n1, n2, length, radius)
        {
            this.midpoint = midpoint;
            this.offset = Vector3.Zero;
        }

        public override void Constrain()
        {
            Vector3 realMid = (node2.Position + node1.Position)*0.5f;
            Vector3 error = realMid - (midpoint.Position+offset);
            Vector3 qoff = 0.25f * error;
            midpoint.Position += 0.5f * error;
            node1.Position -= qoff+offset;
            node2.Position -= qoff+offset;

            base.Constrain();
        }
    }
}
