// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A playground merry-go-round the children can spin around on.
    /// </summary>
    public class MerryGoRound : TwigObject
    {
        /// <summary>
        /// Create a merry-go-round in the specified position
        /// </summary>
        public MerryGoRound(TwigGame g, string name, Vector3 position)
            : base(g, name, true)
        {
            Color = Color.SteelBlue;

            bottom = new Node(this, name+".bottom", position);
            platform = new Node(this, name+".platform", position + new Vector3(0, platformAltitude, 0));
            top = new Node(this, name+".top", position + new Vector3(0, height, 0));
            bottom.ControlMode = platform.ControlMode = top.ControlMode = ControlMode.Fixed;
            platformMesh = CylinderMesh.Get(g.Device, radius, platformThickness, 20);

            List<Link> visibleLinks = new List<Link>();
            List<Link> invisibleLinks = new List<Link>();
            List<Link> verticalLinks = new List<Link>();
            List<Node> periphery = new List<Node>();
            List<Node> pipebases = new List<Node>();
            Node previousTop=null, previousBottom=null, originalTop=null, originalBottom=null;
            visibleLinks.Add(new Link(name+"spine", this, bottom, top, height, pipeRadius));
            for (int i=0; i<sectionCount; i++) {
                Vector3 offset = position+radius*new Vector3((float)Math.Sin(sectionAngle*i), 0, (float)Math.Cos(sectionAngle*i));
                Node bnode = new Node(this, name+" periphery bottom", offset+Vector3.UnitY*platformAltitude);
                Node tnode = new Node(this, name+" periphery top", offset+Vector3.UnitY*(height));
                bnode.BodyReference = tnode;
                tnode.BodyReference = bnode;
                bnode.BodyCenteredDamping = tnode.BodyCenteredDamping = 1;
                pipebases.Add(bnode);
                periphery.Add(bnode);
                periphery.Add(tnode);

                Link vertical = new Link(name + " vertical", this, bnode, tnode, platformHeight, pipeRadius);
                verticalLinks.Add(vertical); 
                visibleLinks.Add(vertical);
                invisibleLinks.Add(new Link(name+" crosspiece", this, bnode, platform, radius, 0));
                visibleLinks.Add(new Link(name+" crosspiece", this, tnode, top, radius, pipeRadius));
                if (previousTop == null)
                {
                    originalBottom = previousBottom = bnode;
                    originalTop = previousTop = tnode;
                }
                else
                {
                    invisibleLinks.Add(new Link(name + " upper stabilizer", this, tnode, previousTop,
                        (tnode.Position - previousTop.Position).Length(), 0));
                    invisibleLinks.Add(new Link(name + " lower stabilizer", this, bnode, previousBottom,
                        (tnode.Position - previousTop.Position).Length(), 0));

                    previousBottom = bnode;
                    previousTop = tnode;
                }
            }

            invisibleLinks.Add(new Link(name + " upper stabilizer", this, originalTop, previousTop,
                (originalTop.Position - previousTop.Position).Length(), 0));
            invisibleLinks.Add(new Link(name + " lower stabilizer", this, originalBottom, previousBottom,
                (originalBottom.Position - previousBottom.Position).Length(), 0));

            this.periphery = periphery.ToArray();
            this.verticalLinks = verticalLinks.ToArray();
            this.visibleLinks = visibleLinks.ToArray();
            this.invisibleLinks = invisibleLinks.ToArray();

            List<CollisionElement> ces = new List<CollisionElement>();
            foreach (Link l in visibleLinks)
            {
                l.painMultiplier = 0.25f;  // Otherwise, the weight is so high it registers pain even for slow-moving collisions
                ces.Add(l.CollisionElement);
            }
            collisionElements = ces.ToArray();
            Dynamic = false;
            ComputeMasses();

            nodeFrame = new NodeFrame(pipebases[0], pipebases[sectionCount / 3], pipebases[2 * sectionCount / 3]);
        }

        /// <summary>
        /// Updates the masses of the nodes by computing the masses of the links under the
        /// assumption they have the density of water, then distributing that mass among the nodes.
        /// </summary>
        void ComputeMasses()
        {
            float totalMass = 0;

            //  Nodes default to unit mass.
            foreach (Node n in Nodes)
            {
                n.Mass = 0;
            }

            foreach (Link l in visibleLinks)
            {
                // Volume of link in cubic meters
                float linkVolume = l.Length * 2 * (float)Math.PI * l.radius * l.radius;
                // Mass of link in kilograms, assuming it's made of a substance twice the density of iron
                // (this makes up for the fact that the platform isn't modeled here as having any mass).
                // The multiplications convert from m^3 to cm^3;  Iron density is roughly 7.5g/cm^3, so
                // we multiply by 15 to fudge a reasonable total mass.  We then divide by 1000 to get KG.
                float linkMass = 15*(linkVolume * 100 * 100 * 100) / 1000.0f;

                totalMass += linkMass;

                // Divide the link's mass evently between its two nodes
                l.node1.Mass += linkMass / 2f;
                l.node2.Mass += linkMass / 2f;
            }

            m_mass = totalMass;
        }


        const float platformHeight = 1f;
        const float platformAltitude = 0.2f;
        const float platformThickness = 0.1f;
        const float height = platformAltitude + platformHeight;
        const float radius = 1.6f;
        const float pipeRadius = 0.05f;
        
        const int sectionCount = 7;
        const float sectionAngle = (float)(2*Math.PI/sectionCount);

        Node bottom, platform, top;
        Node[] periphery;
        Link[] visibleLinks, invisibleLinks, verticalLinks;
        NodeFrame nodeFrame;

        float m_mass;

        CylinderMesh platformMesh;

        /// <summary>
        /// The ground-plane radius of the collision volume of the merry-go-round.
        /// </summary>
        public override float CollisionRadius
        {
            get { return radius; }
        }

        /// <summary>
        /// Draw the merry-go-round
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Draw(GameTime gameTime)
        {
            BasicEffect.Begin();
            ConfigureMaterial();
            foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                foreach (Link l in visibleLinks)
                    l.Draw();
                BasicEffect.World = Matrix.CreateTranslation(Position + Vector3.UnitY * (platformAltitude - platformThickness / 2));
                BasicEffect.CommitChanges();
                platformMesh.Draw();
                pass.End();
            }
            BasicEffect.End();
        }

        /// <summary>
        /// Update node positions.
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Update(GameTime gameTime)
        {
            top.Update();
            bottom.Update();
            platform.Update();
            //periphery[0].Velocity = 3f*Vector3.Cross(periphery[0].Position - bottom.Position, Vector3.Up);
            foreach (Node n in periphery)
                n.Update();
            foreach (Link l in verticalLinks)
            {
                Node b = l.node1;
                Node t = l.node2;
                Vector3 bp = b.Position;
                bp.Y = platformAltitude;
                Vector3 tp = t.Position;
                Vector3 offset = (bp - tp) / 2;
                offset.Y = 0;
                b.Position = bp - offset;
                t.Position = tp + offset;
            }
        }

        /// <summary>
        /// Constrain all links
        /// </summary>
        public override void Constrain()
        {
            for (int i = 0; i < 2; i++)
            {
                foreach (Link l in visibleLinks)
                    l.Constrain();
                foreach (Link l in invisibleLinks)
                    l.Constrain();
            }
        }

        /// <summary>
        /// Ground-plane location of the merry-go-round
        /// </summary>
        public override Vector3 Position
        {
            get { return bottom.Position; }
        }

        /// <summary>
        /// The object's total mass
        /// </summary>
        public override float Mass
        {
            get { return m_mass; }
        }

        /// <summary>
        /// The object's current center of mass in world coordinates
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get { return Position+Vector3.UnitY*height/4; }
        }

        /// <summary>
        /// Location for a character to fixate when looking at the merry-go-round 
        /// </summary>
        public override Vector3 LookPoint
        {
            get { return top.Position; }
        }

#if OldCollision
        /// <summary>
        /// Check for collisions between character, bars, and base of the merry-go-round.
        /// </summary>
        public override bool TestCollisionWithRagdoll(Ragdoll h)
        {
            Vector3 offset = h.Position - Position;
            bool contact = false;
            // Don't bother if they're too far apart
            if (offset.Length() < (h.CollisionRadius + CollisionRadius))
            {
                // Ick.  Collide every link with every other link
                foreach (Link l in visibleLinks)
                {
                    foreach (Link hl in h.Links)
                    {
                        if (l.CheckCollision(hl, h))
                            contact = true;
                    }
                }

                // Check if this is a supporting surface of any part of the ragdoll's body
                foreach (Node n in h.Nodes)
                {
                    if (n.Position.Y <= platformAltitude && IsSupporting(n))
                    {
                        Vector3 p = n.Position;
                        p.Y = platformAltitude;
                        n.Position = p;
                        n.SupportingSurface = this;
                        n.GroundLevel = platformAltitude;
                    }
                }
            }
            
            return contact;
        }
#endif

        /// <summary>
        /// True if the point is over the merry-go-round
        /// </summary>
        public override bool IsSupporting(Vector3 point)
        {
            return MathUtil.ProjectToGroundPlane(point-Position).LengthSquared()<=(radius*radius);
        }

        /// <summary>
        /// True if supporting some node of o.
        /// </summary>
        public override bool IsSupporting(TwigObject o)
        {
            return SupportingSurfaceDefaultIsSupportingMethod(o);
        }

        /// <summary>
        /// Height at which MerryGoRound supports other objects
        /// </summary>
        public override float SupportingSurfaceLevel
        {
            get
            {
                return platformAltitude;
            }
        }

        /// <summary>
        /// This is a supporting surface
        /// </summary>
        public override bool IsASupportingSurface
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Returns a node frame comprised of three of the pipe bases at roughly equidistant angles.
        /// </summary>
        public override NodeFrame NodeFrame
        {
            get
            {
                return nodeFrame;
            }
        }
        /// <summary>
        /// Can sit on the platform
        /// </summary>
        public override Node SittingTarget
        {
            get
            {
                return platform;
            }
        }

        /// <summary>
        /// Can sit anywhere on the platform.
        /// </summary>
        public override float SittingTargetRadius
        {
            get
            {
                return radius*0.9f;
            }
        }
    }
}
