// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

namespace Twig
{
    #region TypeDefinitions
    /// <summary>
    /// Used for rigidly attaching a node to the coordinate frame of a moving object
    /// </summary>
    public class NodeFrame
    {
        /// <summary>
        /// Create a new NodeFrame comprised of the specified nodes.
        /// </summary>
        public NodeFrame(Node a, Node b, Node c)
        {
            this.a = a;
            this.b = b;
            this.c = c;
        }

        /// <summary>
        /// Nodes with respect to which other nodes should be held rigid.
        /// </summary>
        public Node a, b, c;
    }

    /// <summary>
    /// Represents the position of a node relative to NodeFrame
    /// </summary>
    public struct NodeFrameCoordinates
    {
        /// <summary>
        /// Squared distances to respective node in the NodeFrame
        /// </summary>
        public float a, b, c;
        /// <summary>
        /// Determine the a, b, and c distances of position w.r.t. frame
        /// </summary>
        public void Measure(Vector3 position, NodeFrame frame)
        {
            a = MathUtil.DistanceSquared(position, frame.a.Position);
            b = MathUtil.DistanceSquared(position, frame.b.Position);
            c = MathUtil.DistanceSquared(position, frame.c.Position);
        }
    }

    /// <summary>
    /// Determines whether a node can move freely or is fixed in place.
    /// </summary>
    [Serializable]
    public enum ControlMode
    {
        /// <summary>
        /// Node is fixed in place.  Updates to its position will have no effect.
        /// </summary>
        Fixed,
        /// <summary>
        /// Node can move freely (this is the default mode).
        /// </summary>
        Free,
        /// <summary>
        /// Node moves freely, but will switch to the Fixed state upon making contact with the ground.
        /// </summary>
        FallAndStick
    }

    /// <summary>
    /// Controls breakpointing of node update
    /// </summary>
    [Flags]
    public enum TraceMode
    {
        /// <summary>
        /// No tracing
        /// </summary>
        None=0,
        /// <summary>
        /// Break on positionupdates
        /// </summary>
        Position=1,
        /// <summary>
        /// Break on acceleration updates
        /// </summary>
        Acceleration=2,
        /// <summary>
        /// Break on velocity update
        /// </summary>
        Velocity=4,
        /// <summary>
        /// Break on ground velocity update
        /// </summary>
        GroundVelocity=8,
        /// <summary>
        /// Break when Update() is called on node
        /// </summary>
        Update=16,
        /// <summary>
        /// Break when ControlMode is changed
        /// </summary>
        ControlMode=32
    }
    #endregion

    /// <summary>
    /// Class of particles under control of the physics system.
    /// Articulated figures are build from Links whose endpoints are Nodes,
    /// and which impose rigid-body constraints on the Nodes.
    /// </summary>
    [DebuggerDisplay("{name} : {m_position}")]
    [Serializable]
    public sealed class Node
    {
        #region Instance variables
        Vector3 m_position;
        Vector3 m_pposition;
        float mass = 1.0f;
        float invMass = 1.0f;

        #region Tracing
#if NodeTracing
        /// <summary>
        /// If true, will breakpoint when the node updates
        /// </summary>
        private TraceMode trace;

        /// <summary>
        /// Flag set of cases in which to breakpoint for updates of this node.
        /// </summary>
        public TraceMode Trace
        {
            get { return trace; }
            set { trace = value; }
        }
#endif
        #endregion

        /// <summary>
        /// For debugging purposes only
        /// </summary>
        string name;

        /// <summary>
        /// Object to which this node is associated.
        /// </summary>
        public readonly TwigObject twigObject;

        ControlMode m_mode = ControlMode.Free;

        Vector3 m_acceleration = Physics.gVector;

        TwigObject supportingSurface = null;

        /// <summary>
        /// The object upon which this node is resting, if any.
        /// </summary>
        public TwigObject SupportingSurface
        {
            get { return supportingSurface; }
            set {
                if (value != supportingSurface)
                    OnDismount(supportingSurface);
                if (value != null)
                    OnMount(supportingSurface);
                supportingSurface = value;
                groundLevel = (supportingSurface==null)?0:value.SupportingSurfaceLevel;
                if (m_position.Y < groundLevel)
                {
                    // Move it into position
                    m_position.Y = groundLevel;
                    // Prevent it from suddently acquiring huge upward momentum
                    m_pposition = m_position;
                }
            }
        }

        /// <summary>
        /// Handler for when a Node or TwigObject moves over/off-of a suppporting surface.
        /// </summary>
        public delegate void MountEventHandler(TwigObject supportingSurface);

        /// <summary>
        /// Called when a new object becomes the supporting surface of the node.
        /// </summary>
        public event MountEventHandler Mount;
        /// <summary>
        /// Called when an object ceases to be the supporting surface of the node.
        /// </summary>
        public event MountEventHandler Dismount;

        void OnMount(TwigObject newSupportingSurface)
        {
            if (Mount != null)
                Mount(newSupportingSurface);
            twigObject.OnMount(this, newSupportingSurface);
        }

        void OnDismount(TwigObject oldSupportingSurface)
        {
            if (Dismount != null)
                Dismount(oldSupportingSurface);
            twigObject.OnDismount(this, oldSupportingSurface);
        }

        float groundLevel = 0f;

        /// <summary>
        /// Current height of the SupportingSurface.
        /// </summary>
        public float GroundLevel
        {
            get { return groundLevel; }
            //set { groundLevel = value; }
        }

        NodeFrame currentNodeFrame;

        NodeFrame CurrentNodeFrame
        {
            get { return currentNodeFrame; }
            set { 
                currentNodeFrame = value;
                if (value!=null)
                    frameCoordinates.Measure(m_position, currentNodeFrame);
            }
        }
        NodeFrameCoordinates frameCoordinates;

        void EnforceDistanceConstraint(Node n, float distanceSquared)
        {
            float invSum = n.InvMass + invMass;
            Vector3 delta = n.Position - Position;
            delta *= (distanceSquared / (delta.LengthSquared() + distanceSquared) - 0.5f) / invSum;
            n.Position += n.InvMass*delta;
            Position -= invMass*delta;
#if notdef
            Vector3 delta = n.Position - Position;
            delta *= (distanceSquared / (delta.LengthSquared() + distanceSquared) - 0.5f) ;
            n.Position += delta;
            m_position -= delta;
#endif
        }

        void EnforceFrameCoordinates()
        {
            if (Physics.tickCount % 2 == 0)
            {
                EnforceDistanceConstraint(currentNodeFrame.a, frameCoordinates.a);
                EnforceDistanceConstraint(currentNodeFrame.b, frameCoordinates.b);
                EnforceDistanceConstraint(currentNodeFrame.c, frameCoordinates.c);
            }
            else
            {
                EnforceDistanceConstraint(currentNodeFrame.c, frameCoordinates.c);
                EnforceDistanceConstraint(currentNodeFrame.b, frameCoordinates.b);
                EnforceDistanceConstraint(currentNodeFrame.a, frameCoordinates.a);
            }
        }


        #endregion

        #region Static variables

        static float airDamping = 0.01f;
        static float groundDamping = 0.9f;
        static float staticFrictionThresholdSq = 1.0f;

        static Random random = new Random();
        #endregion

        #region Accessor and mutators for Physical properties
        /// <summary>
        /// The Node's acceleration.
        /// </summary>
        public Vector3 Acceleration
        {
            get
            {
                return m_acceleration;
            }
            set
            {
#if NodeTracing
                if ((trace & TraceMode.Acceleration) != 0)
                    System.Diagnostics.Debugger.Break();
#endif

                Debug.Assert(MathUtil.IsProperVector(value), "Setting Acceleration to invalid vector");

                m_acceleration = value;
            }
        }

        /// <summary>
        /// The Nodes's mass.
        /// NB: most calculations presently ignore mass.
        /// </summary>
        public float Mass
        {
            get
            {
                return mass;
            }
            set
            {
                mass = value;
                invMass = 1f / mass;
            }
        }
        /// <summary>
        /// Reciprocal of mass.
        /// </summary>
        public float InvMass
        {
            get
            {
                return invMass;
            }
            set
            {
                invMass = value;
                mass = 1f / invMass;
            }
        }

        /// <summary>
        /// Receives notifications of updates to node positions.
        /// </summary>
        /// <param name="newPosition"></param>
        public delegate void NodePositionListener(Vector3 newPosition);

        /// <summary>
        /// Signaled when a node's position is updated.
        /// </summary>
        public event NodePositionListener PositionUpdated;

        /// <summary>
        /// The node's current position.
        /// Assignments enforces ground non-penetration.  Also assignment will have no effect if object is fixed.
        /// </summary>
        public Vector3 Position
        {
            get
            {
                return m_position;
            }
            set
            {
                Debug.Assert(MathUtil.IsProperVector(value), "Setting Position to invalid vector");
#if NodeTracing
                if ((trace&TraceMode.Position)!=0)
                    System.Diagnostics.Debugger.Break();
                if (PositionUpdated!=null)
                    PositionUpdated(value);
#endif
                if (m_mode!=ControlMode.Fixed && !(OnGround&&m_mode==ControlMode.FallAndStick))
                    m_position = new Vector3(value.X, Math.Max(0f, value.Y), value.Z);
                else if (m_mode != ControlMode.Fixed)
                    m_position = value;

                if (m_position.Y < GroundLevel)
                {
                    m_position.Y = GroundLevel;
                    if (m_mode == ControlMode.FallAndStick)
                    {
                        if (supportingSurface != null && supportingSurface.NodeFrame!=null)
                        {
                            CurrentNodeFrame = supportingSurface.NodeFrame;
                            m_mode = ControlMode.Free;
                        } else
                            m_mode = ControlMode.Fixed;
                    }
                }
            }
        }

        /// <summary>
        /// The node's position projected into the plane Y=0.
        /// </summary>
        public Vector3 GroundPlanePosition
        {
            get
            {
                Vector3 p = Position;
                p.Y = 0;
                return p;
            }
        }

        /// <summary>
        /// Returns Position, if we aren't looping the log, otherwise the logged position.
        /// Always returns Position if the node isn't Logged.
        /// </summary>
        public Vector3 PositionForDisplay
        {
            get
            {
#if NodeTracing
                if (Logged && nowPlayingBackLog)
                    return nodeLog[logPlaybackFrame, logPosition];
                else
                    return Position;
#else
                return Position;
#endif
            }
        }

        /// <summary>
        /// The node's velocity, i.e. rate of movement.
        /// </summary>
        public Vector3 Velocity
        {
            get
            {
                return (m_position - m_pposition) * (1.0f / Physics.deltaT);
            }
            set
            {
#if NodeTracing
                if ((trace & TraceMode.Velocity) != 0)
                    System.Diagnostics.Debugger.Break();
#endif

                Debug.Assert(MathUtil.IsProperVector(value), "Setting Velocity to invalid vector");
                m_pposition = m_position - (value * Physics.deltaT);
            }
        }

        /// <summary>
        /// The node's (vector) momentum, i.e. velocity*mass
        /// </summary>
        public Vector3 Momentum
        {
            get
            {
                return Velocity * mass;
            }
        }

        /// <summary>
        /// The node's Kinetic energy, i.e. 1/2 mv^2
        /// </summary>
        public float KineticEnergy
        {
            get
            {
                return Velocity.LengthSquared() * mass * 0.5f;
            }
        }

        /// <summary>
        /// Projection of velocity into the XZ plane.
        /// </summary>
        public Vector3 GroundVelocity
        {
            get
            {
                return MathUtil.ProjectToGroundPlane(Velocity);
            }
            set
            {
#if NodeTracing
                if ((trace & (TraceMode.Velocity|TraceMode.GroundVelocity)) != 0)
                    System.Diagnostics.Debugger.Break();
#endif
                Debug.Assert(MathUtil.IsProperVector(value), "Setting GroundVelocity to invalid vector");
                Vector3 v = m_position - (value * Physics.deltaT);
                m_pposition = new Vector3(v.X, m_pposition.Y, v.Z);
            }
        }


        /// <summary>
        /// True if the object is at Y=0.
        /// </summary>
        public bool OnGround
        {
            get
            {
#if GroundAtZero
                return Position.Y <= 0.0f;
#endif
            }
        }

        /// <summary>
        /// Damping factor applied to node's velocity.
        /// This varies depending on whether the node is in the air or on the ground.
        /// </summary>
        float Damping
        {
            get
            {
                return OnGround ? groundDamping : airDamping;
            }
        }

        float bodyCenteredDamping=0.1f;

        /// <summary>
        /// Damping coefficient to be used with body-centered damping
        /// </summary>
        public float BodyCenteredDamping
        {
            get { return bodyCenteredDamping; }
            set { bodyCenteredDamping = value; }
        }
        Node bodyReference;

        /// <summary>
        /// Node relative to which this node is damped for body-centered damping
        /// </summary>
        public Node BodyReference
        {
            get { return bodyReference; }
            set { bodyReference = value; }
        }

        #endregion

        /// <summary>
        /// Creates a node in a random position near the origin.
        /// </summary>
        public Node(TwigObject o, string name)
        {
            this.twigObject = o;
            this.name = name;
            m_position = new Vector3((float)random.NextDouble()-0.5f, (float)random.NextDouble()-0.5f, (float)random.NextDouble()-0.5f);
            m_pposition = m_position;
        }

        /// <summary>
        /// Creates a node at the specified position
        /// </summary>
        public Node(TwigObject o, string name, Vector3 position)
        {
            twigObject = o;
            this.name = name;
            Position = position;
            m_pposition = position;
        }

        /// <summary>
        /// Determines whether the node is fixed or moveable.
        /// </summary>
        public ControlMode ControlMode
        {
            get
            {
                return m_mode;
            }
            set
            {
#if NodeTracing
                if ((Trace&TraceMode.ControlMode)!=0)
                    Debugger.Break();
#endif
                m_mode = value;
                if (value == ControlMode.Fixed)
                    m_pposition = m_position;
                if (value != ControlMode.FallAndStick)
                    currentNodeFrame = null;
            }
        }

        /// <summary>
        /// Updates the node's position, velocity, etc.
        /// </summary>
        public void Update()
        {
            Vector3 save = Position;
            float damp = Damping;

            // Handle static friction when on the ground.
            if (OnGround && (save - m_pposition).LengthSquared() < staticFrictionThresholdSq)
                m_pposition = save;

#if NodeTracing
            if ((trace& TraceMode.Update) != 0)
                Debugger.Break();
#endif

            Vector3 newPosition = ((2.0f - damp) * save) - ((1.0f - damp) * m_pposition) + (Acceleration * Physics.deltaTSq);
            if (BodyReference != null)
                newPosition -= BodyCenteredDamping * ((save - m_pposition) - (bodyReference.m_position - bodyReference.m_pposition));
            Position = newPosition;
            m_pposition = save;
            if (currentNodeFrame != null)
                EnforceFrameCoordinates();
        }

        /// <summary>
        /// If move node so as to make its projection onto direction at least as large as the specified value.
        /// </summary>
        public void ConstrainGE(Vector3 direction, float distance)
        {
            float projection = Vector3.Dot(Position, direction);
            if (projection < distance)
            {
                Vector3 delta = direction * (distance - projection);
                Position = Position + delta;
                Velocity = Vector3.Zero;
            }
        }

        /// <summary>
        /// If move node so as to make its projection onto direction no more than the specified value.
        /// </summary>
        public void ConstrainLE(Vector3 direction, float distance)
        {
            float projection = Vector3.Dot(Position, direction);
            if (projection > distance)
            {
                Vector3 delta = direction * (distance - projection);
                Position = Position + delta;
                Velocity = Vector3.Zero;
            }
        }

        #region Motion logging
        /// <summary>
        /// WARNING: Node logging uses static variables, so it isn't reentrant
        /// I.e. you can't run two TwigGame objects in one AppDomain with logging enabled.
        /// Since this is just a debugging feature and it's not clear anyone would ever want
        /// to do this anyway, I'm not going to worry about changing the code to support it.
        /// </summary>
#if NodeLogging
        public bool Logged
        {
            get
            {
                return logPosition >= 0;
            }
        }

        short logPosition = -1;
        static short nextLogIndex = 0;
        /// <summary>
        /// The actual log of old node positions indexed by [frameCount%logLength, node index]
        /// </summary>
        static Vector3[,] nodeLog;
        // Log the last 60 node positions (i.e. 1 second)
        const short logLength = 60;
        static int logPlaybackFrame, logPlaybackEndFrame;
        static bool nowPlayingBackLog = false;
        static int pauseAtEndCounter;

        static List<Node> loggedNodes = new List<Node>();
        static Node[] cachedLoggedNodes;  // cached in an array for iteration w/o GC

        /// <summary>
        /// Arranges for a log of the last second or so of motion of the node to be kept.
        /// </summary>
        public void EnableLogging()
        {
            Debug.Assert(cachedLoggedNodes == null, "Logging must be enabled for a node before update begins.");
            if (logPosition < 0)
                logPosition = nextLogIndex++;
            loggedNodes.Add(this);
        }

        /// <summary>
        /// Called at the end of TwigGame.Update().  Copies all logged node positions to log.
        /// </summary>
        static public void LogNodes()
        {
            InsureLogCreated();
            int logFrame = Physics.tickCount%logLength;
            foreach (Node n in cachedLoggedNodes) {
                nodeLog[logFrame, n.logPosition] = n.m_position;
            }
        }

        static void InsureLogCreated()
        {
            if (cachedLoggedNodes == null)
            {
                cachedLoggedNodes = loggedNodes.ToArray();
                nodeLog = new Vector3[logLength, cachedLoggedNodes.Length];
            }
        }

        /// <summary>
        /// Switches display code to show logged node positions rather than updated positions
        /// </summary>
        static void StartLogPlayback()
        {
            nowPlayingBackLog = true;
            logPlaybackEndFrame = logPlaybackFrame = (Physics.tickCount+1) % logLength;
        }

        /// <summary>
        /// Returns to using current node positions for display
        /// </summary>
        static void CancelLogPlayback()
        {
            nowPlayingBackLog = false;
        }

        /// <summary>
        /// Updates display position within log
        /// </summary>
        static public void AdvanceLog()
        {
            if (logPlaybackFrame == logPlaybackEndFrame && pauseAtEndCounter < 50)
                pauseAtEndCounter++;
            else
            {
                logPlaybackFrame = (logPlaybackFrame + 1) % logLength;
                pauseAtEndCounter = 0;
            }
        }

        /// <summary>
        /// When true, display of logged nodes runs an animation loop
        /// of the last second of recorded motion
        /// </summary>
        static public bool PlaybackLog
        {
            get
            {
                return nowPlayingBackLog;
            }
            set
            {
                if (value != nowPlayingBackLog)
                {
                    if (nowPlayingBackLog)
                        CancelLogPlayback();
                    else
                        StartLogPlayback();
                }
            }
        }
#endif
        #endregion
    }
}