// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A pen/pencil
    /// </summary>
    public class Pen : TwigObject
    {
        #region Instance variables
        Node top;
        Node tip;
        Node grip;
        Link body;
        Chart writingSurface;
        Vector3 writingPosition;
        #endregion

        double scribbleEndTime = 0;

        const float length = 0.17f;
        const float radius = 0.01f;

        #region Constructor
        /// <summary>
        /// Make a pen/pencil
        /// </summary>
        public Pen(TwigGame g, string name, Vector3 position)
            : base(g, name, true)
        {
            grip = new Node(this, name + " grip", position);
            tip = new Node(this, name + " tip", position + 0.5f * length * Vector3.UnitX);
            top = new Node(this, name + " top", position - 0.5f * length * Vector3.UnitX);
            Color = Color.Yellow;
            body = new LinkWithMidpoint(name, this, top, grip, tip, length, radius);
            nodes = new Node[] { tip, grip, top };
            collisionElements = new CollisionElement[] { body.CollisionElement };
        }
        #endregion

        #region Update, Constrain, Draw
        /// <summary>
        /// Draws body of pen
        /// </summary>
        public override void Draw(GameTime t)
        {
            BasicEffect.Begin();
            foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                ConfigureMaterial(); // Don't need to CommitChanges() because that will be done in body.Draw().
                body.Draw();
                pass.End();
            }
            BasicEffect.End();
        }

        /// <summary>
        /// Update nodes in object
        /// </summary>
        public override void Update(GameTime gameTime)
        {
            if (!IsGrasped)
                writingSurface = null;

            if (writingSurface != null)
            {
                Vector3 normal = writingSurface.Normal(writingPosition);
                Vector3 writingOffset = scribbleEndTime<Physics.time?
                    new Vector3(0,0.1f, 0) : new Vector3(0.02f * (float)Math.Sin(Physics.time * 20), 0, 0);
                tip.Position = writingSurface.ToGlobal(writingPosition + writingOffset);
                top.Position = tip.Position + normal * length;
                //TwigGame.DrawDebugMarker(tip.Position, Color.Yellow);
                if (IsGrasped)
                    GraspingLimb.SnapToGraspedNode();
            }

            foreach (Node n in nodes)
                n.Update();
        }

        /// <summary>
        /// Fix positioning of nodes.
        /// </summary>
        public override void Constrain()
        {
            body.Constrain();
        }
        #endregion

        /// <summary>
        /// Hold the pen
        /// </summary>
        public override void SetupHold(Mammal reader, Grasp grasp)
        {
            reader.Hold(grip, this, grasp);
        }

        /// <summary>
        /// Select hand for grasping.  Will grasp the *grasp* node, with a single-hand grip.
        /// </summary>
        public override void SelectGrasp(Mammal grasper, Grasp graspRequest,
                                 out Grasp selectedGrasp, out Node leftTarget, out Node rightTarget)
        {
            grasper.SelectGrasp(grip, grip, this, (Grasp)(graspRequest&~Grasp.TwoHanded),
                                out selectedGrasp, out leftTarget, out rightTarget);
        }

        /// <summary>
        /// Hold the pen
        /// </summary>
        public override void SetupHoldForUse(Mammal reader, Grasp grasp)
        {
            SetupHold(reader, grasp);
        }

        /// <summary>
        /// Move the pen tip to a specified writing surface
        /// </summary>
        public void WriteOn(Chart surface, Vector3 position)
        {
            writingSurface = surface;
            writingPosition = position;
        }

        /// <summary>
        /// Move the pen tip to a piece of paper
        /// </summary>
        public void WriteOn(Paper paper, Vector3 position)
        {
            writingSurface = paper.Atlas["front"];
            writingPosition = position;
            TwigObject character = paper.Grasper;
            if (character != null && character is Mammal)
                ((Mammal)character).HoldForUse(paper);
        }

        /// <summary>
        /// Write on the paper at position (0,0,0)
        /// </summary>
        public void WriteOn(Paper paper)
        {
            WriteOn(paper, Vector3.Zero);
        }

        /// <summary>
        /// Scribble on the current writing surface for specified time.  Also causes grasper to look at pen.
        /// </summary>
        public void Scribble(float time)
        {
            Mammal grasper = Grasper as Mammal;
            if (grasper != null)
                grasper.LookAt(this);

            scribbleEndTime = Physics.time + time;
        }

        /// <summary>
        /// True if pen is in the process of writing
        /// </summary>
        public override bool IsBeingManipulated
        {
            get
            {
                return Physics.time < scribbleEndTime;
            }
        }

        #region Derived properties
        /// <summary>
        /// Pen position is its tip.
        /// </summary>
        public override Vector3 Position
        {
            get { return tip.Position; }
        }

        /// <summary>
        /// Mass is mass of link
        /// </summary>
        public override float Mass
        {
            get { return body.Mass; }
        }

        /// <summary>
        /// CoM is grip point
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get { return grip.Position; }
        }

        /// <summary>
        /// Collision radius is length
        /// </summary>
        public override float CollisionRadius
        {
            get { return length; }
        }

        /// <summary>
        /// Lookpoint is grip
        /// </summary>
        public override Vector3 LookPoint
        {
            get { return tip.Position; }
        }

        /// <summary>
        /// Characters will walk over the pen without avoiding it
        /// </summary>
        public override bool IsObstacle
        {
            get
            {
                return false;
            }
        }
        #endregion
    }
}
