// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using System.Diagnostics;

namespace Twig
{
    /// <summary>
    /// Person extends the Mammal class with attention, affect and motivation systems
    /// </summary>
    public abstract class Person : Mammal
    {
        /// <summary>
        /// Create person with specified bone sizes
        /// </summary>
        public Person(TwigGame g, string name, float spineLength, float spineRadius, float armSpacing, float armRadius, float upperArm, float foreArm,
            float pelvis, float legRadius, float femur, float shin, float headHeight)
            : base(g, name, spineLength, spineRadius, armSpacing, armRadius, upperArm, foreArm, pelvis, legRadius, femur, shin, headHeight)
        {
            Height = femur + shin + spineLength + headHeight;
            //EyeLevel = Height-headHeight/2f;
            Atlas["torso"] = new TorsoChart(this);
        }

        /// <summary>
        /// Salience the charater attributes to the object.
        /// </summary>
        public override float Salience(TwigObject o)
        {
            return Appraisal(o).Salience;
        }

        /// <summary>
        /// Enumerates the object in the character's view.
        /// </summary>
        public override IEnumerable<TwigObject> ObjectsInView
        {
            get
            {
                Vector3 myPosition = Position;
                foreach (TwigObject o in TwigObject.AllTwigObjects)
                {
                    Vector3 offset = o.Position - myPosition;
                    // Treat any object in front of the Person as in view
                    if (o != this && Vector3.Dot(offset, lookForward) > 0f)
                        yield return o;
                }
            }
        }

        // Internal state variable
        /// <summary>
        /// True if left leg should be the next to take a step.
        /// </summary>
        bool leftLegNext;
        /// <summary>
        /// Efferent.  Enables the swinging of arms when walking.
        /// </summary>
        public bool swingArms = true;

        const float maximalComfortableLegExtension = 0.85f;

        /// <summary>
        /// Control the actual walking
        /// </summary>
        protected override void GaitControl()
        {
            bool stepInProgress = legs.left.InBallisticMotion || legs.right.InBallisticMotion;
#if DebugGait
            Chart("Left extension", legs.left.Extension, 1.5f);
            Chart("Right extension", legs.right.Extension, 1.5f);
            // Chart("Walk", walk ? 1 : -1, 1.5f);
            Chart("Step possible", !stepInProgress ? 1 : -1, 3f);
            Chart("Left time", legs.left.ballisticTimeRemaining, 0.5f);
            Chart("Right time", legs.right.ballisticTimeRemaining, 0.5f);
#endif
            Vector3 realWalkVector = walkDirection * walkSpeed;
            targetKneeBend = walk ? walkingTargetKneeBend : 0f;
            if (walk && standingUp)
            {
                // Glide the pelvis along the walk vector
                // Don't move until pointed approximately the right direction
                if (Vector3.Dot(walkDirection, pelvisForward) > 0.3f)
                {
                    legs.left.root.GroundVelocity = realWalkVector;
                    legs.right.root.GroundVelocity = realWalkVector;
                }

                Vector3 leftVector = legs.left.root.Position - legs.left.end.Position;
                // Positive when leg is behind body
                float leftStrideProjection = Vector3.Dot(leftVector, walkVector);
                Vector3 rightVector = legs.right.root.Position - legs.right.end.Position;
                // Positive when leg is behind body
                float rightStrideProjection = Vector3.Dot(rightVector, walkVector);


                // Take a step, if appropriate
                if (!stepInProgress
                    || (legs.right.Extension > 1 && leftStrideProjection < 0)
                    || (legs.left.Extension > 1 && rightStrideProjection < 0))
                {
                    // Override leg priority if leg is overstretched
                    if (legs.left.Extension > maximalComfortableLegExtension && legs.left.Extension > legs.right.Extension)
                        leftLegNext = true;
                    else if (legs.right.Extension > maximalComfortableLegExtension && legs.right.Extension > legs.left.Extension)
                        leftLegNext = false;

                    if (leftLegNext)
                    {
                        if (legs.left.Extension > maximalComfortableLegExtension || leftStrideProjection > stride
                            || (leftStrideProjection > 0 && rightStrideProjection > 0))
                        {
                            Vector3 target = stride * walkDirection + legs.left.root.Position + ballisticStepTime * realWalkVector * 0.3f;
                            target.Y = stepTargetHeight;
                            legs.left.Goto(target, ballisticStepTime, ControlMode.FallAndStick);
                            if (swingArms && !hug)
                                swingRightArm = 5f * walkSpeed;
                            leftLegNext = false;
                        }
                        else
                            WaitingFor("GaitControl: Can step left");
                    }
                    else
                    {
                        if (legs.right.Extension > maximalComfortableLegExtension || rightStrideProjection > stride
                            || (leftStrideProjection > 0 && rightStrideProjection > 0))
                        {
                            Vector3 target = stride * walkDirection + legs.right.root.Position + ballisticStepTime * realWalkVector * 0.3f;
                            target.Y = stepTargetHeight;
                            legs.right.Goto(target, ballisticStepTime, ControlMode.FallAndStick);
                            if (swingArms && !hug)
                                swingLeftArm = 5f * walkSpeed;
                            leftLegNext = true;
                        }
                        else
                            WaitingFor("GaitControl: Can step right");
                    }
                }
            }
            else if (!standUpFromSittingPosition)
            {
                if (walk)
                    WaitingFor("GaitControl: standingUp");
                legs.left.root.GroundVelocity = Vector3.Zero;
                legs.right.root.GroundVelocity = Vector3.Zero;
                legs.left.joint.Acceleration = Physics.gVector;
                legs.right.joint.Acceleration = Physics.gVector;
                legs.left.end.Acceleration = beingHugged ? Physics.gVector : 5 * Physics.gVector;
                legs.right.end.Acceleration = beingHugged ? Physics.gVector : 5 * Physics.gVector;
                if (beingHugged)
                {
                    legs.left.end.ControlMode = ControlMode.FallAndStick;
                    legs.right.end.ControlMode = ControlMode.FallAndStick;
                }
            }
        }

        /// <summary>
        /// Ground level for this character
        /// </summary>
        public override float GroundLevel
        {
            get
            {
                return Math.Min(legs.left.end.GroundLevel, legs.right.end.GroundLevel);
#if notdef
                if (legs.left.end.SupportingSurface != null)
                    return legs.left.end.GroundLevel;
                else if (legs.right.end.SupportingSurface != null)
                    return legs.right.end.GroundLevel;
                else 
                    return 0f;
#endif
            }
        }

        /// <summary>
        /// Manage posture to maintain appearance of balance.
        /// </summary>
        protected override void StaticBalance()
        {
            if (walkSpeed == 0)
            {
                Vector3 upperBody = (torso + pelvis) * 0.5f;
                Vector3 offset = upperBody - footMidpoint;
                offset.Y = 0;

                if (offset.Length() > pelvisWidth * 0.8f)
                {
                    float forwardComponent = Vector3.Dot(offset, pelvisForward);
                    float rightComponent = Vector3.Dot(offset, pelvisRight);
                    if (Math.Abs(forwardComponent) > Math.Abs(rightComponent))
                    {  //Step forward/back

                    }
                    else // Step sideways
                    {
                        if (rightComponent > 0)
                            legs.right.Goto(legs.right.end.Position + offset, ballisticStepTime, ControlMode.FallAndStick);
                        else
                            legs.left.Goto(legs.left.end.Position + offset, ballisticStepTime, ControlMode.FallAndStick);
                    }
                }
            }
        }

        /// <summary>
        /// Apply geometric constraints to body
        /// </summary>
        public override void Constrain()
        {
            Vector3 outInFrontOfChest = spineTop.Position + bodyForward * arms.left.upper.Length;
            ConstrainElbow(arms.left, outInFrontOfChest);
            ConstrainElbow(arms.right, outInFrontOfChest);
            base.Constrain();
        }

        /// <summary>
        /// Constrains elbow to lie in the half-plane formed by end, root, and constraintPoint
        /// </summary>
        /// <param name="arm"></param>
        /// <param name="constraintPoint"></param>
        void ConstrainElbow(Limb arm, Vector3 constraintPoint)
        {
            Vector3 planeNormal = Vector3.Cross(arm.end.Position - arm.root.Position, constraintPoint - arm.root.Position);
            planeNormal.Normalize();

            // Force elbow into plane
            Vector3 planeCorrection = Vector3.Dot(arm.joint.Position - arm.end.Position, planeNormal) * 0.5f * planeNormal;
            arm.joint.Position -= planeCorrection;
            arm.end.Position += planeCorrection;
            // Move elbow outside of triangle formed by root, end, constraint
            Vector3 bendDirection = Vector3.Cross(planeNormal, arm.end.Position - arm.root.Position);
            bendDirection.Normalize();
            float bendError = Vector3.Dot(arm.joint.Position - arm.end.Position, bendDirection) * 0.5f;
            if (bendError > 0)
            {
                Vector3 bendCorrection = bendError * bendDirection;
                arm.joint.Position -= bendCorrection;
                arm.end.Position += bendCorrection;
            }
        }
#if notdef
        public void Wave()
        {

            //arms.right.Pose(new LimbPose(LimbNormalization.Limb, 0, 1f, 0.4f), 1000);
            arms.right.Play(new LimbAnimation(new NodeAnimation(LimbNormalization.Limb,
                //new NodeKey(0, 0, -1, 0),
                //new NodeKey(0.25f, 0, 0.8f, 0.2f),
                new NodeKey(0f, 0, 1, 0.4f), new NodeKey(0.75f, 0, 1, 0.4f)),
                null, false), 1000);
        }
#endif

        /// <summary>
        /// Play the named gesture through the specified arm
        /// </summary>
        public Action Gesture(string arm, string name)
        {
            LimbAnimation a = TwigGame.Content.Load<LimbAnimation>(name);
            switch (arm)
            {
                case "left":
                    return new GestureAction(arms.left, a);

                case "right":
                    return new GestureAction(arms.right, a);

                case "both":
                    return new GestureAction(arms, a);

                default:
                    throw new Exception("Bad arm name in gesture: " + arm);
            }
        }

        internal class GestureAction : Action
        {
            public GestureAction(Limb limb, LimbAnimation gesture)
            {
                this.limb = limb;
                secondLimb = null;
                this.gesture = gesture;
            }

            public GestureAction(LimbPair limbs, LimbAnimation gesture)
            {
                limb = limbs.left;
                secondLimb = limbs.right;
                this.gesture = gesture;
            }

            Limb limb, secondLimb;
            LimbAnimation gesture;

            public override void Start()
            {
                limb.Play(gesture);
                if (secondLimb != null)
                    secondLimb.Play(gesture);
            }

            public override void Abort()
            {
                if (limb.CurrentAnimation == gesture)
                    limb.Play(null);
                if (secondLimb != null && secondLimb.CurrentAnimation == gesture)
                    secondLimb.Play(null);
            }

            public override ActionStatus Status
            {
                get
                {
                    bool running = (limb.CurrentAnimation == gesture) || (secondLimb != null && secondLimb.CurrentAnimation == gesture);
                    return running ? ActionStatus.Running : ActionStatus.Succeed;
                }
            }
        }

        class TorsoChart : PlanarChart
        {
            Person me;

            public TorsoChart(Person me)
            {
                this.me = me;
            }

            public override Matrix GlobalTransform
            {
                get { return Matrix.CreateScale(me.arms.left.length) 
                    * MathUtil.ChangeOfBasis(me.bodyRight, me.bodyUp, me.bodyForward)
                    * Matrix.CreateTranslation(me.spineBottom.Position); }
            }

            public override Vector3 Normal(Vector3 local)
            {
                throw new Exception("The method or operation is not implemented.");
            }
        }
    }
}
