// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

namespace Twig
{
    /// <summary>
    /// Static container for physics-related definitions.
    /// </summary>
    public class Physics
    {
        /// <summary>
        /// Acceleration due to gravity
        /// </summary>
        public const float Gravity = -9.797645f;   // We be metric

        /// <summary>
        /// Density of water in kilograms per cubic meter
        /// </summary>
        public const float WaterDensity = 100*100*100/1000;  // Water is 1g/cm3, so convert to kg/m3

        /// <summary>
        /// Gravitational acceleration vector
        /// </summary>
        public static readonly Vector3 gVector = new Vector3(0, Gravity, 0);

        //
        // In theory, I should pass an XNA GameTime object everywhere to figure
        // out elapsed time, but the update code needs to update elapsed time too
        // often and in too many places for that to be reasonable.
        //

        /// <summary>
        /// Game-time duration of a tick.
        /// </summary>
        public const float deltaT = 1 / 60f;
        /// <summary>
        /// Square of game-time duration of a tick.
        /// </summary>
        public const float deltaTSq = deltaT * deltaT;
        /// <summary>
        /// Number of ticks (calls to Update()) that have occurred.
        /// </summary>
        public static int tickCount = 0;
        /// <summary>
        /// Current game time.
        /// </summary>
        public static double time = 0;

        /// <summary>
        /// True if initial settling of the game engine has completed.
        /// </summary>
        public static bool IsSettled = false;

        /// <summary>
        /// True if we're in slow motion (not yet supported).
        /// </summary>
        public static bool SlowMo
        {
            get
            {
                return false;   // Not yet supported
            }
        }

        /// <summary>
        /// Updates physics system.  Right now, this mostly just runs the objects' constrain methods.
        /// </summary>
        public static void Update(GameTime t)
        {
            tickCount++;
            // this behaves *very* strangely
            //time = t.TotalGameTime.Seconds;
            time += deltaT;
            TwigObject.ConstrainAllTwigObjects();
            IsSettled = tickCount>60;
            // Clear out any SupportingSurface relations that aren't active anymore
            foreach (TwigObject o in TwigObject.DynamicTwigObjects)
            {
                foreach (Node n in o.Nodes)
                {
                    if (n.SupportingSurface != null)
                    {
                        if (!n.SupportingSurface.IsSupporting(n.Position))
                            n.SupportingSurface = null;
                    }
                }
            }

            CheckAndHandleCollisions();
        }

        /// <summary>
        /// Check for, and handle, all collisions between all collision elements of all twig objects
        /// </summary>
        static void CheckAndHandleCollisions()
        {
            // Check every TwigObject against every other TwigObject.
            for (int i = 0; i < TwigObject.AllTwigObjects.Count; i++)
            {
                TwigObject o1 = TwigObject.AllTwigObjects[i];
                if (o1.Cloaked)
                    continue;
                // All the dynamic objects come first in the list, so when we hit a static one, we're done.
                if (!o1.Dynamic)
                    return;
                
                for (int j = i + 1; j < TwigObject.AllTwigObjects.Count; j++)
                {
                    TwigObject o2 = TwigObject.AllTwigObjects[j];

                    if (o2.Cloaked)
                        continue;
                    else
                        CollisionElement.CheckAndHandleCollisions(o1, o2);
                }
            }
        }

        /// <summary>
        /// Converts seconds to update cycles.
        /// </summary>
        public static int SecondsToTicks(float seconds)
        {
            return (int)(seconds/deltaT);
        }
    }
}
