// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

//#define DEBUG_POSTURE
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Twig
{
    partial class Mammal
    {
        #region Other posture-specific parameters
        /// <summary>
        /// Ideal knee bend when leg is extended while walking.
        /// </summary>
        public float walkingTargetKneeBend = 0.05f;
        /// <summary>
        /// Ideal knee bend when leg is extended while standing.
        /// </summary>
        public float targetKneeBend = 0.05f;
        #endregion

        // Internal intermediate variables updated in Constrain()
        Vector3 leftLegDirection;
        Vector3 leftLegForward;
        Vector3 leftLegPlaneNormal;
        Vector3 leftKneeForward;
        Vector3 rightLegDirection;
        Vector3 rightLegForward;
        Vector3 rightLegPlaneNormal;
        Vector3 rightKneeForward;
        Vector3 comOffset;

        /// <summary>
        /// True when at least one foot is on the ground.
        /// Can't do much if there's no leverage with the ground.
        /// </summary>
        bool footOnGround;

        /// <summary>
        /// Updates nodes to maintain posture.
        /// Includes sitting up, standing up, orientation control for head, shoulders, pelvis, and feet.
        /// Also respiration.
        /// </summary>
        protected virtual void PostureControl()
        {
            comOffset = CenterOfMass - Position;
            comOffset.Y = 0;

            footOnGround = legs.left.end.OnGround || legs.right.end.OnGround;
            SitUpController();
            StandUpController();
            ShoulderController();
            PelvisController();
            HeadController();
            FootController();
            ArmController();
            RespirationController();
            arms.linkage.offset = bodyUp * diaphragmPosition * spineLength * 0.1f;
            
            // Apply forces and update
            arms.left.root.Acceleration = (tForce - shoulderOffset) - faceGazeDirectionError;
            arms.right.root.Acceleration = tForce + shoulderOffset + faceGazeDirectionError;
            legs.left.root.Acceleration = pForce + faceWalkForce + pelvisOffset;
            legs.right.root.Acceleration = (pForce - faceWalkForce) - pelvisOffset;
            //spineJoint.Acceleration = tForce;
        }

        #region Sitting up
        /// <summary>
        /// Efferent.  Directs posture control to sit up.
        /// </summary>
        public bool sitUp = true;
        /// <summary>
        /// Afferent.  True when sitting up.
        /// </summary>
        public bool sittingUp;

        // Internal
        float tGain = 1000f;
        float sitUpBoost = 500f;
        float spineExtension=1f;
        Vector3 leanForward;

        /// <summary>
        /// Restoring force on torso (top of spine) applied to correct posture.
        /// </summary>
        Vector3 tForce;

        /// <summary>
        /// Compute tForce to move top of torso into sitting-up posture.
        /// </summary>
        protected virtual void SitUpController()
        {
            // Compute force to sit torso up
            if ((standUp || sitUp) &&
                // Temporarily disable situp when reaching
                !(arms.left.InBallisticMotion || arms.right.InBallisticMotion))
            {
                // Center shoulders over pelvis or pseudo-center-of-mass
                Vector3 shoulderMisalignment;

                shoulderMisalignment = (standUp && !standingUp) ? Vector3.Zero : (pelvis - torso);
                if (standUp)
                {
                    shoulderMisalignment -= comOffset / 5f;
                    leanForward = walkDirection * stride * (stride / legLength) * 0.5f;
                    if (walk)
                        shoulderMisalignment += leanForward;
                }

                shoulderMisalignment.Y = 0;

                sittingUp = shoulderMisalignment.Length() < spineLength * 0.2f;

                tForce = shoulderMisalignment * tGain;

                spineConstraint.Length = spineLength * spineExtension;

                if (torso.Y < pelvis.Y + (0.5f * spineLength))
                    // need a boost to overcome gravity
                    tForce.Y = sitUpBoost;
            }
            else
            {
                WaitingFor("SitUpController: Stand up or sit up command");
                tForce = Physics.gVector;
            }
        }
        #endregion

        #region Standing up
        /// <summary>
        /// Efferent.  Directs posture control to stand up.
        /// </summary>
        public bool standUp = true;
        /// <summary>
        /// Afferent.  True when standing up.
        /// </summary>
        public bool standingUp;
        /// <summary>
        /// Afferent.  Pelvis is centered over foot midpoint.
        /// </summary>
        bool pelvisCentered;

        /// <summary>
        /// Restoring force on pelvis to maintain posture.
        /// </summary>
        Vector3 pForce;

        float pGain = 100f;
        float standGain = 2000f;
        /// <summary>
        /// Compute pForce to move pelvis into standing position.
        /// </summary>
        protected virtual void StandUpController()
        {
            float pelvisError = (pelvis.Y-GroundLevel) - legLength;
            float pelvisRelativeError = pelvisError / legLength;
            // We're standing if our pelvis height is within 30% of our leg length
            standingUp = Math.Abs(pelvisRelativeError) < 0.1;

            //Chart("Left foot on ground", legs.left.end.OnGround ? 1 : -1, 1);
            //Chart("Right foot on ground", legs.right.end.OnGround ? 1 : -1, 1.5f);
            Chart("left foot.Y", legs.left.end.Position.Y, 1f);

            forceFeetTogether = (standUp && !standingUp) || IsSitting; // (!standingUp || !walk);
            // Compute force to stand pelvis up
            if (standUp && (standingUp || sittingUp))
            {
                if (pelvisRelativeError > -0.1f)
                {
                    Vector3 pelvisOffset = (footMidpoint - pelvis);
                    pForce = pelvisOffset * pGain;
                    pelvisCentered = pelvisOffset.Length() / legLength < 0.1f;
                    //if (standingUp || (sittingUp && pelvisCentered && feetTogether))
                    pForce.Y = (legLength + GroundLevel - pelvis.Y)
                        * (walk ? 1f : 2f) * standGain;
                    //else
                    //  pForce.Y = 100;
                    //if (walk)
                    //    pForce += 1.0f*pelvisForward;
                }
                else
                {
                    WaitingFor("StandUpController: pelvis in air");
                    // Forcibly levitate the pelvis and spine
                    spineBottom.Velocity = spineJoint.Velocity = legs.left.root.Velocity = legs.right.root.Velocity =
                        legLength * 4f * Math.Max(0.5f, -pelvisRelativeError) * Vector3.UnitY;
                    pForce = -Physics.gVector;
                }
            }
            else
            {
                WaitingFor("StandUpController: Standing up or sitting up");
                pForce = Physics.gVector;
            }
        }
        #endregion

        #region Shoulder control
        Vector3 shoulderOffset;
        /// <summary>
        /// Measure of error between current torso/body forward and gazeDirection
        /// </summary>
        Vector3 faceGazeDirectionError;
        /// <summary>
        /// Gain for rotating torso toward gaze
        /// </summary>
        float faceGazeDirectionGain = 50f;
        /// <summary>
        /// Gain for leveling out shoulders
        /// </summary>
        float shoulderRestorationGain = 2000f;

        /// <summary>
        /// Apply torque to shoulders to face gaze direction.
        /// </summary>
        protected virtual void ShoulderController()
        {
            // Even out shoulders
            shoulderOffset = shoulderRestorationGain
                * MathUtil.ComponentInDirection(arms.left.root.Position - arms.right.root.Position, bodyUp);

#if OLDSHOULDERCONTROL
            // Align shoulders with gaze as well as possible
            if (Vector3.Dot(gazeDirection, pelvisForward) >= 0)
                // Gaze is ahead of pelvis; turn toward gaze
                faceForwardError = lookRight - bodyRight;
            // Gaze is behind character
            else if (Vector3.Dot(gazeDirection, pelvisRight) > 0)
                // On the right.  Turn as far as possible to the right
                faceForwardError = -pelvisForward - bodyRight;
            else
                // Turn as far as possible to the left
                faceForwardError = pelvisForward - bodyRight;
#else
            Vector3 targetDirection = gazeDirection;
            TwigObject faceTarget;
            float strength;
            Mammal.AttentionBuffer.Slot s;

            facingToward.Goal.FindWinner(out faceTarget, out strength, out s);
            if (strength > 0)
            {
                targetDirection = faceTarget.Position - Position;
                targetDirection.Y = 0;
                targetDirection.SafeNormalize();
            }
            float strength2;
            facingAwayFrom.Goal.FindWinner(out faceTarget, out strength2, out s);
            if (strength2 > 0 && strength2>strength)
            {
                // Note reversal of character and target, so this points away from target.
                targetDirection = Position-faceTarget.Position;
                targetDirection.Y = 0;
                targetDirection.SafeNormalize();
            }

            // Saturation is to prevent turning the spine past its joint limit
            float target = MathUtil.Saturate(PelvisAzmuth(targetDirection), (float)Math.PI/2.5f);
            float current = PelvisAzmuth(bodyForward);
            
            float delta = current - target;

            if (delta > (float)Math.PI)
                delta = delta - 2 * (float)Math.PI;

            Chart("delta", delta, 2*(float)Math.PI);
            Chart("target", target, 2 * (float)Math.PI);

            //Chart("torso-to-gaze angle error", delta, (float)Math.PI);

            // Don't try to turn toward it if we're already holding it or if we're hugging the parent
            if (IsHolding(gazeTarget) || hugging)
                delta = 0;

            // If it's too far away, don't try to turn there
            //if (Math.Abs(delta) > 0.8f * Math.PI)
                //delta = 0;

            faceGazeDirectionError = faceGazeDirectionGain * bodyForward * delta;
#endif
            faceGazeDirectionError -= MathUtil.ComponentInDirection(faceGazeDirectionError, bodyUp);
        }

        /// <summary>
        /// Tests whether creature is facing object.
        /// </summary>
        public bool FacingToward(TwigObject o)
        {
            if (o == null)
                return false;
            Vector3 offset = o.Position - footMidpoint;
            offset.Normalize();
            return Vector3.Dot(offset, bodyForward)> 0.75f;
        }

        /// <summary>
        /// The dot product of bodyforward and the vector to this object
        /// </summary>
        public float OrientationCosine(TwigObject o)
        {
            if (o == null)
                return 0;
            Vector3 offset = MathUtil.ProjectToGroundPlane(o.Position - footMidpoint);
            offset.Normalize();
            return Vector3.Dot(offset, bodyForward);
        }
        #endregion

        #region Pelvis control

        /// <summary>
        /// Force to rotating toward walkVector
        /// </summary>
        Vector3 faceWalkForce;
        /// <summary>
        /// Force for leveling hips
        /// </summary>
        Vector3 pelvisOffset;
        /// <summary>
        /// Gain for leveling hips
        /// </summary>
        float pelvisRestorationGain = -1000f;

        /// <summary>
        /// Apply torque to pelvis to make it face in the direction of walkVector (if walking) or gaze (otherwise).
        /// </summary>
        protected virtual void PelvisController()
        {
            // Align pelvis with walkVector, if walking, otherwise torso.
            //Vector3 lookParallelToGroundPlane = lookForward;
            Vector3 lookParallelToGroundPlane = bodyForward;
            lookParallelToGroundPlane.Y = 0;
            MathUtil.SafeNormalize(ref lookParallelToGroundPlane);
            if (lookParallelToGroundPlane == Vector3.Zero || IsHolding(gazeTarget))
                lookParallelToGroundPlane = bodyForward;
            float pelvisError = Vector3.Dot(pelvisRight,
                (sittingOrientation!=Vector3.Zero)?sittingOrientation:( walk ? walkDirection : lookParallelToGroundPlane));
            Debug.AssertProper(pelvisError, "Invalid error vector in pelvis control");
            faceWalkForce = faceWalkGain * pelvisError * pelvisForward;
            pelvisOffset = pelvisRestorationGain *
                MathUtil.ComponentInDirection(legs.left.root.Position - legs.right.root.Position,
                                              bodyUp);
        }
        #endregion

        #region Head control

        /// <summary>
        /// Move head to face gaze target.
        /// BUGs: instantaneous head saccades, looking backward
        /// </summary>
        protected virtual void HeadController()
        {
            if (hugging)
                lookForward = bodyForward;
            Vector3 lookSide = Vector3.Cross(lookForward, Vector3.Up);
            if (lookSide.LengthSquared() < 0.2f)
                // Singularity
                lookSide = bodyRight;
            Vector3 headDirection = Vector3.Cross(lookSide, lookForward);
            if (headDirection.Y < 0f)
                headDirection.Y = 0f;

            if (headDirection.LengthSquared() > 0.0)
                headDirection.Normalize();
            else
                headDirection = Vector3.Up;
            headTop.Position = spineTop.Position + headHeight * headDirection;
            headTop.Acceleration = new Vector3(0, 0, 0);
        }
        #endregion

        #region Arm control
        /// <summary>
        /// Efferent.  Reach arms toward approach target.
        /// </summary>
        public bool reach;
        /// <summary>
        /// Efferent lift arms to shoulder level.
        /// Used for reaching, hugging.
        /// </summary>
        public bool liftArms;
        /// <summary>
        /// Force to apply to swing arm.
        /// </summary>
        public float swingRightArm, swingLeftArm;
        float reachGain = 500f;
        /// <summary>
        /// Vector pointing in the direction to reach.
        /// Need not be a unit vector, but if reachDirection is zero, bodyForward is used instead.
        /// </summary>
        public Vector3 reachDirection=Vector3.Zero;

        /// <summary>
        /// Control arms for reaching, lifting, and swinging.
        /// </summary>
        protected void ArmController()
        {
            arms.left.lower.painMultiplier = arms.right.lower.painMultiplier =
                (grapple || punch) ? 20f : 1f;

            //Chart("reach", reach ? 1 : -1, 2.5f);
            if ((reach || grapple) && gazeTarget==approach.target)
            {
                System.Diagnostics.Debug.Assert(MathUtil.IsProperVector(bodyForward), "Body forward direction is an invalid vector");
                Vector3 reachUnit = (reachDirection.LengthSquared() > 0.01f) ? reachDirection : bodyForward;
                reachUnit.Normalize();
                Debug.Assert(MathUtil.IsProperVector(reachUnit), "Reach direction is an invalid vector");
                Vector3 leftHandTarget, rightHandTarget;
                Vector3 shoulderRight = arms.right.root.Position - arms.left.root.Position;
                leftHandTarget = arms.left.root.Position + armLength * reachUnit - 0.1f*shoulderRight;
                rightHandTarget = arms.right.root.Position + armLength * reachUnit + 0.1f*shoulderRight;
                float gain = (grapple?2:1)*reachGain;
                arms.left.end.Acceleration = (leftHandTarget - arms.left.end.Position) * gain;
                arms.right.end.Acceleration = (rightHandTarget - arms.right.end.Position) * gain;
            }
            else if (liftArms)
            {
                // Give the arms antigravity or they'll drag the parent's leg around.
                arms.left.end.Acceleration = arms.right.end.Acceleration = new Vector3(0, 10, 0);
                arms.left.joint.Acceleration = arms.right.joint.Acceleration = new Vector3(0, 10, 0);
                arms.left.root.Acceleration = arms.right.root.Acceleration = new Vector3(0, 10, 0);
            }
            else if (!reach && !hug && (swingLeftArm != 0 || swingRightArm != 0))
            {
                 arms.left.joint.Acceleration = 
                    arms.left.end.Acceleration = bodyForward * swingLeftArm;
                 arms.right.joint.Acceleration = 
                    arms.right.end.Acceleration = bodyForward * swingRightArm;
            }
            else
            {
                if (arms.left.CurrentPose==null)
                    arms.left.end.Acceleration = arms.left.joint.Acceleration = Physics.gVector;
                if (arms.right.CurrentPose == null)
                    arms.right.end.Acceleration = arms.right.joint.Acceleration = Physics.gVector;
            }
            swingLeftArm = swingRightArm = 0;
            reach = liftArms = false;

            TwigObject graspTarget;
            Mammal.AttentionBuffer.Slot graspSlot;
            float graspPriority;

            handInGraspingPosition.Goal.FindWinner(out graspTarget, out graspPriority, out graspSlot);
            if (graspPriority > 0)
            {
                float reachTime = 0.4f;

                Grasp chosenGrasp;
                Node leftTarget;
                Node rightTarget;
                graspTarget.SelectGrasp(this, Grasp.Any, out chosenGrasp, out leftTarget, out rightTarget);
                if (leftTarget != null)
                {
                    if (!arms.left.InBallisticMotion)
                    {
                        leftHandReachTarget = graspTarget;
                        leftHandReachNode = leftTarget;
                        arms.left.Limp();
                        arms.left.Goto(leftTarget.Position, reachTime, ControlMode.Free);
                    }
                }
                else
                {
                    leftHandReachTarget = null;
                    leftHandReachNode = null;
                }
                if (rightTarget != null)
                {
                    if (!arms.right.InBallisticMotion)
                    {
                        rightHandReachTarget = graspTarget;
                        rightHandReachNode = rightTarget;
                        arms.right.Limp();
                        arms.right.Goto(rightTarget.Position, reachTime, ControlMode.Free);
                    }
                }
                else
                {
                    rightHandReachTarget = null;
                    rightHandReachNode = null;
                }
            }
        }

        TwigObject leftHandReachTarget = null;
        TwigObject rightHandReachTarget = null;
        Node leftHandReachNode = null;
        Node rightHandReachNode = null;

        bool LeftHandInPositionToGrasp(TwigObject o)
        {
            if (leftHandReachTarget == null || leftHandReachTarget != o)
                return false;
            return !arms.left.InBallisticMotion 
                && MathUtil.DistanceLessThan(arms.left.end.Position, leftHandReachNode.Position, 0.8f);
        }

        bool RightHandInPositionToGrasp(TwigObject o)
        {
            if (rightHandReachTarget == null || rightHandReachTarget != o)
                return false;
            return !arms.right.InBallisticMotion
                && MathUtil.DistanceLessThan(arms.right.end.Position, rightHandReachNode.Position, 0.8f);
        }

        #endregion

        #region Foot control
        /// <summary>
        /// Efferent.  Move feet together in preparation for standing.
        /// </summary>
        public bool forceFeetTogether;
        float footTogetherGain = 100f;

        /// <summary>
        /// Control feel when not stepping.
        /// Stepping is controlled by GaitController().
        /// </summary>
        protected void FootController()
        {
            // Bring feet together in preparation for standing
            Vector3 footOffset = legs.left.end.Position - legs.right.end.Position;
            bool feetTogether = footOffset.LengthSquared() <= pelvisWidth * pelvisWidth * 1.1f;
            if (forceFeetTogether && !feetTogether)
            {
                if (!legs.left.InBallisticMotion)
                    legs.left.end.Acceleration = -footOffset * footTogetherGain;
                if (!legs.right.InBallisticMotion)
                    legs.right.end.Acceleration = footOffset * footTogetherGain;
            }
            else
            {
                if (!legs.left.InBallisticMotion)
                {
                    legs.left.end.Acceleration = Physics.gVector;
                    //legs.left.end.ControlMode = ControlMode.FallAndStick;
                }
                if (!legs.right.InBallisticMotion)
                {
                    legs.right.end.Acceleration = Physics.gVector;
                    //legs.right.end.ControlMode = ControlMode.FallAndStick;
                }
            }
            //Chart("Left foot planted", legs.left.end.OnGround ? 1 : -1, 1.5f);
            //Chart("Right foot planted", legs.right.end.OnGround ? 1 : -1, 1.3f);
#if DEBUG_POSTURE
            if (Physics.IsSettled)
            {
                Debug.Assert(legs.left.Extension <= 1.2, "Left leg overextended");
                Debug.Assert(legs.right.Extension <= 1.2, "Right leg overextended");
            }
#endif
        }
        #endregion

        #region Respiration
        /// <summary>
        /// Rate at which shoulders rise and fall.
        /// </summary>
        public float respirationRate = 0.1f;
        /// <summary>
        /// Distance traveled by shoulders when rising and falling for respiration.
        /// </summary>
        public float respirationAmplitude = 0.01f;
        /// <summary>
        /// Height shoulders are raised or dropped for respiration.
        /// </summary>
        public float diaphragmPosition = 0f;
        float respirationPhase = 0f;

        /// <summary>
        /// Adjust respiration phase (shoulder height) of character
        /// </summary>
        protected void RespirationController()
        {
            respirationPhase += 5*(respirationRate * Physics.deltaT); //%(float)(2f*Math.PI);
            diaphragmPosition = respirationAmplitude * height * (float)Math.Cos(respirationPhase);
        }
        #endregion
    }
}
