﻿// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A chair with a simple box for its collision volume
    /// </summary>
    public class SimpleChair : BoxModelSupportingSurface
    {
        float seatHeight;
        float totalHeight;
        float backDepth = 0.02f;
        Node sitTarget;

        /// <summary>
        /// Make a chair with a simple box for its collision volume
        /// </summary>
        public SimpleChair(TwigGame g, string name, Model m, float seatHeight, Vector3 bottomCenter)
            : base(g, name, m, seatHeight, bottomCenter)
        {
            // This is inefficient because we're computing the BBox twice, but it's still a rare operation
            totalHeight = m.BoundingBox().Height;
            this.seatHeight = seatHeight;
            sitTarget = new Node(this, name + "-sit-point");
            collisionElements = new CollisionElement[] { collisionElements[0], new ChairBackCollisionElement(this) };
        }

        /// <summary>
        /// Make a chair with a simple box for its collision volume
        /// </summary>
        public SimpleChair(TwigGame g, string name, string modelName, float seatHeight, Vector3 bottomCenter)
            : this(g, name, g.Content.Load<Model>(modelName), seatHeight, bottomCenter)
        { }

        /// <summary>
        /// Update the nodes' positoins.
        /// </summary>
        public override void Update(GameTime gameTime)
        {
            // Kluge: force the sitPoint's position
            sitTarget.Position = TopPosition;
            sitTarget.Velocity = Vector3.Zero;
            base.Update(gameTime);
        }

        /// <summary>
        /// Approach from the front (+Z direction)
        /// </summary>
        public override Vector3 DockingDirection
        {
            get
            {
                return LocalZ;
            }
        }

        /// <summary>
        /// Close to front of chair
        /// </summary>
        public override float DockingDistance
        {
            get
            {
                return depth/2;
            }
        }

        /// <summary>
        /// Return whether this object can be sat on.  Since this is a chair, the answer is always true.
        /// </summary>
        public override bool IsSittable
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Point to sit at (top of the seat, in this case)
        /// </summary>
        public override Node SittingTarget
        {
            get
            {
                return sitTarget;
            }
        }

        class ChairBackCollisionElement : CollisionElement
        {
            SimpleChair me;
            float backHeight;
            float backElevation;
            public ChairBackCollisionElement(SimpleChair me)
                : base(TypeCode.OBB, me)
            {
                this.me = me;
                backHeight = me.totalHeight - me.seatHeight;
                backElevation = me.seatHeight + backHeight * 0.5f;
            }

            public override float Mass
            {
                get { return me.mass; }
            }

            public override float InvMass
            {
                get { return 1 / me.mass; }
            }

            protected override GeometryPrimitives.OBB OBBGeometry
            {
                get
                {
                    return new GeometryPrimitives.OBB(me.Position+backElevation*me.LocalY-0.5f*me.depth*me.LocalZ,
                        me.LocalX, me.LocalY, me.LocalZ,
                        me.width * 0.5f, me.height * 0.5f, me.backDepth);
                }
            }

            public override Vector3 Velocity
            {
                get { return me.Velocity; }
            }

            public override float PainMultiplier
            {
                get { return 1; }
            }

            protected override void HandleCollision(CollisionElement collider, Vector3 contactPoint, Vector3 contactNormal, float penetrationDepth)
            {
                penetrationDepth = Math.Min(me.width, Math.Min(me.height, Math.Min(me.depth, penetrationDepth)));
                Vector3 offset = contactNormal * penetrationDepth * InvMass / (InvMass + collider.InvMass);
                foreach (Node n in me.nodes)
                    n.Position += offset;
            }
        }

    }
}