// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

using SlotSet = Twig.Mammal.AttentionBuffer.SlotSet;
using Slot = Twig.Mammal.AttentionBuffer.Slot;

namespace Twig
{
    partial class Mammal
    {
        TwigObject sittingSurface;

        /// <summary>
        /// Attach rump of character to this node.
        /// </summary>
#if OldSitOn
        public void SitOn(TwigObject o)
        {
            sittingSurface = o;
            if (o == null)
                standUpFromSittingPosition = true;
            else
            {
                standUp = false;
                legs.left.end.Acceleration = legs.right.end.Acceleration = Physics.gVector;
                legs.left.end.ControlMode = legs.right.end.ControlMode = ControlMode.Free;
            }
        }
#else
        public Action SitOn(TwigObject o)
        {
            if (o.IsSittable)
                return new GoalAchievementAction(sittingOn, o);
            else
                return new FailAction();
        }
#endif

        /// <summary>
        /// True when we've been sitting and are trying to stand up.
        /// </summary>
        protected bool standUpFromSittingPosition = false;
        float sitSpeed = 2f;
        /// <summary>
        /// Used by PelvisController() to orient the body while sitting.
        /// </summary>
        Vector3 sittingOrientation;
        void SittingControl()
        {
#if notdef
            Printf(inSittingPosition.DebugString);
            Printf(inSittingPosition.Goal.DebugString);
            Printf(facingAwayFrom.DebugString);
            Printf(facingAwayFrom.Goal.DebugString);
            Printf(sittingOn.DebugString);
            Printf(sittingOn.Goal.DebugString);
#endif
            // Check sittingOn.Goal to see if we should sit someplace.
            TwigObject tobject;
            float strength;
            Slot s;

            // Check whether we want to sit someplace else, and if so, stand up.
            inSittingPosition.Goal.FindWinner(out tobject, out strength, out s);
            if (strength > 0 && tobject != sittingSurface)
            {
                sittingSurface = null;
                standUp = true;
            }

            sittingOn.Goal.FindWinner(out tobject, out strength, out s);
            if (strength > 0 && inSittingPosition[s])
            {
                // We're ready to sit on it
                sittingSurface = tobject;
                standUp = false;
                legs.left.end.Acceleration = legs.right.end.Acceleration = Physics.gVector;
                legs.left.end.ControlMode = legs.right.end.ControlMode = ControlMode.Free;
            }

            if (sittingSurface == null)
                sittingOrientation = Vector3.Zero;

            // SITTING DOWN
            if (sittingSurface != null && sittingSurface.SittingTarget != null)
            {
                Node target = sittingSurface.SittingTarget;
                Vector3 targetOffset = target.Position - spineBottom.Position;
                // Orient the pelvis away from the sitting point
                sittingOrientation = MathUtil.SafeNormalize(-targetOffset);
                if (targetOffset.LengthSquared() > MathUtil.Square(sittingSurface.SittingTargetRadius))
                    spineBottom.Velocity = MathUtil.SetLength(targetOffset, sitSpeed);
                if (legs.left.joint.Position.Y < target.Position.Y)
                {
                    Vector3 temp = legs.left.joint.Position;
                    temp.Y = target.Position.Y;
                    legs.left.joint.Position = temp;
                }
                if (legs.right.joint.Position.Y < target.Position.Y)
                {
                    Vector3 temp = legs.right.joint.Position;
                    temp.Y = target.Position.Y;
                    legs.right.joint.Position = temp;
                }
            }
            // STANDING UP AGAIN
            else if (standUpFromSittingPosition)
            {
                sittingOrientation = Vector3.Zero;
                if (comOffset.LengthSquared() > MathUtil.Square(legs.left.upper.Length * 0.6f))
                {
                    // Can't stand yet, move feet closer
                    Vector3 footOffset = spineBottom.Position - legs.left.end.Position;
                    footOffset.Y = 0;
                    if (footOffset.Length() > legs.left.upper.Length * 0.4)
                    {
                        if (legs.left.Extension > 0.7)
                        {
                            legs.left.joint.Velocity = Vector3.Up;
                            legs.left.end.ControlMode = ControlMode.Free;
                        }
                        else
                        {
                            legs.left.end.Velocity = MathUtil.SetLength(footOffset, sitSpeed);
                            legs.left.end.ControlMode = ControlMode.Free;
                        }
                    }
                    footOffset = spineBottom.Position - legs.right.end.Position;
                    footOffset.Y = 0;
                    if (footOffset.Length() > legs.right.upper.Length * 0.4)
                    {
                        if (legs.right.Extension > 0.7)
                        {
                            legs.right.joint.Velocity = Vector3.Up;
                            legs.right.end.ControlMode = ControlMode.Free;
                        }
                        else
                        {
                            legs.right.end.Velocity = MathUtil.SetLength(footOffset, sitSpeed);
                            legs.right.end.ControlMode = ControlMode.Free;
                        }
                    }
                }
                else
                {
                    // Feet are close enough
                    standUp = true;
                    standUpFromSittingPosition = false;
                    legs.left.end.ControlMode = ControlMode.FallAndStick;
                    legs.right.end.ControlMode = ControlMode.FallAndStick;
                }
            }
        }

        /// <summary>
        /// True if character is sitting down.
        /// </summary>
        public bool IsSitting
        {
            get
            {
                return sittingSurface != null
                    && MathUtil.Difference(sittingSurface.SittingTarget.Position.Y, spineBottom.Position.Y) < 0.1f;
            }
        }

        /// <summary>
        /// True if the character is sitting on that object.
        /// </summary>
        public Predicate sittingOn;
        /// <summary>
        /// True if the character is in position to sit on the object in question.
        /// </summary>
        public Predicate inSittingPosition;

        void InitializeSitting()
        {
            sittingOn = new Predicate("sittingOn", this,
                delegate()
                {
                    if (IsSitting)
                    {
                        return SlotSet.Empty;
                    }
                    else
                    {
                        int sitslot = Attention.FastObjectSlotNumber(sittingSurface);
                        if (sitslot >= 0)
                            return new SlotSet((byte)(1 << sitslot));
                        else
                            return SlotSet.Empty;
                    }
                });
            inSittingPosition = new Predicate("inSittingPosition", this,
                delegate(Slot s)
                {
                    return dockingDistance[s] < 1;
                });

            sittingOn.DeclarePreconditions(inSittingPosition, facingAwayFrom);

            ApproachParameters sitApproachParameters = new ApproachParameters();
            approachController.Add(new PredicateAchievingBehavior<ApproachParameters>(inSittingPosition, false,
                delegate(TwigObject o)
                {
                    if (o == null)
                        sitApproachParameters.Approach(null);
                    else
                        sitApproachParameters.Approach(o, o.DockingDistance, o.DockingDirection);
                    return sitApproachParameters;
                }));
        }
    }
}
