// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A tree stump (i.e. a short cylinder with a height suitable for sitting).
    /// </summary>
    public class Stump : TwigObject
    {
        Link body;
        const float height = 0.6f;
        const float radius = 0.25f;
        const float mass = height * 2 * (float)Math.PI * radius * radius * 100 * 100 * 100 / 1000;

        static int stumpCounter;
        /// <summary>
        /// Makes a new name of the form "Stump n" where n is a unique integer.
        /// </summary>
        static string Gensym()
        {
            return "Stump" + stumpCounter++.ToString();
        }

        /// <summary>
        /// Makes a new stump with a generated name.
        /// </summary>
        public Stump(TwigGame g, Vector3 basePosition)
            : this(g, Gensym(), basePosition)
        { }

        /// <summary>
        /// Makes a new stump with a generated name.
        /// </summary>
        public Stump(TwigGame g, string name, Vector3 basePosition)
            : base(g, name, false)
        {
            body = new Link(name, this, new Node(this, name+" base", basePosition), new Node(this, name+" top", basePosition + height * Vector3.Up), height, radius);
            body.node1.Mass = 999f;
            body.node2.Mass = 999f;
            body.node1.ControlMode = ControlMode.Fixed;
            body.node2.ControlMode = ControlMode.Fixed;
            Color = new Color(150, 70, 20);
            Dynamic = false;
            collisionElements = new CollisionElement[] { };
            nodes = new Node[] { body.node1, body.node2 };
        }

        /// <summary>
        /// Draws base and top of stump
        /// </summary>
        /// <param name="t"></param>
        public override void Draw(GameTime t)
        {
            BasicEffect.Begin();
            foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                ConfigureMaterial(); // Don't need to CommitChanges() because that will be done in body.Draw().
                body.Draw();
                pass.End();
            }
            BasicEffect.End();
        }

        /// <summary>
        /// Position of the stump
        /// </summary>
        public override Vector3 Position
        {
            get { return body.node1.Position; }
        }

        /// <summary>
        /// The object's current center of mass in world coordinates
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get { return Position+Vector3.UnitY*height*0.5f; }
        }

        /// <summary>
        /// The object's total mass
        /// </summary>
        public override float Mass
        {
            get { return mass; }
        }

        /// <summary>
        /// Returns the position to look at on the stump.
        /// The lookpoint is in the stumptops
        /// </summary>
        public override Vector3 LookPoint
        {
            get { return body.node2.Position; }
        }

        /// <summary>
        /// The radius of the collision cylinder for the stump
        /// </summary>
        public override float CollisionRadius
        {
            get { return radius; }
        }

        /// <summary>
        /// True if the point is in, on, or directly over the stump
        /// </summary>
        public override bool IsSupporting(Vector3 point)
        {
            Vector3 delta = point - Position;
            delta.Y = 0;
            return delta.LengthSquared() <= (radius * radius);
        }

        /// <summary>
        /// True if supporting some node of o.
        /// </summary>
        public override bool IsSupporting(TwigObject o)
        {
            return SupportingSurfaceDefaultIsSupportingMethod(o);
        }

        /// <summary>
        /// Height at which MerryGoRound supports other objects
        /// </summary>
        public override float SupportingSurfaceLevel
        {
            get
            {
                return height;
            }
        }

        /// <summary>
        /// This is a supporting surface
        /// </summary>
        public override bool IsASupportingSurface
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Sitting target it node for the top of the body.
        /// </summary>
        public override Node SittingTarget
        {
            get
            {
                return body.node2;
            }
        }

        /// <summary>
        /// Sitting target radius is half the stump's overall radius
        /// </summary>
        public override float SittingTargetRadius
        {
            get
            {
                return radius/2f;
            }
        }

        /// <summary>
        /// Nothing to constrain
        /// </summary>
        public override void Constrain()
        { }
    }
}
