// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// Displays white titles on a black background
    /// </summary>
    public class TitleScreen : HUD, IDrawable
    {
        /// <summary>
        /// A title screen that displays white titles on a black background.
        /// </summary>
        public TitleScreen(TwigGame g, string name)
            : base(g)
        {
            g.NamedObjects[name] = this;
            background = new Texture2D(g.Device, 1, 1, 1, TextureUsage.Tiled, SurfaceFormat.Color);
            background.SetData<Color>(new Color[] { Color.Black });
            titleColor = Color.White;
        }

        float alpha = 255;
        Texture2D background;
        string text;
        Vector2 textPosition = new Vector2(100, 200);
        Color titleColor;
        bool forceDisplay = false;

        void Transparent()
        {
            forceDisplay = true;
            alpha = 255;
            text = null;
        }

        void Off()
        {
            forceDisplay = false;
            alpha = 0;
        }

        /// <summary>
        /// Draw a black background and any titles displayed over it
        /// </summary>
        void IDrawable.Draw(GameTime gameTime)
        {
            if (forceDisplay || alpha > 0)
            {
                HUDEffect.Begin();
                foreach (EffectPass pass in HUDEffect.CurrentTechnique.Passes)
                {
                    pass.Begin();
                    HUDSprites.Begin();
                    if (!forceDisplay)
                        HUDSprites.Draw(background, new Rectangle(0, 0, Device.Viewport.Width, Device.Viewport.Height),
                                        new Color(0, 0, 0, (byte)alpha));
                    //Device.Clear(new Color(0, 0, 0, (byte)alpha));
                    if (text != null)
                        HUDSprites.DrawString(TwigGame.DialogFont, text, textPosition,
                            new Color(titleColor.R, titleColor.B, titleColor.G, (byte)alpha));
                    HUDSprites.End();
                    pass.End();
                }
                HUDEffect.End();
            }
            else
                text = null;
            base.Draw(gameTime);
        }

        int IDrawable.DrawOrder
        {
            get
            {
                return HUD.HUDDrawOrder+1;
            }
        }

        /// <summary>
        /// Display specified text
        /// </summary>
        public void Say(string text)
        {
            this.text = text;
        }

        /// <summary>
        /// Display specified text
        /// </summary>
        public void Say(string text, float x, float y)
        {
            textPosition = new Vector2(x, y);
            this.text = text;
        }

        /// <summary>
        /// Fades the title screen to be opaque, making the other objects invisible.
        /// </summary>
        public Action FadeToBlack(float time)
        {
            return new FadeAction(this, 255f / Physics.SecondsToTicks(time));
        }

        /// <summary>
        /// Fades the title screen out
        /// </summary>
        public Action FadeFromBlack(float time)
        {
            return new FadeAction(this, -255f / Physics.SecondsToTicks(time));
        }

        class FadeAction : Action
        {
            float delta;
            bool done;
            TitleScreen me;

            public FadeAction(TitleScreen me, float delta)
            {
                this.me = me;
                this.delta = delta;
                done = true;
            }

            public override void Abort()
            {
                done = true;
            }

            public override void Start()
            {
                done = false;
            }

            public override void Update()
            {
                me.alpha += delta;
                if (me.alpha < 0)
                {
                    me.alpha = 0;
                    done = true;
                }
                else if (me.alpha > 255)
                {
                    me.alpha = 255;
                    done = true;
                }
            }

            public override ActionStatus Status
            {
                get { return done?ActionStatus.Succeed:ActionStatus.Running; }
            }
        }
    }
}
