// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A (passive) tree
    /// </summary>
    public class Tree : TwigObject
    {
        Vector3 basePosition;
        Link body;
        static SimpleMesh top;
        const float height = 3f;
        const float baseRadius = 0.25f;
        const float topRadius = 1.5f;
        const float mass = height * 2 * (float)Math.PI * baseRadius * baseRadius * 100 * 100 * 100 / 1000;

        static int treeCounter;
        /// <summary>
        /// Makes a new name of the form "Tree n" where n is a unique integer.
        /// </summary>
        static string Gensym()
        {
            return "Tree " + treeCounter++.ToString();
        }

        /// <summary>
        /// Makes a new tree with a generated name.
        /// </summary>
        public Tree(TwigGame g, Vector3 basePosition)
            : this(g, Gensym(), basePosition)
        {
            collisionElements = new CollisionElement[] { body.CollisionElement };
        }

        /// <summary>
        /// Makes a new tree with a generated name.
        /// </summary>
        public Tree(TwigGame g, string name, Vector3 basePosition)
            : base(g, name, false)
        {
            this.basePosition = basePosition;
            body = new Link(name, this, new Node(this, name+" base", basePosition), new Node(this, name+" top", basePosition + height * Vector3.Up), height, baseRadius);
            body.node1.Mass = 999f;
            body.node2.Mass = 999f;
            body.node1.ControlMode = ControlMode.Fixed;
            body.node2.ControlMode = ControlMode.Fixed;

            if (top == null)
            {
                top = SimpleMesh.Sphere(g.Device, topRadius, 2, 0.9f);
            }
            Color = new Color(150, 70, 20);
            Dynamic = false;
        }

        /// <summary>
        /// Nothing to constrain
        /// </summary>
        public override void Constrain()
        { }

        /// <summary>
        /// Draws base and top of tree
        /// </summary>
        /// <param name="t"></param>
        public override void Draw(GameTime t)
        {
            BasicEffect.Begin();
            foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                ConfigureMaterial(); // Don't need to CommitChanges() because that will be done in body.Draw().
                body.Draw();
                BasicEffect.DiffuseColor = Color.Green.ToVector3();
                BasicEffect.AmbientLightColor = Color.Green.ToVector3() * 0.5f;
                BasicEffect.World = Matrix.CreateTranslation(basePosition + new Vector3(0, topRadius + height - 0.3f, 0));
                BasicEffect.CommitChanges();
                top.Draw();
                pass.End();
            }
            BasicEffect.End();
        }

        /// <summary>
        /// Position of the tree
        /// </summary>
        public override Vector3 Position
        {
            get { return basePosition; }
        }

        /// <summary>
        /// The object's current center of mass in world coordinates
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get { return Position+Vector3.UnitY*height; }
        }

        /// <summary>
        /// The object's total mass
        /// </summary>
        public override float Mass
        {
            get { return mass; }
        }

#if OldCollision
        /// <summary>
        /// Tests collisions of the tree with a ragdoll
        /// </summary>
        /// <param name="h"></param>
        /// <returns></returns>
        public override bool TestCollisionWithRagdoll(Ragdoll h)
        {
            Vector3 offset = h.Position - basePosition;
            bool contact = false;
            // Don't bother if they're too far apart
            if (offset.Length() < CollisionRadius+h.CollisionRadius)
            {
                    foreach (Link hl in h.Links)
                        if (hl.CheckCollision(body, this))
                            contact = true;
            }
            return contact;
        }

        /// <summary>
        /// Tests collisions of the tree with a ball
        /// </summary>
        /// <param name="b"></param>
        /// <returns></returns>
        public override bool TestCollisionWithBall(Ball b)
        {
            return b.TestCollisionWithLink(body);
        }
#endif

        /// <summary>
        /// Returns the position to look at on the tree.
        /// The lookpoint is in the treetops
        /// </summary>
        public override Vector3 LookPoint
        {
            get { return basePosition+new Vector3(0, height, 0); }
        }

        /// <summary>
        /// The radius of the collision cylinder for the tree
        /// </summary>
        public override float CollisionRadius
        {
            get { return baseRadius; }
        }
    }
}
