// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Twig
{
    /// <summary>
    /// Base class of TwigObjects with animal-like behavior
    /// </summary>
    partial class Mammal
    {
        /// <summary>
        /// Coefficient of repulsion for avoiding negatively-valenced objects
        /// </summary>
        public float navigationAvoidance = 2f;

        #region Walking-specific control variables
        /// <summary>
        /// Enables display of target velocity vector for walking system
        /// </summary>
        public static bool showWalkVector = false;

        /// <summary>
        /// Enables walking.
        /// </summary>
        public bool walk = false;
        /// <summary>
        /// Target velocity vector for walking system.
        /// </summary>
        public Vector3 walkVector = new Vector3(-2, 0, 0);
        #endregion


        #region Intermediate values computed for walking

        /// <summary>
        /// Scalar speed at which the character is walking.
        /// May be less than the magnitude of walkVector if walkVector is faster than
        /// the creature can move.
        /// </summary>
        public float walkSpeed;
        /// <summary>
        /// Unit vector in the direction of the creature's motion.
        /// </summary>
        public Vector3 walkDirection;
        /// <summary>
        /// The actual velocity vector for the body (limited by the character's maximum speed)
        /// </summary>
        public Vector3 realWalkVector;
        #endregion

        #region Walking-specific gains and other parameters

        /// <summary>
        /// Gain for rotating pelvis roward walkVector
        /// </summary>
        public float faceWalkGain = 20f;
        /// <summary>
        /// Height creature raises foot to when taking a step.
        /// </summary>
        public float stepTargetHeight = 0.1f;
        /// <summary>
        /// Distance ahead of body foot moves during a step.
        /// </summary>
        public float stride = 0.75f;
        /// <summary>
        /// Duration of a single step.
        /// </summary>
        public float ballisticStepTime = 0.2f;
        #endregion 

        /// <summary>
        /// Selects between locomotion behaviors to determine walkVector.
        /// Initialized in Mammal constructor
        /// </summary>
        MaxBehavior<Vector3> locomotionController;

        // Information that's used outside the walking system, so needs to
        // be computed before the walking code actually runs
        void ComputeWalkingParameters()
        {
            walkDirection = walkVector = locomotionController.Activation > 0 ? locomotionController.MotorOutput : Vector3.Zero;
            //CheckChatter("walkVector", walkVector, 0.5f, 0.005f);
            System.Diagnostics.Debug.Assert(MathUtil.IsProperVector(walkVector));
            walkSpeed = Math.Min(maxSpeed, Math.Max(1f, walkVector.Length()));
            //CheckChatter("Walk speed", walkSpeed, 0.25f, 0.01f);
            realWalkVector = walkDirection * walkSpeed;
            

            MathUtil.SafeNormalize(ref walkDirection);

            Debug.Assert(legLength != 0, "legLength is 0");
            stride = Math.Min(0.5f * walkSpeed / legLength, legLength / 2f);
            Debug.Assert(walkSpeed != 0, "walkSpeed is 0");
            ballisticStepTime = Math.Min(0.25f, 2f * stride / walkSpeed);

            // Can't walk if walkVector is zero
            bool localMinimum = (walkVector == Vector3.Zero);
            // withdrawl from pain overrides all other considerations, like looking where we're going
            bool withdrawingFromPain = (locomotionController.maxBehavior == painWithdrawalReflex);
            // Need to look where we're going unless we're walking slowly
            bool slow = (walkSpeed <= 1);
            bool lookingWhereWalking = (Vector3.Dot(approach.targetDirection, lookForward) > 0)
                || approach.targetDirection==Vector3.Zero;
            //bool facingWhereWalking = (Vector3.Dot(walkDirection, bodyForward) > 0);
            //walk = !localMinimum && (withdrawingFromPain || (!hugging && (slow || lookingWhereWalking) && facingWhereWalking));
            walk = !localMinimum && (withdrawingFromPain || (!hugging && (slow || lookingWhereWalking)
                ));
#if DEBUG
            if (locomotionController.Activation > 0 && locomotionController.MotorOutput != Vector3.Zero && !walk)
            {
                if (localMinimum)
                    WaitingFor("Escape local minimum");
                if (!lookingWhereWalking)
                    WaitingFor("Look where walking");
            }
#endif
            Chart("walk", walk ? 1 : -1, 2);
            //&& (slow || facingWhereWalking)));
#if old
            //walk = approach.target != null &&
            walk = walkVector != Vector3.Zero
                && (locomotionController.maxBehavior == painWithdrawlReflex ||
                    (!hugging
                    && (walkSpeed <= 1 || 
                    // true)));
                         (Vector3.Dot(approach.targetDirection, lookForward) > 0)) && Vector3.Dot(walkDirection, bodyForward)>0));
#endif
        }

        /// <summary>
        /// Threshold distance below which the creature will actively avoid this object.
        /// If the object is farther than this, it will be ignored for purposes of navigation
        /// </summary>
        /// <param name="o">obstacle TwigObject</param>
        /// <returns>Threshold distance</returns>
        protected float AvoidDistance(TwigObject o)
        {
            return Math.Max(0f, -navigationAvoidance * Appraisal(o).valence);
        }

        /// <summary>
        /// Controls legs based on walk and walkVector.
        /// </summary>
        protected abstract void GaitControl();

        /// <summary>
        /// Posture control that's still active when not walking.
        /// </summary>
        protected abstract void StaticBalance();

        /// <summary>
        /// Behavior that stops walking when active.
        /// This is the default behavior and only runs when nothing else does.
        /// </summary>
        Behavior<Vector3> freezeLocomotion;
        /// <summary>
        /// Behavior that stops walking when active.
        /// This is the default behavior and only runs when nothing else does.
        /// </summary>
        class FreezeLocomotion : Behavior<Vector3>
        {
            public override string Name
            {
                get
                {
                    return "Freeze";
                }
            }

            protected override void Update()
            {
                this.mActivation = TwigObject.AllCharactersFreeze ? 1000 : 0;
            }
        }
    }
}
