// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using Microsoft.Xna.Framework;
using System.Reflection;

namespace Twig
{
    public class TwigServer : ScriptScriptInterpreter
    {
        public const int DefaultPort = 10001;

        Socket serverSocket;
        Socket clientConnection = null;
        SocketAsyncEventArgs pendingAccept = null;
        NetworkStream readerStream;

        public delegate void TraceHandler(string message);
        public event TraceHandler TraceMessage;

        [System.Diagnostics.Conditional("DEBUG")]
        void Trace(string message)
        {
            if (TraceMessage != null)
                TraceMessage(message);
        }

        public TwigServer(TwigGame g) : this(g, DefaultPort) { }

        public TwigServer(TwigGame g, int port) : base(g)
        {
            g.NamedObjects["server"] = this;
            messageControl = delegate(Action a, bool start)
            {
                if (start)
                {
                    output.Write("{0}: Message ", a.requestID);
                    return output;
                }
                else
                {
                    output.WriteLine();
                    return null;
                }
            };
            serverSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint endpoint = new IPEndPoint(IPAddress.Any, port);
            serverSocket.Bind(endpoint);
            serverSocket.Listen(1);
        }

        public override void Update(GameTime gameTime)
        {
            if (clientConnection != null)
            {
                if (clientConnection.Connected) // && IckyTestConnectionState())
                    ServiceClient();
                else
                {
                    Trace("Client disconnected");
                    try
                    {
                        clientConnection.Close();
                    }
                    finally
                    {
                        clientConnection = null;
                        ListenForConnection();
                    }
                }
            }
            else if (pendingAccept == null)
                ListenForConnection();
        }

        void ListenForConnection()
        {
            Trace("Waiting for connection");
            pendingAccept = new SocketAsyncEventArgs();
            pendingAccept.AcceptSocket = null;
            pendingAccept.Completed += delegate(object sender, SocketAsyncEventArgs e)
            {
                ProcessAccept(e);
            };
            bool willRaiseAccept = serverSocket.AcceptAsync(pendingAccept);
            if (!willRaiseAccept)
                ProcessAccept(pendingAccept);
        }

        void ProcessAccept(SocketAsyncEventArgs e)
        {
            Socket s = e.AcceptSocket;
            pendingAccept = null;
            readerStream = new NetworkStream(s, FileAccess.Read);
            output = new StreamWriter(new NetworkStream(s, FileAccess.Write));
            output.WriteLine();
            output.WriteLine("TwigServer");
            output.WriteLine(System.Reflection.Assembly.GetAssembly(typeof(TwigGame)).FullName);
            output.Flush();
            clientConnection = s;
            Trace("Connected");
        }

        int testConnectionCounter = 0;
        bool IckyTestConnectionState()
        {
            // This routine is going to get called 60 times a second and we don't really
            // want to do an extra 60 system calls.  So just do it once a second and lie
            // and say we're connected the rest of the time.
            if ((testConnectionCounter++ % 63) != 0)
                return true;

            bool stillConnected = true;
            // Cribbed from Microsoft's documentation
            // This is how you can determine whether a socket is still connected.
            bool blockingState = serverSocket.Blocking;
            try
            {
                byte[] tmp = new byte[1];

                serverSocket.Blocking = false;
                serverSocket.Send(tmp, 0, 0);
            }
            catch (SocketException e)
            {
                // 10035 == WSAEWOULDBLOCK
                stillConnected = e.NativeErrorCode.Equals(10035);
            }
            finally
            {
                serverSocket.Blocking = blockingState;
            }
            return stillConnected;
        }

        byte[] linebuffer = new byte[512];
        int bufferPosition = 0;

        TextReader MaybeReadLine()
        {
            while (readerStream.DataAvailable)
            {
                int c = readerStream.ReadByte();
                switch (c)
                {
                    case '\n':
                    case '\r':
                        if (bufferPosition == 0)
                            // We have either a blank line or the cr after an lf.  Ignore it.
                            continue;
                        // We have a newly completed line
                        int length = bufferPosition;
                        bufferPosition = 0;
                        Trace(new StreamReader(new MemoryStream(linebuffer, 0, length)).ReadLine());
                        return new StreamReader(new MemoryStream(linebuffer, 0, length));

                    case '\b': // Control-H/backspace
                    case '\x7f':   // Delete
                        bufferPosition = Math.Max(bufferPosition - 1, 0);
                        break;

                    default:
                        linebuffer[bufferPosition++] = (byte)c;
                        break;
                }
            }
            return null;
        }

        void ServiceClient()
        {
            currentRequestID = -1;
            bool completed = false;
            TextReader r = MaybeReadLine();
            if (r != null)
                try
                {
                    //string command = reader.ReadLine();
                    //TextReader r = new StringReader(command);
                    if (char.IsDigit((char)r.Peek()))
                    {
                        currentRequestID = (int)Math.Round((double)ReadNumber(r));
                        SkipWhiteSpace(r);
                    }
                    Execute(r);
                    if (blockingAction == null)
                        completed = true;
                    else
                    {
                        // We inherit from ScriptScriptInterpreter which thinks it's interpreting a linear script
                        // and can only block on one action at a time.  So if we see it's waiting on an action, we
                        // add it to out list of pending actions and then clear it.
                        pendingActions.Add(new PendingAction(currentRequestID, blockingAction));
                        blockingAction = null;
                    }
                }
                catch (Exception e)
                {
                    if (e is System.Reflection.TargetInvocationException)
                        e = e.InnerException;
                    output.WriteLine("{0}: fail {1}: {2}", currentRequestID, e.GetType().Name, e.Message);
                    output.Flush();
                }
            if (completed)
            {
                output.WriteLine("{0}: succeed", currentRequestID);
                output.Flush();
            }
            CheckPendingActions();
        }

        void CheckPendingActions()
        {
            // We shouldn't have to check this, but output.Flush sometimes gives a
            // NullReferenceException if it's called when there's no output pending.
            if (pendingActions.Count > 0)
            {
                for (int i = pendingActions.Count - 1; i >= 0; i--)
                {
                    PendingAction a = pendingActions[i];
                    ActionStatus status = a.action.Status;
                    if (status != ActionStatus.Running)
                    {
                        output.WriteLine("{0}: {1}", a.requestID, status == ActionStatus.Succeed ? "succeed" : "fail");
                        pendingActions.RemoveAt(i);
                    }
                    else
                        a.action.Update();
                }
                output.Flush();
            }
        }

        class PendingAction
        {
            public Action action;
            public int requestID;
            public PendingAction(int id, Action a)
            {
                requestID = id;
                action = a;
            }
        }

        List<PendingAction> pendingActions = new List<PendingAction>();

        Action Report(string fieldName)
        {
            return new ReportAction(fieldName, false);
        }
        
        Action Report(string fieldName, bool includeStatic)
        {
            return new ReportAction(fieldName, includeStatic);
        }

        Action Report(string fieldName, bool includeStatic, float updateTime)
        {
            return new ReportAction(fieldName, includeStatic, updateTime);
        }

        class ReportAction : Action
        {
            string fieldName;
            int intervalTicks;
            bool includeStatic;
            bool onceOnly;
            bool doneOne = false;

            public ReportAction(string fieldName, bool includeStatic)
            {
                this.fieldName = fieldName;
                this.includeStatic = includeStatic;
                this.intervalTicks = 1;
                this.onceOnly = true;
            }

            public ReportAction(string fieldName, bool includeStatic, float time)
            {
                this.fieldName = fieldName;
                this.includeStatic = includeStatic;
                this.intervalTicks = Physics.SecondsToTicks(time);
                this.onceOnly = false;
            }

            public override void Update()
            {
                if (onceOnly || (Physics.tickCount % intervalTicks) == 0)
                {
                    object[] empty = new object[0];

                    System.IO.TextWriter output = StartMessage();
                    for (int i = 0; i < TwigObject.AllTwigObjects.Count; i++)
                    {
                        TwigObject o = TwigObject.AllTwigObjects[i];
                        if (includeStatic || o.Dynamic)
                        {
                            output.Write(o.Name);
                            output.Write("=");
                            object val = o.GetType().InvokeMember(fieldName,
                                BindingFlags.GetField | BindingFlags.GetProperty |
                                BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.IgnoreCase,
                                null, o, empty);
                            if (val is Vector3)
                            {
                                Vector3 v = (Vector3)val;
                                output.Write("(");
                                output.Write(v.X);
                                output.Write(" ");
                                output.Write(v.Y);
                                output.Write(" ");
                                output.Write(v.Z);
                                output.Write(")");
                            } else
                                output.Write(val);
                            output.Write(" ");
                        }
                    }
                    EndMessage();
                }
                doneOne = true;
            }

            public override ActionStatus Status
            {
                get { return (onceOnly&&doneOne)?ActionStatus.Succeed:ActionStatus.Running; }
            }
        }
    }
}
